/******************************************************************
 *
 *  File:   fiona_ioc_hal.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   12/07/05
 *
 * Copyright 2005, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *      HAL layer for interface to IOC chip
 *
 ******************************************************************/
#include <linux/autoconf.h>
#include <linux/types.h>
#include <asm/fcntl.h>
#include <asm/unistd.h>
#include <linux/string.h>
#include <linux/ioctl.h>
#include <linux/termios.h>
#include <linux/list.h>
#include <linux/syscalls.h>
#include <linux/slab.h>
#include <asm/arch/system.h>
#include <asm/irq.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/fs.h>

#include <asm/arch/fiona.h>
#include <linux/ioc.h>
#include "ioc_hal.h"
#include "fiona_ioc_hal.h"
#include "ioc_protocol.h"

#include <asm/arch/fpow.h>

//
// External Globals
//
extern int g_do_core_debug;

#define cdebug(x...)    if (g_do_core_debug) printk(x)

//////////////////////////////

#define memclr(a,l) memset(a,0,l)
#define KEYBOARD_BUF_SIZE 256
#define MIN(x,y) ((x<y)?x:y)


//////////////////////////////
// External Prototypes
//////////////////////////////
extern int report_fw_engine_version(void);
extern int fasync_ioc(int fd, struct file *filp, int on);
extern void fiona_receive(struct fiona_struct *k);
extern int IOC_Reset_Last_Wake_Condition(void);


//////////////////////////////
// Local Prototypes
//////////////////////////////

// HAL Primitives
//
static int FionaIOC_Hal_Send(IOC_HAL_Interface_T *hal, char *buf, int size);
static int FionaIOC_Hal_DataReady(IOC_HAL_Interface_T *hal);
static int FionaIOC_Hal_Receive(IOC_HAL_Interface_T *hal, char *buf, int size);
static int FionaIOC_Hal_SendRcv(IOC_HAL_Interface_T *hal, char *sndBuf, int sendSize, char *rcvBuf, int rcvSize);
static int FionaIOC_Hal_ResetChip(IOC_HAL_Interface_T *hal);
static int FionaIOC_Hal_FlushRxQueue(IOC_HAL_Interface_T *hal);
static int FionaIOC_Hal_FlushTxQueue(IOC_HAL_Interface_T *hal);
static struct packet_element *FionaIOC_Hal_GetRxPacket(IOC_HAL_Interface_T *hal);
static struct packet_element *FionaIOC_Hal_GetWaitRxPacket(IOC_HAL_Interface_T *hal);



// HW Interface Prototypes
//
static int FionaIOC_Hal_InitSerialPort(void);
static int FionaIOC_Hal_DestroySerialPort(void);


// Local Prototypes
//
static int __FionaIOC_Hal_FlushRxQueue(struct rx_queue *rx_q);
int ioc_put_packet_irq(fiona_packet_t *k);
static int ioc_get_packet_irq(packet_element_t *p);
static int ioc_xfer_packets_to_rx_q(void);

static void ioc_bh(unsigned long dummy);

typedef struct Fiona_Board {
    int ioc_reset_gpio;
} Fiona_Board_T;


//
// Globals & Static variables
//
static fiona_packet_t               g_fiona_rx_packet;
static volatile unsigned char       reply_expected;
static volatile unsigned char       acknowledge;
static volatile unsigned char       resend;
struct Fiona_IOC_HAL_Globals        *g_hal_globs = NULL;
kbd_report_key_proc                 *g_report_key = NULL;
Fiona_Board_T                       board;


// Keep track of last command issued for debugging purposes
#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
packet_element_t    g_last_command;
#endif


DECLARE_TASKLET_DISABLED(ioc_hal_tasklet, ioc_bh, (unsigned long) &g_fiona_rx_packet);


/***************************************************************
 * Interrupt Handler
 ***************************************************************/
static irqreturn_t fiona_ioc_interrupt(int irq, void *data, struct pt_regs *regs)
{
	struct fiona_struct *k = &g_fiona_rx_packet;
	u8 msr, lsr;

#ifdef DEBUG_RX_BUG
    cdebug("< IOC RX INTERRUPT >\n");
#endif
	msr = BTMSR;
	
	if ((msr & MSR_CTS) && !(k->last_msr & MSR_CTS)) {
		if (k->cached_mcr & MCR_RTS)
			printk("confused: RTS already high\n");
		/* CTS went high.  Send RTS. */
		k->cached_mcr |= MCR_RTS;
		BTMCR = k->cached_mcr;
	} else if ((!(msr & MSR_CTS)) && (k->last_msr & MSR_CTS)) {
		/* CTS went low. */
		if (!(k->cached_mcr & MCR_RTS))
			printk("confused: RTS already low\n");
		k->cached_mcr &= ~MCR_RTS;
		BTMCR = k->cached_mcr;
	}

	k->last_msr = msr;

	lsr = BTLSR;
	if (lsr & LSR_DR)
		fiona_receive(k);

    // This routine handles receive ONLY
    
    return IRQ_HANDLED;
}


/**************************************************************************
 * Event Queue
 **************************************************************************/
static void ioc_clear_init_rx_queue(struct rx_queue *q)
{
#ifdef DEBUG_RX_BUG
    printk("Calling __FionaIOC_Hal_FlushRxQueue(%08x)\n",(int)q);
#endif
    // Flush packet_list
    __FionaIOC_Hal_FlushRxQueue(q);

#ifdef DEBUG_RX_BUG
    printk("Initializing client_list (%08x)\n",(int)&q->client_list);
#endif

    // Reinitialize the client_list
    init_waitqueue_head(&q->client_list);

#ifdef DEBUG_RX_BUG
    printk("Done initializing client_list\n");
#endif
}


/**************************************************************************
 * Keyboard Handlers
 **************************************************************************/

int register_ioc_keyboard_handler(kbd_report_key_proc *handler)
{
    g_report_key = handler;
    return 0;
}

void unregister_ioc_keyboard_handler(kbd_report_key_proc *handler)
{
    // if it's the right handler, remove it
    if (g_report_key == handler)
        g_report_key = NULL;
}



void dprint_packet(char *k, int length, char *debug_str)
{
    int i;
    if (g_do_core_debug)  {
        printk("%s [ ",debug_str);
        for(i=0; i<length; i++)
            printk("%02x ", k[i]);
        printk("]\n");
    }
}

void dprint_fiona_packet(struct fiona_struct *k, char *debug_str)
{
    dprint_packet(k->packet,k->length,debug_str);
}


static void dprint_packet_element(struct packet_element *k, char *debug_str)
{
    dprint_packet(k->packet,k->length,debug_str);
}



static void
ioc_bh(unsigned long dummy)
{
    unsigned long flags = 0;

#ifdef DEBUG_FIFOS
    printk("ioc_bh(): Getting lock...");
#endif

    // If we have rx packets ready to deliver, wake up any clients
    spin_lock_irqsave(&g_hal_globs->rx_q.lock,flags);
#ifdef DEBUG_FIFOS
    printk("GOT IT\n");
#endif

    ioc_xfer_packets_to_rx_q();

    // Wake up clients if we have packet(s) available
    if (!list_empty(&g_hal_globs->rx_q.packet_list))  {
        kill_fasync(&g_hal_globs->rx_q.fasync, SIGIO, POLL_IN);
        wake_up_interruptible(&g_hal_globs->rx_q.client_list);
    }
    spin_unlock_irqrestore(&g_hal_globs->rx_q.lock,flags);

#ifdef DEBUG_FIFOS
    printk("ioc_bh exiting\n");
#endif
}


void __do_init_uart_hw(void)
{
    char    burn = 0;

    // Setup the BTUART to do 9600, 8N1
    CKEN = ((CKEN & CKEN_RESPECT) | (CKEN7_BTUART));

    BTIER = 0;
    BTFCR = 0x0;                    // Disable FIFO's first....
    BTFCR = 0x07;                   // FIFO mode, 1 byte interrupt, tx empty int, clear fifos
    BTLCR = LCR_DLAB;               // Prepare to access DLL
    BTDLL = 0x60;                   // Set port speed to 9600 Baud
    BTDLH = 0x0;                    // Set port speed to 9600 Baud
    BTLCR = (LCR_WLS0 | LCR_WLS1);  // 8-bit chars, No parity, and 1 stop bit
    BTIER = IER_UUE;                // Enable this UART
	g_fiona_rx_packet.last_msr = BTMSR;
	g_fiona_rx_packet.cached_mcr = (BTMCR & MCR_RESPECT);

    // Suck off any stale rx data...
    wakelog("__do_init_uart_hw() checking for stale data...\n");
    while (BTLSR & LSR_DR)  {
        burn = BTRBR;
        wakelog("__do_init_uart_hw() burned 0x%x\n",burn);
    }
}

void __do_init_uart_interrupts(void)
{
    /* Enable interrupts at the UART */
    BTIER = (IER_UUE | IER_RTIOE | IER_RLSE | IER_RAVIE);
    BTMCR = (g_fiona_rx_packet.cached_mcr | MCR_OUT2);
}

void __do_full_init_uart(void)
{
    __do_init_uart_hw();
    __do_init_uart_interrupts();
}


static int
FionaIOC_Hal_InitSerialPort(void)
{
    int rc = 0;

    // Init UART hardware
    __do_init_uart_hw();

	/* Ok, finally allocate the IRQ, and off we go.. */
	rc = request_irq(IRQ_BTUART, fiona_ioc_interrupt, 0, "IOC", NULL);
    if (rc) {
        printk("ERROR !!  request_irq returned %d\n",rc);
        free_irq(IRQ_BTUART, NULL);
        return(rc);
    }

    /* Enable our tasklet */
    tasklet_enable(&ioc_hal_tasklet);

    /* Enable interrupts at the UART */
    __do_init_uart_interrupts();

    return(rc);
}

//////////////////////////////
// Local Routines
//////////////////////////////

int
OSCR_timeout(unsigned int start, unsigned int end)
{
    unsigned int oscr = OSCR;

    // If wrap case, use different logic to determine timeout
    if (start > end)
        return((oscr > end) && (oscr < start)); // Wrap case
    else
        return((oscr > end) || (oscr < start)); // Normal case
}


static int 
btuart_send(char *buf, int size)
{
    int bytes_sent = 0;
    char *cur_byte_ptr = buf;
    unsigned int start;
    unsigned int end;

    // Loop through, writing out data
    while (bytes_sent != size)  {
        // Wait until we can send another byte and UART is ready for it
        start = OSCR;
        end = start + IOC_SEND_TIMEOUT;
        while ((BTLSR & (LSR_TEMT | LSR_TDRQ)) != (LSR_TEMT | LSR_TDRQ)) {
            if (OSCR_timeout(start,end))  {
                printk("IOC_Hal_Send PID %x timed out waiting for TX data to be sent (lsr %x, fcr %x).\n",(int) sys_getpid(),BTLSR,BTFCR);
                cdebug(" Time constants: s:%x, e:%x, c:%x\n",start,end,OSCR);
#if 1
                {
                    int i;
                    int lsr = 0;
                    int iir = 0;
                    int ier = BTIER;

                    // Check to see that rx interrupts still enabled...
                    if ((ier & IER_RAVIE) != IER_RAVIE)
                        printk("\n\n!! UART Error: RX ints disabled (ier 0x%x\n",ier);

                    // Check to see if we can identify a problem
                    lsr = (BTLSR & LSR_RESPECT);
                    if (lsr & (LSR_FIFOE | LSR_BI | LSR_FE | LSR_PE | LSR_OE))
                        printk("\n\n!! UART Error: BTLSR = 0x%08x\n",lsr);
                    iir = (BTIIR & IIR_RESPECT);
                    if ((iir & (IIR_FIFOES1 | IIR_FIFOES0)) != (IIR_FIFOES1 | IIR_FIFOES0))
                        printk("\n\n!! UART Error: FIFO not enabled (iir 0x%x, ier 0x%x).\n",iir,ier);
                    if (iir & IIR_TOD)
                        printk("\n\n!! UART Error: IIR TOD timeout detected (iir 0x%x, ier 0x%x).\n",iir,ier);

                    printk("Sent bytes [ ");
                    for (i=0;i<bytes_sent;i++)
                        printk("%02x ",buf[i]);
                    printk("] :: Bytes left to send [ ");
                    for (i=bytes_sent;i<size;i++)
                        printk("%02x ",buf[i]);
                    printk("]\n");
                }
#endif 
                return(0);
            }
        }

        // NOTE: We're ignoring flow control here !!

        // Write the character
        BTTHR = (int)*cur_byte_ptr++;
        bytes_sent++;
    }

    dprint_packet(buf, bytes_sent, "IOC TX packet");
        
    if (bytes_sent != size) {
        cdebug("btuart_send returning %d bytes sent (size = %d)\n",bytes_sent,size);
    }
    return(bytes_sent);
}



static int
FionaIOC_Hal_DestroySerialPort(void)
{
    // Disable the serial port
    BTIER = 0;
    BTMCR = g_fiona_rx_packet.cached_mcr;
    CKEN = ((CKEN & CKEN_RESPECT) & (~CKEN7_BTUART));
    free_irq(IRQ_BTUART, NULL);
    g_hal_globs->serial_inited = 0;
    return 0;
}


static int
FionaIOC_Hal_Send(IOC_HAL_Interface_T *hal, char *buf, int size)
{
    int rc = 0;
    int reset_once = 1;     // Allow 1 reset

    spin_lock(&g_hal_globs->ioc_lock);

    if (g_hal_globs->serial_inited != 1)  {
        printk("Serial port not initialized in FionaIOC_Hal_Send()\n");
        return(0);
    }

    // If no data to write, we're done
    if (size == 0) 
        goto exit;

#ifdef DEBUG_IOC_COMMAND_SEND
    {
        char    logStr[256];
        int x;
        sprintf(logStr,"Sending %d byte command to IOC(",size);
        for (x=0;x<size;x++) {
            sprintf(&logStr[strlen(logStr)], " %02x",buf[x]);
        }
        strcat(logStr, " )\n");
        printk(logStr);
    }
#endif // DEBUG_IOC_COMMAND_SEND

#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
    if (size > MAX_COMMAND_BYTES)  {
        printk("COMMAND size of %d is too large a command size (must be <= %d)\n",size,MAX_COMMAND_BYTES);
        goto exit;
    }
    strncpy(g_last_command.packet,buf,size);
    g_last_command.length = size;
#endif

send_data:
    rc = btuart_send(buf,size);
    if ((rc == -1) && (reset_once))  {
        // Something happened, UART is not setup properly
        cdebug("Hal_Send(): Error -1 from btuart_send(), resetting IOC\n");
        spin_unlock(&g_hal_globs->ioc_lock);
        FionaIOC_Hal_ResetChip(hal);
        // IOC forgot wake condition across reset
        if (IOC_Reset_Last_Wake_Condition()) {
            printk("Hal_Send(): Call to IOC_Reset_Last_Wake_Condition() failed !!\n"); 
        }

        spin_lock(&g_hal_globs->ioc_lock);
        reset_once = 0; // Don't allow looping on failures...
        goto send_data;
    }

#ifdef DEBUG_IOC_COMMAND_SEND
    if (rc != 0)  {
        PDEBUG("Back from writing to IOC, wrote %d bytes\n",rc);
    }
    else  {
        PDEBUG("Back from calling write, failed to write...\n");
    }
#endif // DEBUG_IOC_COMMAND_SEND

exit:
    spin_unlock(&g_hal_globs->ioc_lock);
if (rc != size) printk("Hal_Send() returning %d bytes sent (size = %d)\n",rc,size);
    return(rc);
}


static int
FionaIOC_Hal_Receive(IOC_HAL_Interface_T *hal, char *response, int size)
{
    int rc = 0;                     // Default as error condition
    struct packet_element *p = NULL;
    unsigned int start;
    unsigned int end;
    int copySize = 0;

    // If there's no data to receive, we're done
    if (size == 0) {
        rc = size;
        goto exit;
    }
    
    // Go and get keyboard input from serial port
    if (g_hal_globs->serial_inited != 1)  {
        printk("Serial port not initialized in FionaIOC_Hal_Receive()\n");
        goto exit;
    }

    spin_lock(&g_hal_globs->ioc_lock);

    start = OSCR;
    end = start + IOC_RECEIVE_TIMEOUT;
    while (p == NULL)  {
        p = FionaIOC_Hal_GetRxPacket(hal);
        if (p != NULL)  {
#ifdef DEBUG_IOC_COMMAND_RECEIVE
            {
                int x;
                printk("FionaIOC_Hal_Receive() retrieved packet [");
                for (x=0; x<p->length; x++)
                    printk(" %02x",p->packet[x]);
                printk(" ];\n");
            }
#endif
            copySize = MIN(size,p->length);

            // If the receive buffer is not large enough, let the user know of the error !!
            if (copySize != p->length) {
                int orig_core_debug = g_do_core_debug;
                g_do_core_debug = 1;
                printk("WARNING !!  Packet size of %d was larger than rx buffer size of %d.\n",p->length,size);
                dprint_packet_element(p,"Packet received was");
#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
                dprint_packet_element(&g_last_command,"Last command packet sent:");
#endif
                g_do_core_debug = orig_core_debug;
            }
    
            memcpy(response,p->packet,copySize);
            rc = copySize;
            kfree(p);   // Free up the packet, we're done with it
            break;
        }
        else if (OSCR_timeout(start,end)) {
            printk("IOC_Hal_Receive timed out waiting for a response (%x,%x).\n",start,end);
            goto exit;
        }
    }

#ifdef DEBUG_IOC_COMMAND_RECEIVE
    if (rc != 0)  {
        char    logStr[256];
        int x;
        sprintf(logStr,"Back from read , read %d bytes (",rc);
        for (x=0;x<rc;x++) {
            sprintf(&logStr[strlen(logStr)], " %02x",response[x]);
        }
        strcat(logStr, " )\n");
        printk(logStr);
    }
#endif // DEBUG_IOC_COMMAND_RECEIVE

exit:
    //spin_unlock_irqrestore(&g_hal_globs->ioc_lock,flags);
    spin_unlock(&g_hal_globs->ioc_lock);
    return(rc);
}


#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
// Returns number of bytes read in response, or -1 on failure
extern int g_test_toggle;
#endif

static int
FionaIOC_Hal_SendRcv(IOC_HAL_Interface_T *hal, char *cmd, int cmdSize, char *response, int responseSize)
{
    int rc = 0;

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    // Toggle gpio for scope trigger if needed
    if (g_test_toggle)  {
        GPSR(22) = GPIO_bit(22);
    }
#endif

    // Send the command
    rc = (hal->IOC_HAL_Send(hal, cmd, cmdSize));
    if (rc != cmdSize)  {
        // we weren't able to send all data, try and
        // correct for it now by resetting chip ??
        cdebug("ERROR:  FionaIOC_Hal_SendRcv() failed to send all %d bytes of command.\n",cmdSize);
#ifdef DEBUG_VERBOSE
#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
        dprint_packet_element(&g_last_command,"Last command sent was:");
#endif
#endif
        rc = -1;
        goto enable_exit;
    }


    // To make sure we don't get several back-to-back no-response calls that could
    // overflow a uart fifo, we delay here in the case of commands where the
    // IOC doesn't return a response.
    // We delay ~5mSec for each byte (9600 baud connection to IOC)
    //
    if (responseSize == 0)  {
        mdelay((cmdSize*IOC_NO_RESPONSE_CMDS_MDELAY));
        rc = 0;
        goto enable_exit;
    }


    // Read back the response
    rc = hal->IOC_HAL_Receive(hal,response,responseSize);

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    if (g_test_toggle)  {
        GPCR(22) = GPIO_bit(22);
        g_test_toggle = 0;
    }
#endif

    if ((rc != responseSize) && !((rc == 1) && (response[0] == IOC_NO_BATTERY_RESPONSE))) {
        cdebug("ERROR:  FionaIOC_Hal_SendRcv() failed to receive all %d bytes of command response.\n",responseSize);
#ifdef DEBUG_VERBOSE
#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
        dprint_packet_element(&g_last_command,"Last command sent was:");
#endif
#endif
        rc = -1;
        goto enable_exit;
    }

enable_exit:
    return(rc);
}


static int
FionaIOC_Hal_DataReady(IOC_HAL_Interface_T *hal)
{
    // For now, stub and return no data ready
    printk("@@@@ FionaIOC_Hal_DataReady Stubbed @@@@\n");
    return(0);
}


int g_ioc_in_reset = 0;

static int 
FionaIOC_Hal_ResetChip(IOC_HAL_Interface_T *hal)
{
    unsigned int start;
    unsigned int end;

#ifdef DEBUG_RX_BUG
    int uartErr = 0;
#endif
    unsigned long flags = 0;
    //int orig_core_debug = g_do_core_debug;

#ifdef DEBUG_RESET_BUG
    g_do_core_debug = 1;
#endif

    spin_lock_irqsave(&g_hal_globs->ioc_lock,flags);

    printk("@@@@ FionaIOC_Hal_ResetChip @@@@\n");

    // Disable the UART interrupts
    // FIXME - check to assure UART was originally enabled before re-enabling
    disable_irq(IRQ_BTUART);

#ifdef DEBUG_RX_BUG
    cdebug("ResetChip(): FOUND IIR = 0x%x, IER = 0x%x on entry.\n",BTIIR, BTIER);
#endif

    /******************************************************************
     * EVT2 and older boards have a bug where reseting IOC will cause
     * GPIO1 to transition.
     ******************************************************************/
    g_ioc_in_reset = 1;

    // Drive IOC into reset
    if (board.ioc_reset_gpio) {
        GPCR(board.ioc_reset_gpio) = GPIO_bit(board.ioc_reset_gpio);
    }

    // Remember the time we asserted reset
    start = OSCR;

    // Make sure we've held reset long enough before deasserting it
    end = start + IOC_HW_RESET_ASSERT_TIME;

    // Wait for any data crossing from IOC to PXA to clear UART...
    mdelay(20);

    // Fully reset the UART
    __do_full_init_uart();

    // Clear out RX and TX Buffered Data
    FionaIOC_Hal_FlushRxQueue(hal);
    FionaIOC_Hal_FlushTxQueue(hal);

    while (!OSCR_timeout(start,end))  {
#ifdef DEBUG_RX_BUG
        printk("Delaying to respect IOC reset assert time\n");
#endif
        mdelay(1);
    }

    // Bring IOC out of reset
    if (board.ioc_reset_gpio) {
        GPSR(board.ioc_reset_gpio) = GPIO_bit(board.ioc_reset_gpio);
    }

    /******************************************************************
     * EVT2 and older boards have a bug where reseting IOC will cause
     * GPIO1 to transition.
     ******************************************************************/
    g_ioc_in_reset = 0;

    // Remember when we brought IOC out of reset
    start = OSCR;
    end = start + IOC_HW_POST_RESET_TIME;

    //Make sure we give the IOC enough time to do internel checksum
    while (!OSCR_timeout(start,end))  {
#ifdef DEBUG_RX_BUG
        printk("Delaying to respect IOC post-reset startup time\n");
#endif
        mdelay(10);
    }

#ifdef DEBUG_RX_BUG
    cdebug("ResetChip(): FOUND IIR = 0x%x, IER = 0x%x on exit.\n\n",BTIIR, BTIER);
#endif

    // Enable the UART interrupts
    enable_irq(IRQ_BTUART);

    spin_unlock_irqrestore(&g_hal_globs->ioc_lock,flags);
    return(0);
}


/*********************************************************************************
 * Routine:     __FionaIOC_Hal_FlushRxQueue
 * Purpose:     Clear the Receive Queue Packet List
 * Paramters:   rx_q - pointer to rx_queue to be cleared
 * Returns:     0 - SUCCESS
 *              non-zero - ERROR, Queue not flushed
 * ASSUMPTIONS: spinlock rx_q.lock is held by caller
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************************************/
static int
__FionaIOC_Hal_FlushRxQueue(struct rx_queue *rx_q)
{
    packet_element_t    *p = NULL;
    packet_element_t    *q = NULL;  // Temp storage for list_for_each_entry_safe

    // Keep deleting items in list until list is empty
#if ((defined(DEBUG_FIFOS) || defined(DEBUG_RX_BUG)) && defined(DEBUG_VERBOSE))
    printk("__FionaIOC_Hal_FlushRxQueue() iterating through packet list.\n");
    printk(" packet_list 0x%08x, next 0x%08x, prev 0x%08x\n",(int)&rx_q->packet_list, (int) rx_q->packet_list.next, (int)rx_q->packet_list.prev);
#endif

    // For each packet in list, remove and delete it
    list_for_each_entry_safe(p,q,&rx_q->packet_list, list)  {

#if defined(DEBUG_FIFOS) || defined(DEBUG_RX_BUG)
        dprint_packet_element(p, "Flushing and deleting RX packet");
#endif

        list_del(&p->list);
        kfree(p);

#if defined(DEBUG_FIFOS) || defined(DEBUG_RX_BUG)
        printk("Removed another packet from the packet_list queue.\n");
#endif
    }

#if (defined(DEBUG_FIFOS) || defined(DEBUG_RX_BUG)) && defined(DEBUG_VERBOSE)
    printk("Nil'ing out rx_holding_bin\n");
#endif

    // Remove any packets in holding bin
    g_hal_globs->rx_q.rx_holding_bin.head = g_hal_globs->rx_q.rx_holding_bin.tail = 0;

    return(0);
}

/*********************************************************************************
 * Routine:     FionaIOC_Hal_FlushRxQueue
 * Purpose:     Clear the Receive Queue
 * Paramters:   hal - pointer to HAL's interface structure
 * Returns:     0 - SUCCESS
 *              non-zero - ERROR, Queue not flushed
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************************************/
static int
FionaIOC_Hal_FlushRxQueue(IOC_HAL_Interface_T *hal)
{
    unsigned long       flags;

    // FIXME - TODO flush rx FIFO at uart
    spin_lock_irqsave(&g_hal_globs->rx_q.lock,flags);
    __FionaIOC_Hal_FlushRxQueue(&g_hal_globs->rx_q);
    spin_unlock_irqrestore(&g_hal_globs->rx_q.lock,flags);

    return(0);
}


/*********************************************************************************
 * Routine:     FionaIOC_Hal_FlushTxQueue
 * Purpose:     Clear the Transmit Queue
 * Paramters:   hal - pointer to HAL's interface structure
 * Returns:     0 - SUCCESS
 *              non-zero - ERROR, Queue not flushed
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************************************/
static int
FionaIOC_Hal_FlushTxQueue(IOC_HAL_Interface_T *hal)
{
#if 0
    struct list_head *entry = NULL;
    packet_element *p = NULL;
    unsigned long flags;

    // FIXME - TODO flush tx FIFO at uart
    spin_lock_irqsave(&g_hal_globs->tx_q.lock,flags);

    // Keep deleting items in list until list is empty
    entry = g_hal_globs->tx_q.head.next;
    list_for_each_entry(entry,&g_hal_globs->tx_q.head, head)  {
//    while (entry != &g_hal_globs->tx_q.head)  {
        struct packet_element *p = list_entry(entry,struct packet_element, entry);
        list_del(&p->head);

#ifdef DEBUG_FIFOS
        {
            int x;
            printk("Flushing TX packet [");
            for (x=0; x<p->length; x++)
                printk(" %02x",p->packet[x]);
            printk("]\n");
        }
#endif

        kfree(p);
    }

    spin_unlock_irqrestore(&g_hal_globs->tx_q.lock,flags);
#endif
    return(0);
}

static int
ioc_rx_holding_bin_empty(void)
{
    return(g_hal_globs->rx_q.rx_holding_bin.head == g_hal_globs->rx_q.rx_holding_bin.tail);
}


int ioc_put_packet_irq(fiona_packet_t *k)
{
    packet_element_t    *p = NULL;
    rx_holding_bin_t    *bin = &g_hal_globs->rx_q.rx_holding_bin;
    int size = 0;
    int err = 0;

    // We don't grab a lock as anyone accessing temp bin needs to spin_lock_irqsave()
    p = &bin->packet[bin->head++];
    size = MIN(sizeof(p->packet),k->length);
    memcpy(p->packet,k->packet,size);
    p->length = size;

    // Wrap our pointer if needed
    if (bin->head >= MAX_RX_PACKETS_TO_BUFFER) {
        bin->head = 0;
    }

    // If head = tail, we've overflowed the buffer, so remove the last item in circular queue
    if (bin->head == bin->tail)  {
        printk("DROPPED RX Packet on floor - rx_holding_bin buffer full !!\n");
        dprint_packet_element(&bin->packet[bin->tail],"Dropped RX packet");
        bin->tail = ((bin->tail+1) >= MAX_RX_PACKETS_TO_BUFFER) ? 0 : bin->tail+1;
    }

#ifdef DEBUG_IOC_COMMAND_RECEIVE
    {
        int x;
        printk("** DATA QUEUED: ioc_put_packet_irq queueing [");
        for (x=0;x<size;x++)
            printk(" %02x",p->packet[x]);
        printk(" ]\n");
    }
#endif

    return(err);
}

static int
ioc_get_packet_irq(packet_element_t *k)
{
    packet_element_t    *p = NULL;
    rx_holding_bin_t    *bin = &g_hal_globs->rx_q.rx_holding_bin;
    int                 size = 0;
    int                 err = 0;
    
    if (bin->tail == bin->head) {
        err = 1;    // Indicate empty queue
        goto exit;
    }

    p = &bin->packet[bin->tail++];
    size = MIN(sizeof(k->packet),p->length);
    memcpy(k->packet,p->packet,size);
    k->length = size;

    // Wrap our pointer if needed
    if (bin->tail >= MAX_RX_PACKETS_TO_BUFFER) {
        bin->tail = 0;
    }

#ifdef DEBUG_IOC_COMMAND_RECEIVE
    {
        int x;
        printk("** DATA DEQUEUED: ioc_get_packet_irq dequeueing [");
        for (x=0;x<size;x++)
            printk(" %02x",p->packet[x]);
        printk(" ]\n");
    }
#endif
exit:
    return(err);
}



static int
ioc_xfer_packets_to_rx_q(void)
{
    int                 err = 0;
    packet_element_t    *p = NULL;

    while (!ioc_rx_holding_bin_empty()) {
        p = kmalloc(sizeof(struct packet_element), GFP_ATOMIC); 
        if (p == NULL)  {
            err = -ENOMEM;
            goto exit;
        }
                
        // Get the next packet
        err = ioc_get_packet_irq(p);

        if (err)
            goto exit;

        // Queue up that packet in the client rx packet queue
        list_add(&p->list, &g_hal_globs->rx_q.packet_list);

#ifdef DEBUG_IOC_COMMAND_RECEIVE
        {
            int x;
            printk("** DATA TRANSFERED: ioc_xfer_packet_to_rx_q xfered [");
            for (x=0;x<p->length;x++)
                printk(" %02x",p->packet[x]);
            printk(" ]\n");
        }
#endif
    }

exit:
    return(err);
}


/*********************************************************************************
 * Routine:     FionaIOC_Hal_GetRxPacket
 * Purpose:     Get the next fully received packet if there's one ready.
 *              This routine removes the packet from the receive queue.
 * Paramters:   hal - pointer to HAL's interface structure
 * Returns:     NULL is returned if no complete packets are ready
 *              Otherwise, a pointer to the received packet is returned.
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************************************/
static struct packet_element *
FionaIOC_Hal_GetRxPacket(IOC_HAL_Interface_T *hal)
{
    struct list_head *entry = g_hal_globs->rx_q.packet_list.next;
    packet_element_t *p = NULL;
    unsigned long flags;

    spin_lock_irqsave(&g_hal_globs->rx_q.lock,flags);

    // If we have a packet ready, remove it from the queue and return it...
    if (entry != &g_hal_globs->rx_q.packet_list)  {
        p = list_entry(entry,struct packet_element, list);
        list_del(&p->list);
    }

    spin_unlock_irqrestore(&g_hal_globs->rx_q.lock,flags);

#ifdef DEBUG_FIFOS
    {
        int x;

        if (p == NULL)
            printk("Fiona_IOC_Hal_GetRxPacket() found NO packets waiting.\n");
        else  {
            printk("Fiona_IOC_Hal_GetRxPacket() returning [");
            for (x=0; x<p->length; x++)
                printk(" %02x",p->packet[x]);
            printk("]\n");
        }
    }
#endif

    return(p);
}


/*********************************************************************************
 * Routine:     FionaIOC_Hal_GetWaitRxPacket
 * Purpose:     Get the next fully received packet if there's one ready.
 *              If a packet is not ready, wait for it.
 *              This routine removes the packet from the receive queue.
 * Paramters:   hal - pointer to HAL's interface structure
 * Returns:     packet_element* - pointer to the received packet
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************************************/
static struct packet_element *
FionaIOC_Hal_GetWaitRxPacket(IOC_HAL_Interface_T *hal)
{
    struct list_head *entry = NULL;
    packet_element_t *p = NULL;
    unsigned long   flags;

    spin_lock_irqsave(&g_hal_globs->rx_q.lock,flags);

    // If we have a packet ready, remove it from the queue and return it...
    entry = g_hal_globs->rx_q.packet_list.next;
    if (entry == &g_hal_globs->rx_q.packet_list)  {
        // TODO - put us on a wait queue until a packet arrives
        printk("Fiona_IOC_Hal_GetWaitRxPacket() waiting on data....\n");
    }

    // If we have a packet ready, remove it from the queue and return it...
    entry = g_hal_globs->rx_q.packet_list.next;
    if (entry != &g_hal_globs->rx_q.packet_list)  {
        struct packet_element *p = list_entry(entry,struct packet_element, list);
        list_del(&p->list);
    }

    spin_unlock_irqrestore(&g_hal_globs->rx_q.lock,flags);

#ifdef DEBUG_FIFOS
    {
        int x;
        printk("Fiona_IOC_Hal_GetWaitRxPacket() returning [");
        for (x=0; x<p->length; x++)
            printk(" %02x",p->packet[x]);
        printk("]\n");
    }
#endif
    return(p);
}


////////////////////////////////////////////////////////
//  Start of IOC Hardware Interface Code.
//  On Fiona, the IOC is connected to the BTUART
//  of the Cotulla.
////////////////////////////////////////////////////////

//
// These two routines are called from the base IOC driver 
// to allow us to initialize HAL and pass back our HAL table
//
//
extern struct fiona_hw g_fiona_hw;

IOC_HAL_Interface_T *
IOC_HAL_Init(void)
{
    int rc = 0;

    // Let user know what engine is installed
    report_fw_engine_version();

	g_hal_globs = kmalloc(sizeof(Fiona_IOC_HAL_Globals_T), GFP_KERNEL);
	if (g_hal_globs == NULL)  {
		printk("Could not kmalloc space for IOC HAL table !\n");
		return(NULL);
	}

    // Initialize the HAL world
    memclr(g_hal_globs,sizeof(Fiona_IOC_HAL_Globals_T));
    g_hal_globs->signature = (unsigned int) IOC_GLOBALS_SIGNATURE;

    // Save off where the IOC reset GPIO is
    board.ioc_reset_gpio = g_fiona_hw.ioc_reset_gpio;

    // Setup our primitives
    g_hal_globs->hal.IOC_HAL_Send = FionaIOC_Hal_Send;
    g_hal_globs->hal.IOC_HAL_DataReady = FionaIOC_Hal_DataReady;
    g_hal_globs->hal.IOC_HAL_Receive = FionaIOC_Hal_Receive;
    g_hal_globs->hal.IOC_HAL_SendRcv = FionaIOC_Hal_SendRcv;
    g_hal_globs->hal.IOC_HAL_ResetChip = FionaIOC_Hal_ResetChip;
    g_hal_globs->hal.private = (void*)g_hal_globs;
    g_hal_globs->hal.IOC_HAL_GetRxPacket = FionaIOC_Hal_GetRxPacket;
    g_hal_globs->hal.IOC_HAL_GetWaitRxPacket = FionaIOC_Hal_GetWaitRxPacket;
    g_hal_globs->hal.IOC_HAL_FlushRxQueue = FionaIOC_Hal_FlushRxQueue;
    g_hal_globs->hal.IOC_HAL_FlushTxQueue = FionaIOC_Hal_FlushTxQueue;

    // Setup our transmit and receive FIFO's
#ifdef DEBUG_FIFOS
    printk( "Intializing g_hal_globs->rx_q.packet_list and rx_q.lock\n");
#endif
    g_hal_globs->rx_q.packet_list.next = &g_hal_globs->rx_q.packet_list;
    g_hal_globs->rx_q.packet_list.prev = &g_hal_globs->rx_q.packet_list;
    spin_lock_init(&g_hal_globs->rx_q.lock);
    spin_lock_init(&g_hal_globs->ioc_lock);
    g_hal_globs->rx_q.client_list.lock = SPIN_LOCK_UNLOCKED;
    g_hal_globs->rx_q.client_list.task_list.next = &g_hal_globs->rx_q.client_list.task_list;
    g_hal_globs->rx_q.client_list.task_list.prev = &g_hal_globs->rx_q.client_list.task_list;

    g_hal_globs->rx_q.rx_holding_bin.head = g_hal_globs->rx_q.rx_holding_bin.tail = 0;

#if defined(DEBUG_IOC_COMMAND_RECEIVE) || defined(DEBUG_RX_BUG)
    g_last_command.list.next = &g_last_command.list;
    g_last_command.list.prev = &g_last_command.list;
    g_last_command.length = 0;
#endif

    // Setup the serial port to talk to the IOC
    rc = FionaIOC_Hal_InitSerialPort();
    if (rc != 0)  {
        PDEBUG("FionaIOC_Hal_InitSerialPort(): Error %d opening serial port %s.\n",rc,BTUART_SERIAL_PORT_NAME);
        kfree(g_hal_globs);
        return(NULL);
    }
    g_hal_globs->serial_inited = 1;
        
    // Pass back our HAL vector table
	return(&g_hal_globs->hal);
}


int 
IOC_HAL_Destroy(IOC_HAL_Interface_T *hal)
{
    // Tear down the serial port connection
    FionaIOC_Hal_DestroySerialPort();

    // Clear our the FIFOs
    ioc_clear_init_rx_queue(&g_hal_globs->rx_q);
    FionaIOC_Hal_FlushTxQueue(hal);

	// Free up any other space allocated by us...
    if (NULL != hal->private)
	    kfree(hal->private);
	return 0;
}

