/******************************************************************
 *
 *  File:   ioc_cmds.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   01/04/06
 *
 * Copyright 2006, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *      IOC Commands for Protocol v7 and later
 *
 ******************************************************************/

#include <linux/autoconf.h>
#ifdef CONFIG_IOC

#include <stdarg.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/errno.h>
#include <asm/string.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/fiona.h>
#include <asm/arch/boot_globals.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/syscalls.h>

#include <linux/ioc.h>
#include <linux/pn_lcd.h>
#include <net/anydata.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/fiona_keymap.h>
#include "fiona_ioc_hal.h"
#include "ioc_hal.h"
#include "ioc_protocol.h"
#include "ioc_cmds.h"

// Set to 1 to print out the voltage 
// averaging tables when we hit a watermark.
#define PRINT_VTABLE_ON_WATERMARK_HIT   1


/*
 * CONFIG_SAFETY_NET
 *
 * If defined, the watermark check mechanism (wcm) will use a safety net.
 * The wcm currently adjusts voltage spikes reported into actual calculated
 * voltage at the battery.  This calculation can be off due to a couple of
 * factors, the biggest one being that the IOC doesn't snapshot voltage and
 * current draws at the exact same time, and since both of these values are
 * used when determining adjusted voltage, it's important that they're from
 * the same sample (and they're not, really).
 *
 * To try and be more robust / account for this fact as best we can in SW,
 * we can add a safety net, which means that if any unadjusted voltage (the
 * actual spike read) hits the net, we'll treat that the same as if the adjusted
 * value hit the watermark.
 *
 * Unfortunately, we can't control the depth of the spike coming, and once we're
 * told about it, it's already too late, so this is not a "catch-all" net.  
 * If the spike is large enough, it will "hit the ground through the net". 
 *
 * It is better than doing nothing, so we're adding a safety net to try and catch
 * a large percentage of "overweight jumpers".
 */

#undef LOG_EVERY_VOLT_READING

#define RESISTANCE_CALCULATION_DEBUG
#ifdef RESISTANCE_CALCULATION_DEBUG
  #define resLog(x...) printk(x)
#else
  #define resLog(x...)  do{}while(0)
#endif


int Do_IOC_IssueCmd(char *cmd, int cmdSize, char *response, int responseSize, int resetChipOnErr);
#define IOC_IssueCmd(w,x,y,z)   Do_IOC_IssueCmd(w,x,y,z,1)

////////////////////////////
// External Globals 
////////////////////////////
extern int              g_ioc_enabled;   // ENABLE or DISABLE
extern Battery_Info_T   g_battery_info;
extern int              g_voltage_dump_debug;

#ifdef LOCAL_CONFIG_PM
extern int g_ioc_wake_state;
#endif

extern struct tasklet_struct ioc_hal_tasklet;

// Since these two keys default to 0x2f and 0x26 inside the IOC, we define
// them here based on rawcode instead of keycode since keycode would change on
// different rev boards, but not inside the IOC !!
//
//static int g_wake_rawkey_1 = 0x2F;      // Home key on DVT1
//static int g_wake_rawkey_2 = 0x26;      // SYM key on DVT1
static int g_last_wake_condition = IOC_CMD_ENABLE_ALL_KEY_WAKE;

//
// Next time to print out voltage level
static unsigned long next_voltage_level_log_time = 0;
unsigned long voltage_level_log_freq = VOLTAGE_LOGGING_FREQUENCY_SECS;

// Values to use for voltage smoothing algorithm
int g_icon_rising_lambda = 2;
int g_icon_falling_lambda = 3;
int g_icon_mvolt_average = 0;

int g_icon_board_resistance = BATTERY_ICON_BOARD_RESISTANCE;
int g_min_board_resistance = MIN_BOARD_RESISTANCE;
int g_max_board_resistance = MAX_BOARD_RESISTANCE;

static unsigned long g_critical_watermark_timeout = INITIAL_JIFFIES;
static unsigned long g_wan_watermark_timeout = INITIAL_JIFFIES;
static unsigned long g_low_watermark_timeout = INITIAL_JIFFIES;

static int sleep_flag = 1; // Force update first time, then everytime after sleep

#ifdef CONFIG_IOC_VOLTAGE_CHECKER
#ifdef CONFIG_SAFETY_NET
static unsigned long g_ioc_host_safety_net = 0;
#endif
static unsigned long g_ioc_host_watermark = 0;
static unsigned long g_ioc_voltage_check_enabled = 0;
static int g_voltage_checker_needs_reset = 0;

static unsigned long    g_voltage_sample_count = 0;
static unsigned long    g_next_temp_sample = 0;
static unsigned long    g_next_voltage_sample = 0;
static unsigned long    g_next_icon_voltage_sample = 0;
static unsigned long    g_next_amperage_sample = 0;

static unsigned long    g_temperature_samples[TEMPERATURE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_amperage_samples[AMPERAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_voltage_samples[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_icon_voltage_samples[VOLTAGE_AVERAGE_SAMPLE_COUNT];

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static unsigned long    g_2_sample_count = 0;
static unsigned long    g_5_sample_count = 0;
static unsigned long    g_8_sample_count = 0;
static unsigned long    g_10_sample_count = 0;
static unsigned long    g_20_sample_count = 0;
static unsigned long    g_next_voltage_sample_2_sec = 0;
static unsigned long    g_next_voltage_sample_5_sec = 0;
static unsigned long    g_next_voltage_sample_8_sec = 0;
static unsigned long    g_next_voltage_sample_10_sec = 0;
static unsigned long    g_next_voltage_sample_20_sec = 0;

static unsigned long    g_voltage_samples_2_sec[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_voltage_samples_5_sec[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_voltage_samples_8_sec[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_voltage_samples_10_sec[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 
static unsigned long    g_voltage_samples_20_sec[VOLTAGE_AVERAGE_SAMPLE_COUNT]; 

#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR

static unsigned long    g_vc_sample_count = VOLTAGE_AVERAGE_SAMPLE_COUNT;

#endif // CONFIG_IOC_VOLTAGE_CHECKER


//////////////////////////////
// External Prototypes
//////////////////////////////
extern int ioc_put_packet_irq(fiona_packet_t *k);
extern signed int ioc_convert_raw_temp_to_degrees_celcius(unsigned int raw);
extern unsigned int ioc_convert_raw_voltage_to_mvolts(unsigned int raw);
extern unsigned int ioc_convert_rve_voltage_to_icon_adj_mvolts(unsigned int rve_mvolts, unsigned int draw);
extern unsigned int ioc_convert_rve_voltage_to_adj_mvolts(unsigned int rve_mvolts, unsigned int draw);
extern unsigned int ioc_convert_raw_voltage_to_icon_mvolts(unsigned int raw);
extern unsigned int ioc_convert_mvolts_to_raw_voltage(unsigned int mvolts);
extern unsigned int ioc_convert_raw_draw_to_mamps(unsigned int raw);

extern int fpowGetWanWatermark(void);
extern int fpowGetLowWatermark(void);
extern int fpowGetCriticalWatermark(void);

//////////////////////////////
// Local Prototypes
//////////////////////////////
// HW Function Enable/Disable Routines
static int  Do_Enable_IOC(unsigned short enable);
static int  IOC_Enable_Keyboard(unsigned short enable);
static int  IOC_Query_Cmd(int query_cmd, char *response, char parm);
static int  VERIFY_BAD_RESPONSE(char cmd, int response_size, char *response);
static int  IOC_Flash_Write(char *data, int length);
static int  IOC_All_Key_Wake_Enable(unsigned short enable);
void        get_battery_info_from_samples(Battery_Info_T *info);
static int  rangeCheckBatteryValues(Battery_Info_T *info);

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static int ga(int table_id);
#endif

// Segment On/Off Functions
#define SEGMENT_ON  1           // Values for "on" parameter
#define SEGMENT_OFF 0
static int  IOC_Sequential_Segment(unsigned short on, char startSeg, char endSeg, int vertical);
static int  IOC_Multi_Segment(unsigned short on, char num_segments, char *segment_array);
static int  IOC_Multi_Segment_Pairs(unsigned short on, char num_pairs, char *paired_segment_array, int vertical);

// Receive Engine
static void ioc_response_seen(char *packet);
static int ioc_process_rx_packet(struct fiona_struct *k);

static inline void handle_keyboard_event(unsigned char scancode)
{
#ifdef CONFIG_IOC_KEYBOARD
    fiona_kbd_report_key((scancode&0x7F),((scancode&0x80)==0));
#endif
}

void report_fw_engine_version()
{
    PDEBUG("**** IOC FW Protocol Engine Version 7\n");
}

void ioc_reset_voltage_averages(void) {
    waterLog("ioc_reset_voltage_averages()");
    g_voltage_checker_needs_reset = 1;

    // Wake up voltage timer now so it can reset timer frequency...
    mod_timer(&vc_timer,jiffies);
}


/*********************************************************
 * Routine:     verify_firmware_protocol_match
 * Purpose:     Verify IOC firmware version matches currently
 *              enabled protocol stack.
 * Paramters:   none
 * Returns:     -EIO - Firmware IS NOT supported by this protocol stack
 *              0 - Firmware IS supported by this protocol stack
 * Side Effects: Sets g_fw_version
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *********************************************************/
int get_ioc_firmware_version(int *version, int force_update)
{
    int retVal = 0;
    char fw_version_str[32];
    int pass = 1;

    if (version == NULL)
        return(-EINVAL);

    if ((g_fw_version != 0) && (!force_update)) {
        *version = g_fw_version;
        return(0);
    }

    *version = 0;

getVersion:
    retVal = IOC_Query_Cmd(IOC_QUERY_FW_VERSION, fw_version_str, 0);
    if (retVal == 0) {

        // If a crush condition occurred, check again
        if ((pass != 2) && (fw_version_str[0] == 0xE3))  {
            // WE got a crush condition from IOC.  Ignore for now and ask for firmware version again.
            pass = 2;
            goto getVersion;
        }

        if ((fw_version_str[0] != 0xE0) ||
            (fw_version_str[1] < MIN_VALID_IOC_VERSION) ||
            (fw_version_str[1] > MAX_VALID_IOC_VERSION))  {
            printk("IOC Firmware Version #%d too %s for this kernel version"
                " (version must be >= %d and <= %d).\n",
                (int)fw_version_str[1],
                (fw_version_str[1] < MIN_VALID_IOC_VERSION)?"old":"new",
                MIN_VALID_IOC_VERSION,MAX_VALID_IOC_VERSION);
             retVal = -EIO;
             goto exit;
        }
        else {
            g_fw_version = fw_version_str[1];   // Already in decimal format, not char
            *version = g_fw_version;
            PDEBUG("IOC: Firmware Version #%d \n",g_fw_version);
        }
    }
exit:
    return(retVal);
}


int verify_firmware_protocol_match(void)
{
    int retVal = 0;
    int version = 0;

    retVal = get_ioc_firmware_version(&version,0);
    if (retVal)  {
        printk("IOC Firmware too old for this kernel version (version must be >= %d).\n",MIN_VALID_IOC_VERSION);
        retVal = -EIO;
        goto exit;
    }
exit:
    return(retVal);
}

/////////////////////////////////////////////
// Battery Information Routine
/////////////////////////////////////////////

/*******************************************************************
 * Routine:     battery_installed
 * Purpose:     Returns whether there's a battery plugged in or not.
 * Paramters:   none
 * Returns:     0 - No battery installed
 *              1 - Battery present
 * Assumptions: IOC will return IOC_NO_BATTERY_RESPONSE if no battery installed
 *              on calls to query capacity, voltage, or temperature
 * Author:      Nick Vaccaro <nvaccaro@lab126.com>
 *******************************************************************/
int battery_installed(void)
{
    int retVal;
    int version = 0;
    int battery_installed = 0;
    char response[MAX_IOC_RX_PACKET_SIZE];

    retVal = get_ioc_firmware_version(&version, 0);
    if (retVal != 0)
        return(0);

    if (version >= 20) {
        retVal = IOC_Query_Cmd(IOC_QUERY_CHARGE_OR_DISCHARGE, response, 0);
        if (retVal == 0) {
            battery_installed = (response[1] == 3) ? 0 : 1;
            BDEBUG("Queried Charge State (battery installed) Status: [ %02x = %s]\n",response[1],battery_installed?"TRUE":"FALSE");
        }
    }
    else  {
        retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_TEMPERATURE, response, 0);
        battery_installed = (response[0] != IOC_NO_BATTERY_RESPONSE);
        BDEBUG("battery_installed returned %s\n",(battery_installed)?"TRUE":"FALSE");
    }
    return (battery_installed);
}

// Version of IOC firmware that implements battery_info command

int get_battery_info(Battery_Info_T *info, int raw)  {
    int raw_volts = 0;
    int raw_amps = 0;
    int raw_temp = 0;
    int charger = 0;
    int version = 0;
    int retVal;
    char response[MAX_IOC_RX_PACKET_SIZE];
    unsigned long adjusted_mvolts = 0;
    Battery_Info_T local_info;

    // Clear struct
    memset(&local_info,0,sizeof(local_info));

    if (info == NULL)
        return(-1);
    retVal = get_ioc_firmware_version(&version, 0);
    if (retVal != 0)
        return(-1);
    if (version >= IOC_QUERY_BATTERY_INFO_CMD_VERSION) {
        retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_INFO, response, 0);
        if (retVal == 0) {
            raw_volts = response[IOC_QUERY_BATTERY_INFO_VOLT_RESPONSE_OFFSET];
            raw_amps = response[IOC_QUERY_BATTERY_INFO_AMP_RESPONSE_OFFSET];
            raw_temp = response[IOC_QUERY_BATTERY_INFO_TEMP_RESPONSE_OFFSET];
            charger = ((response[IOC_QUERY_BATTERY_INFO_CHARGE_RESPONSE_OFFSET]==IOC_QUERY_BATTERY_INFO_CHARGING_RESPONSE)?1:0);
            BDEBUG("Queried Battery Info, vat=0x%x 0x%x 0x%x\n",raw_volts,raw_amps, raw_temp);
        }
        else BDEBUG("ERROR !! Tried Queried Battery Info\n");
    }
    else  {
        // Get it piece-meal, open to flaws due to sampling time
        // Get Voltage Level
        BDEBUG("Querying Battery Level...\n");
        retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_LEVEL, response, 0);
        if (retVal == 0) {
            raw_volts = (int) response[1];
        }
        else
            BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery voltage.\n",retVal);

        // Get Battery Current Draw
        BDEBUG("Querying Battery Current Draw...\n");
        retVal = IOC_Query_Cmd(IOC_QUERY_CURRENT_DRAW, response, 0);
        if (retVal == 0) {
            raw_amps = (int) response[1];

            // If it's a charge response and not a discharge response,
            // then we're actually drawing "negative" current as we're charging.
            if (response[0] == IOC_BATTERY_CHARGE_RESPONSE)
                charger = 1;
            else
                charger = 0;
            BDEBUG("Queried Battery Current Draw: [ %02x %02x ]\n",response[0],response[1]);
        }
        else
            BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery current draw.\n",retVal);

        retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_TEMPERATURE, response, 0);
        if (retVal == 0) {
            raw_temp = response[1];
        }
        else
            BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery temperature.\n",retVal);
        
    }

    if (raw_volts == 0)
        waterLog("WARNING: IOC returned raw_volts = 0\n");

    if (raw_amps == 0)
        waterLog("WARNING: IOC returned raw_amps = 0\n");

    if (charger)
        info->draw = MIN_MAMP_VALUE;
    else
        info->draw = ioc_convert_raw_draw_to_mamps(raw_amps);

    info->raw_mvolts = ioc_convert_raw_voltage_to_mvolts(raw_volts);

    info->icon_mvolts = ioc_convert_rve_voltage_to_icon_adj_mvolts(info->raw_mvolts, info->draw);
    adjusted_mvolts = ioc_convert_rve_voltage_to_adj_mvolts(info->raw_mvolts,info->draw);
    info->temperature = ioc_convert_raw_temp_to_degrees_celcius(raw_temp);
    info->charging = charger;

    // These fields are not really used...
    info->installed = battery_installed();

    powdebug("%d mv = (raw %d + ((%d * %d)/1000)) (s=%d)\n",(int)adjusted_mvolts,(int)info->raw_mvolts,info->draw,board_resistance,(int)g_voltage_sample_count);


#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    get_battery_info_from_samples(&local_info);
    waterLog("VOLT: %d,%d,%d,%d,%d,%d,%d,%d,%d\n",local_info.mvolts,ga(2),ga(5),ga(8),ga(10),ga(20),(int)adjusted_mvolts,(int)info->raw_mvolts,info->draw);
#else
    get_battery_info_from_samples(&local_info);
    waterLog("VOLT: %d,%d,%d,%d\n",local_info.mvolts,(int)adjusted_mvolts,(int)info->raw_mvolts,info->draw);
#endif  // CONFIG_MULTI_RATE_AVERAGING_MONITOR

    // Return adjusted voltage (voltage at battery) instead of raw voltage (measured at board)
    if (raw == 0)
        info->mvolts = adjusted_mvolts;
    else
        info->mvolts = info->raw_mvolts;

    BDEBUG("Queried Battery Info: [ %d.%03d Volts, %d.%03d RawV, %d mAmps, %d Celcius, %sCHARGING, %sINSTALLED\n",info->mvolts/1000,info->mvolts%1000,info->raw_mvolts/1000,info->raw_mvolts%1000,info->draw,info->temperature,(charger?"":"DIS"),(info->installed?"":"NOT "));

    // Return SUCCESS
    return(0);
}

int ac_adapter_installed(void)
{
    int ac = 0;
    int retVal = 0;
    int version;
    char str[MAX_IOC_RX_PACKET_SIZE];

    // Get Battery Charging Status
    BDEBUG("Querying AC Adapter Status...\n");
    retVal = IOC_Query_Cmd(IOC_QUERY_CHARGE_OR_DISCHARGE, str, 0);
    if (retVal == 0) {
        retVal = get_ioc_firmware_version(&version, 0);
        if ((retVal != 0) || (version < 20)) {
            ac = str[0];
            BDEBUG("Queried AC Adapter Status: [ %02x ]\n",str[0]);
        }
        else {
            ac = (str[1] == 0) ? 0 : 1;
            BDEBUG("Queried AC Adapter Status: [ %02x ]\n",str[1]);
        }
    }
    else
        BDEBUG("ac_adapter_installed(): Error %d getting battery charge/discharge status.\n",retVal);

    return(ac);
}


static void update_ac_adapter_status(int on)
{
    if (on)
        globs.flags |= AC_ADAPTER_PRESENT_BIT;
    else
        globs.flags &= ~AC_ADAPTER_PRESENT_BIT;
}


static int ac_not_adjusted_since_sleep(void) {
    return(sleep_flag);
}
static void ac_adjusted(void) {
    sleep_flag = 0;
}
void ioc_set_sleep_flag(void) {
    sleep_flag = 1;
}

extern int g_ioc_sleeping;
int ac_adapter_status(void)
{
    if (ac_not_adjusted_since_sleep() && (!g_ioc_sleeping))  {
        ac_adjusted();  // reset flag
        update_ac_adapter_status(ac_adapter_installed());
    }
    return (globs.flags & AC_ADAPTER_PRESENT_BIT);
}

void IOC_Query_Battery_Info(Battery_Info_T *info)
{
    int retVal = 0;
    int version = 0;
    char str[MAX_BYTES_IOC_RESPONSE];
    int battery_present = 1;            // Assume a battery is present

    // First, check if battery is installed
    if (!battery_installed()) {
        battery_present = 0;
        info->mvolts = info->installed = info->temperature = 0;
        goto no_battery_continue;
    }
    info->installed = 1;

    // Get Battery Temperature
    retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_TEMPERATURE, str, 0);
    if (retVal == 0)
        info->temperature = ioc_convert_raw_temp_to_degrees_celcius((int)str[1]);
    else
        BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery temperature.\n",retVal);


    // Get Voltage Level
    retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_LEVEL, str, 0);
    if (retVal == 0)
        info->mvolts = ioc_convert_raw_voltage_to_mvolts((int)str[1]);
    else
        BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery voltage.\n",retVal);

no_battery_continue:
    // Get Battery Current Draw
    retVal = IOC_Query_Cmd(IOC_QUERY_CURRENT_DRAW, str, 0);
    if (retVal == 0) {
        info->draw = ioc_convert_raw_draw_to_mamps(str[1]);

        // If it's a charge response and not a discharge response,
        // then we're actually drawing "negative" current as we're charging.
        if (str[0] == IOC_BATTERY_CHARGE_RESPONSE)
            info->draw *= -1;
    }
    else
        BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery current draw.\n",retVal);

    // Get Battery Charging Status
    retVal = IOC_Query_Cmd(IOC_QUERY_CHARGE_OR_DISCHARGE, str, 0);
    if (retVal == 0) {
        retVal = get_ioc_firmware_version(&version, 0);
        if ((retVal != 0) || (version < 20))
            info->charging = str[0];
        else
            info->charging = (str[1] == 0) ? 0 : 1;
    }
    else
        BDEBUG("IOC_Query_Battery_Info(): Error %d getting battery charge/discharge status.\n",retVal);

    // Report our battery status if someone's interested
    if (battery_present) {
        BDEBUG("Battery: %d.%03d Volts, %03d Degrees C, %d mA Draw, %sHARGING\n", \
            (info->mvolts/1000),(info->mvolts%1000),info->temperature, info->draw, \
            (info->charging)?"C":"DISC");
    }
    else
        BDEBUG("Battery: NOT INSTALLED, Draw %d, A/C Charger is %s\n",info->draw,(info->charging)?"IN":"OUT");
}

int fiona_get_current_draw(void)
{
    char str[MAX_BYTES_IOC_RESPONSE];
    int draw = 0;
    int retVal = 0;

    retVal = IOC_Query_Cmd(IOC_QUERY_CURRENT_DRAW, str, 0);
    if (retVal == 0)
        draw = ioc_convert_raw_draw_to_mamps(str[1]);
    return(draw);
}

void update_battery_info(struct fiona_struct *k, Battery_Info_T *info)
{
    switch(k->packet[0])  {

    case IOC_BATTERY_LOW_RESPONSE:          // Battery Low Voltage Warning
        info->mvolts = ioc_convert_raw_voltage_to_mvolts((int)k->packet[1]);
        BDEBUG("update_battery_info(): LOW VOLTAGE RESPONSE, voltage = %02x\n",info->mvolts);
        break;
        
    case IOC_BATTERY_CAPACITY_RESPONSE:     // Battery Capacity Data
        BDEBUG("update_battery_info(): CAPACITY RESPONSE, capacity = %02x\n",k->packet[1]);
        // We don't use this, throw it away...
        break;
        
    case IOC_BATTERY_DISCHARGE_RESPONSE:    // Battery Discharge Current
        info->charging = 0;
        info->draw = ioc_convert_raw_draw_to_mamps(k->packet[1]);
        BDEBUG("update_battery_info(): DISCHARGE RESPONSE, draw = %02x\n",info->draw);
        break;
        
    case IOC_BATTERY_CHARGE_RESPONSE:       // Battery Charge Current
        info->charging = 1;
        info->draw = ioc_convert_raw_draw_to_mamps(k->packet[1]);
        BDEBUG("update_battery_info(): CHARGE RESPONSE, draw = %02x\n",info->draw);
        break;
        
    default:
        BDEBUG("update_battery_info(): UNKNOWN BATTERY PACKET TYPE %02x\n",k->packet[0]);
        break;

    }
}

int get_raw_battery_mili_voltage(void)
{
    int mvolts = 0;
    int retVal = 0;
    char response[MAX_IOC_RX_PACKET_SIZE];

    if (battery_installed()) {
        retVal = IOC_Query_Cmd(IOC_QUERY_BATTERY_LEVEL, response, 0);
        if (retVal == 0)
            mvolts = ioc_convert_raw_voltage_to_mvolts((unsigned int)response[1]);
    }
    return(mvolts);
}

// FIXME - Make this routine actually do something
static unsigned long generateBoardActivity(int stage)
{
    switch(stage)  {
        case 0:
            // Do something here to generate activity
            set_wan_power(WAN_ON);
            break;

        case 1:
            // Do something here to stop generated activity
            set_wan_power(WAN_OFF_KILL);
            break;

        default:
            break;
    }
}



/*******************************************************************
 * Routine:     calculate_board_resistance
 * Purpose:     Dynamically calculate the board's resistance value.
 * Paramters:   none
 * Returns:     0 - Couldn't calculate resistance
 *              non-0 - board resistance value
 * Side Effects:
 *  Updates watermark boot globals if resitance calculated properly
 *  Updates watermark fields in g_fiona_hw structure.
 *******************************************************************/
unsigned long calculate_board_resistance(void)
{
    unsigned long resistance = board_resistance; // value passed from bootloader
    unsigned long amp_sum = 0, volt_sum = 0, temp_sum = 0;
    unsigned long temp[RESISTANCE_CALCULATION_SAMPLE_COUNT];
    unsigned long volts[RESISTANCE_CALCULATION_SAMPLE_COUNT];
    unsigned long amps[RESISTANCE_CALCULATION_SAMPLE_COUNT];
    Battery_Info_T  info;

    int sample_count = 0;
    int x = 0, y = 0;
    int volt_1 = 0, volt_2 = 0;
    int amp_1 = 0, amp_2 = 0;
    int temp_1 = 0, temp_2 = 0;
    int delta_v = 0, delta_a = 0, delta = 0;
    int ioc_version = 0;

    // Verify we have a new enough version of firmware
    // If AC plug is inserted, we can't calculate resistance, so bail...
    get_ioc_firmware_version(&ioc_version, 0);
    if ( ac_adapter_installed() || (ioc_version < IOC_QUERY_BATTERY_INFO_CMD_VERSION)) {
        resLog("fiona_calculate_board_resistance(AC/Old IOC) returning default value %d\n",board_resistance);
        return(board_resistance);
    }

    for (x=0; x<2; x++)  {
        // Go perform the calculation
        sample_count = 0;
        while (sample_count < RESISTANCE_CALCULATION_SAMPLE_COUNT) {

            // Get battery info
#if 0
            dbg_bt_putc(IOC_QUERY_BATTERY_INFO);
            if (0 == get_ioc_response(IOC_QUERY_BATTERY_INFO_RESPONSE_PACKET_TYPE, 0, response)) {
                resLog("calculate_board_resistance() ERRORED getting battery info from IOC..\n");
                return 0;
            }
#else
           get_battery_info(&info,1);   // Make sure to get raw values
#endif

            volts[sample_count] = info.mvolts;
            amps[sample_count] = info.draw;
            temp[sample_count] = info.temperature;
            sample_count++;
        }

        // Now average our samples
        amp_sum = volt_sum = temp_sum = 0;
        for (y=0;y<sample_count;y++) {
            resLog("Resistance sample %d: %d v, %d mA, %d C\n",y+1,(int)volts[y],(int)amps[y],(int)temp[y]);
            volt_sum += volts[y];
            amp_sum += amps[y];
            temp_sum += temp[y];
        }
        volt_sum /= sample_count;
        amp_sum /= sample_count;
        temp_sum /= sample_count;

        if (x==0)  {
            volt_1 = volt_sum;
            amp_1 = amp_sum;
            temp_1 = temp_sum;
        }
        else  {
            volt_2 = volt_sum;
            amp_2 = amp_sum;
            temp_2 = temp_sum;
        }

        // Generate (or stop) a load to draw amperage up/down
        generateBoardActivity(x);
    }

    // Take 2 readings...
    if ((volt_2 != volt_1) && (amp_2 != amp_1)) {
        resistance = 0;
        delta_v = ((volt_2-volt_1) * 1000);   // 1000 converts to mOhms
        delta_a = amp_1-amp_2;
        delta = (delta_v / delta_a);
        resistance = abs(delta);

        resLog("Resistance Calc: %d  ( abs((%d-%d)*1000/(%d-%d)))\n", \
                (int)resistance,volt_2,volt_1,amp_1,amp_2);

        if ((resistance >= MIN_BOARD_RESISTANCE) && 
            (resistance <= MAX_BOARD_RESISTANCE))  {
            // Set the kernel global
            board_resistance = resistance;
            resLog("Board Resistance Calc: %d mOhms)\n",(int)resistance);
        }
        else
            resLog("Board Resistance Calc OUT-OF-BOUNDS: %d mOhms)\n",(int)resistance);
    }
    else resLog("Not enough good mixed samples to do delta method !!\n");

    return(resistance);
}

/*******************************************************************
 * Routine:     calculate_battery_watermarks
 * Purpose:     Dynamically calculate the watermarks based on the
 *              board's maximum current draw for the various modes
 *              and the board's resistance value.
 * Paramters:   none
 * Returns:     0 - Couldn't calculate watermarks
 *              1 - Calculated watermarks and updated boot global
 * Side Effects:
 *  Updates watermark boot globals if resitance calculated properly
 *  Updates watermark fields in g_fiona_hw structure.
 *******************************************************************/
int calculate_battery_watermarks(void)
{
    int wan = 0;
    int low = 0;
    int critical = 0;
    bat_level_t battery_watermarks;
    int resistance = g_max_board_resistance;

    //
    // If the board does not have the rve calibration number set, we need to make sure we have
    // enough safety buffer so that the device is not in danger of batt_fault'ing
    //
    int conservative_factor = ((g_rve_correction_value == UNINITIALIZED_RVE_VALUE) ? WATERMARK_OP_AMP_CONSERVATIVE_FACTOR_MV : 0);
    memcpy(&battery_watermarks, get_battery_watermarks(), sizeof(bat_level_t));
    battery_watermarks.update_env = 0;

#ifndef USE_DEFAULT_BOARD_RESISTANCE
    //
    // Since we're using the device's current board resistance, we have to make sure watermarks
    // are set conservatively to handle resistance gain of the battery over time and
    // still have the units function reliably.  WATERMARK_CONSERVATIVE_RESISTANCE_GAIN
    // is added to reported board resistance when determining actual watermark settings.
    // By doing this, we sacrifice a bit of battery life in trade for boards not crashing
    // at some point in the future as their battery gains resistance.
    //
    resistance += WATERMARK_CONSERVATIVE_RESISTANCE_GAIN;
#endif

    // Calculate the lowest level we can go with WAN enabled
    //
    wan = (HW_VOLTAGE_SHUTOFF_LEVEL + ((MAX_MAMPS_WAN_MODE * resistance)/1000) \
                + conservative_factor);

    // Calculate the lowest level we can go and safely update the display
    //
    critical = (HW_VOLTAGE_SHUTOFF_LEVEL + ((MAX_MAMPS_NON_WAN_MODE * resistance)/1000) \
                + conservative_factor + WATERMARK_CONSERVATIVE_CRITICAL_FACTOR_MV);

    // Critical will be the lowest level we can safely go, so we set low to be 
    // WATERMARK_CONSERVATIVE_LOW_FACTOR_MV above that line...
    //
    low = (critical + WATERMARK_CONSERVATIVE_LOW_FACTOR_MV);


    resLog("Recalculated watermarks:\n  wan:      %d (was %d)\n"           \
                                     "  low:      %d (was %d)\n"           \
                                     "  critical: %d (was %d)\n",       \
                wan,(int)battery_watermarks.stage[BAT_LEVEL_WAN],       \
                low, (int)battery_watermarks.stage[BAT_LEVEL_LOW],      \
                critical, (int)battery_watermarks.stage[BAT_LEVEL_CRIT]);


    // Store off the new calculated watermarks
    g_fiona_hw.wan_watermark = wan;
    g_fiona_hw.low_watermark = low;
    g_fiona_hw.critical_watermark = critical;

#ifdef CONFIG_SAFETY_NET
    // For the safety nets, sit 100mV above the HW cutoff point
    g_fiona_hw.wan_safety_net = WAN_SAFETY_NET_MVOLTS; 
    g_fiona_hw.low_safety_net = LOW_SAFETY_NET_MVOLTS;
    g_fiona_hw.critical_safety_net = CRITICAL_SAFETY_NET_MVOLTS;
    
    resLog("Watermark Safety Nets :\n  wan:      %d\n"           \
                                    "  low:      %d\n"           \
                                    "  critical: %d\n",       \
                (int)g_fiona_hw.wan_safety_net,(int)g_fiona_hw.low_safety_net,(int)g_fiona_hw.critical_safety_net);
#endif

    // Don't store them back to the boot globals unless one or more have changed.
    if (wan != battery_watermarks.stage[BAT_LEVEL_WAN]) {
        battery_watermarks.stage[BAT_LEVEL_UPDT] = (unsigned short)CALC_WATERMARK_UPDATE(wan);
        battery_watermarks.stage[BAT_LEVEL_WAN]  = (unsigned short)wan;
        battery_watermarks.update_env = 1;
    }
    
    if (low != battery_watermarks.stage[BAT_LEVEL_LOW]) {
        battery_watermarks.stage[BAT_LEVEL_LOW] = (unsigned short)low;
        battery_watermarks.update_env = 1;
    }
    
    if (critical != battery_watermarks.stage[BAT_LEVEL_CRIT]) {
        battery_watermarks.stage[BAT_LEVEL_CRIT] = (unsigned short)critical;
        battery_watermarks.update_env = 1;
    }
    
    if (battery_watermarks.update_env) {
        set_battery_watermarks(&battery_watermarks);
    }

    return(1);
}


/////////////////////////////////////////////
// HW Function Enable/Disable Routines
/////////////////////////////////////////////

void IOC_Reset_Fiona(void)
{
    char cmd = IOC_CMD_RESET_FIONA;
    IOC_IssueCmd(&cmd, 1, NULL, 0);
}

static int
Do_Enable_IOC(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE) ? IOC_CMD_ENABLE_IOC : IOC_CMD_DISABLE_IOC );
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_IOC_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_IOC_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];

    memclr(response,MAX_BYTES_IOC_RESPONSE+1);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("IOC", enable, rc, response_size);
        return(1);  // Return error
    }
    g_ioc_enabled = enable;

#ifdef LOCAL_CONFIG_PM
    g_ioc_wake_state = enable;  // Save off this state
#endif
    return(0);
}

int get_current_lowbatt_mvolt_level(void)  {
    return(g_ioc_host_watermark); 
}

int
IOC_Enable_Battery_Voltage_Low_Reporting_PRE_43(unsigned short enable, int level)
{
    int rc = 0;
    char cmd;
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_LOW_WARNING_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_BATTERY_LOW_WARNING_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];

    memclr(response,MAX_BYTES_IOC_RESPONSE+1);
    cmd = ((enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_LOW_REPORTING : \
                                    IOC_CMD_DISABLE_BATTERY_LOW_REPORTING );

    powdebug("IOC: IOC_Enable_Battery_Voltage_Low_Reporting_PRE_43(%d,%d)\n",enable,level);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Battery Low Voltage Reporting", enable, rc, response_size);
        return(1);  // Return error
    }
    powdebug("BATTERY: %sabled low voltage reporting.\n",(enable==ENABLE)?"En":"Dis");
    return(0);
}


#ifdef CONFIG_IOC_VOLTAGE_CHECKER
void printBatteryInfo(char *str, Battery_Info_T *info)
{
    printk("***********************************\n");
    printk("%s\n",str);
    printk("  mvolts   : %d\n",info->mvolts);
    printk("  raw mv   : %d\n",info->raw_mvolts);
    printk("  icon mv  : %d\n",info->icon_mvolts);
    printk("  temp     : %d\n",info->temperature);
    printk("  mamps    : %d\n",info->draw);
    //printk("  charging : %d\n",info->charging);
    printk("***********************************\n");
}

static void printVoltageInfo(Battery_Info_T *current_info, Battery_Info_T *average_info)
{
    if (g_voltage_dump_debug)
        printk("VOLTAGE: %d, %d, %d, %d, %d\n",average_info->icon_mvolts, average_info->mvolts, current_info->mvolts,current_info->raw_mvolts, current_info->draw);
}


void printVoltageTable(int force)
{
    Battery_Info_T info;
    int sampleCount = (int)MIN(g_voltage_sample_count,g_vc_sample_count);

    // If we're not supposed to force the print, and the global voltage logging flag
    // is turned off, then return now and don't allow printing of the table
    if ((g_fiona_vtable_debug == 0) && (force == 0) && (g_fiona_voltage_debug == 0))
        return;

    memset(&info,0,sizeof(info));
    get_battery_info_from_samples(&info);
    printk("Battery: %d mV, %d imV, %d mA, %d C, S:%d\n",info.mvolts, info.icon_mvolts, info.draw, info.temperature,sampleCount);
}

void printVoltageTableFull(int force)
{
    int x = 0;
    int sampleCount = (int)MIN(g_voltage_sample_count,g_vc_sample_count);

    // If we're not supposed to force the print, and the global voltage logging flag
    // is turned off, then return now and don't allow printing of the table
    if ((g_fiona_vtable_debug == 0) && (force == 0) && (g_fiona_voltage_debug == 0))
        return;

    printk("V[%d] : ",sampleCount);
    for (x=0; x<sampleCount; x++)  {
        if (x == (g_next_voltage_sample?g_next_voltage_sample-1:g_vc_sample_count-1)){
            // Make last sample value apparent for user
            if (x==(sampleCount-1))
                printk("~%d~",(int)g_voltage_samples[x]);
            else
                printk("~%d~, ",(int)g_voltage_samples[x]);
        }
        else {
            if (x==(sampleCount-1))
                printk("%d",(int)g_voltage_samples[x]);
            else
                printk("%d, ",(int)g_voltage_samples[x]);
        }
    }
    printk("\n");


    printk("ICON[%d] : ",sampleCount);
    for (x=0; x<sampleCount; x++)  {
        if (x == (g_next_icon_voltage_sample?g_next_icon_voltage_sample-1:g_vc_sample_count-1)){
            // Make last sample value apparent for user
            if (x==(sampleCount-1))
                printk("~%d~",(int)g_icon_voltage_samples[x]);
            else
                printk("~%d~, ",(int)g_icon_voltage_samples[x]);
        }
        else {
            if (x==(sampleCount-1))
                printk("%d",(int)g_icon_voltage_samples[x]);
            else
                printk("%d, ",(int)g_icon_voltage_samples[x]);
        }
    }
    printk("\n");

    printk("A[%d] : ",sampleCount);
    for (x=0; x<sampleCount; x++)  {
        if (x == (g_next_amperage_sample?g_next_amperage_sample-1:g_vc_sample_count-1)){
            // Make last sample value apparent for user
            if (x==(sampleCount-1))
                printk("~%d~",(int)g_amperage_samples[x]);
            else
                printk("~%d~, ",(int)g_amperage_samples[x]);
        }
        else {
            if (x==(sampleCount-1))
                printk("%d",(int)g_amperage_samples[x]);
            else
                printk("%d, ",(int)g_amperage_samples[x]);
        }
    }
    printk("\n");

    printk("T[%d] : ",sampleCount);
    for (x=0; x<sampleCount; x++)  {
        if (x == (g_next_amperage_sample?g_next_amperage_sample-1:g_vc_sample_count-1)){
            // Make last sample value apparent for user
            if (x==(sampleCount-1))
                printk("~%d~",(int)g_temperature_samples[x]);
            else
                printk("~%d~, ",(int)g_temperature_samples[x]);
        }
        else {
            if (x==(sampleCount-1))
                printk("%d",(int)g_temperature_samples[x]);
            else
                printk("%d, ",(int)g_temperature_samples[x]);
        }
    }
    printk("\n");
}

static int rangeCheckBatteryValues(Battery_Info_T *info)
{
    if ((info->mvolts >= MIN_MVOLT_VALUE) && (info->mvolts <= MAX_MVOLT_VALUE)) {
        if ((info->draw >= MIN_MAMP_VALUE) && (info->draw <= MAX_MAMP_VALUE))
            return(1);  // Range check OK
        else
            printk("ERROR: Amperage %d out-of-range of %d --> %d\n",info->draw,MIN_MAMP_VALUE,MAX_MAMP_VALUE);
    }
    else
        printk("ERROR: Voltage %d out-of-range of %d --> %d\n",info->mvolts,MIN_MVOLT_VALUE,MAX_MVOLT_VALUE);
    return(0);          // Range check failed....

}

static int adjust_icon_voltage_average(int icon_mvolts)
{
    if (g_icon_mvolt_average == 0) {
        g_icon_mvolt_average = icon_mvolts;
    }
    else if (icon_mvolts > g_icon_mvolt_average) {
        g_icon_mvolt_average = g_icon_mvolt_average + ((icon_mvolts - g_icon_mvolt_average)/(1<<g_icon_rising_lambda));
    }
    else {
        g_icon_mvolt_average = g_icon_mvolt_average + ((icon_mvolts - g_icon_mvolt_average)/(1<<g_icon_falling_lambda));
    }
    return(g_icon_mvolt_average); 
}

static int get_icon_voltage_average(void)
{
    return(g_icon_mvolt_average);
}

static int clear_icon_voltage_average(void)
{
    g_icon_mvolt_average = 0;
    return(g_icon_mvolt_average);
}


void add_battery_info_to_samples(Battery_Info_T *info)
{
    // Set our charge and installed information, they should 
    // be accurate even if range check fails....
    g_battery_info.charging = info->charging;
    g_battery_info.installed = info->installed;

    // Range-check values
    if (0 == rangeCheckBatteryValues(info)) {
        return;     // bogus values, don't add to table...
    }

    // Average in current.
    g_amperage_samples[g_next_amperage_sample++] = info->draw; 
    if (g_next_amperage_sample >= g_vc_sample_count) //AMPERAGE_AVERAGE_SAMPLE_COUNT)
        g_next_amperage_sample = 0;

    // Average in adjusted voltage...
    g_voltage_samples[g_next_voltage_sample++] = info->mvolts; 
    if (g_next_voltage_sample >= g_vc_sample_count)
        g_next_voltage_sample = 0;

    // Average in battery gauge voltage, which uses a different board resistance
    // for it's calculation....
    g_icon_voltage_samples[g_next_icon_voltage_sample++] = info->icon_mvolts;
    if (g_next_icon_voltage_sample >= g_vc_sample_count)
        g_next_icon_voltage_sample = 0;

    // Adjust battery icon voltage average
    adjust_icon_voltage_average(info->icon_mvolts);

    // Average in temperature
    g_temperature_samples[g_next_temp_sample++] = info->temperature; 
    if (g_next_temp_sample >= g_vc_sample_count) // TEMPERATURE_AVERAGE_SAMPLE_COUNT)
        g_next_temp_sample = 0;

    g_voltage_sample_count++;
}


void add_battery_info_to_samples_full(Battery_Info_T *info)
{
#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    int mv = info->mvolts;
    
    // Set our charge and installed information, they should 
    // be accurate even if range check fails....
    g_battery_info.charging = info->charging;
    g_battery_info.installed = info->installed;

    // Range-check values
    if (0 == rangeCheckBatteryValues(info))
        return;     // bogus values, don't add to table...

    // Add to appropriate tables....
    if ((g_voltage_sample_count % 2) == 0) {
        g_voltage_samples_2_sec[g_next_voltage_sample_2_sec++] = mv; 
        g_2_sample_count++;
        if (g_next_voltage_sample_2_sec >= g_vc_sample_count)
            g_next_voltage_sample_2_sec = 0;
    }

    if ((g_voltage_sample_count % 5) == 0) {
        g_voltage_samples_5_sec[g_next_voltage_sample_5_sec++] = mv; 
        g_5_sample_count++;
        if (g_next_voltage_sample_5_sec >= g_vc_sample_count)
            g_next_voltage_sample_5_sec = 0;
    }

    if ((g_voltage_sample_count % 8) == 0) {
        g_voltage_samples_8_sec[g_next_voltage_sample_8_sec++] = mv; 
        g_8_sample_count++;
        if (g_next_voltage_sample_8_sec >= g_vc_sample_count)
            g_next_voltage_sample_8_sec = 0;
    }

    if ((g_voltage_sample_count % 10) == 0) {
        g_voltage_samples_10_sec[g_next_voltage_sample_10_sec++] = mv; 
        g_10_sample_count++;
        if (g_next_voltage_sample_10_sec >= g_vc_sample_count)
            g_next_voltage_sample_10_sec = 0;
    }

    if ((g_voltage_sample_count % 20) == 0) {
        g_voltage_samples_20_sec[g_next_voltage_sample_20_sec++] = mv; 
        g_20_sample_count++;
        if (g_next_voltage_sample_20_sec >= g_vc_sample_count)
            g_next_voltage_sample_20_sec = 0;
    }
#endif

    // Add sample post so g_voltage_sample_count doesn't change prior to 
    // above code running.
    // Add the sample to the main table
    add_battery_info_to_samples(info);
}


#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
static int ga(int table_id)
{
    unsigned long count = 0;
    unsigned long sum = 0, avg = 0;
    unsigned int x = 0;
    unsigned long *samples = NULL;

    switch(table_id) {
        case 2:
            count = MIN(g_2_sample_count,g_vc_sample_count);
            samples = &g_voltage_samples_2_sec[0];
            break;
        case 5:
            count = MIN(g_5_sample_count,g_vc_sample_count);
            samples = &g_voltage_samples_5_sec[0];
            break;
        case 8:
            count = MIN(g_8_sample_count,g_vc_sample_count);
            samples = &g_voltage_samples_8_sec[0];
            break;
        case 10:
            count = MIN(g_10_sample_count,g_vc_sample_count);
            samples = &g_voltage_samples_10_sec[0];
            break;
        case 20:
            count = MIN(g_20_sample_count,g_vc_sample_count);
            samples = &g_voltage_samples_20_sec[0];
            break;
        default:
            printk("Invalid table id %d\n",table_id);
            return(0);
    }

    // If we don't have any samples, average is 0
    if (count == 0)  {
        return(0);
    }
    for (x=0; x<count; x++)  {
        sum += samples[x];
    }
    avg = (sum/count);
    return(avg);
}
#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR

static unsigned long get_average_of_table(unsigned long samples[], unsigned long count) {
    unsigned        x = 0;
    unsigned long   sum = 0, avg = 0;
    for (x=0; x<count; x++)
        sum += samples[x];
    avg = (sum/count);
    return(avg);
}


void get_battery_info_from_samples(Battery_Info_T *info)
{
    unsigned long v_count = MIN(g_voltage_sample_count,g_vc_sample_count);
    unsigned long a_count = MIN(g_voltage_sample_count,g_vc_sample_count);
    unsigned long t_count = MIN(g_voltage_sample_count,g_vc_sample_count);

    // If we don't have any samples, clear fields and return
    if (v_count == 0)  {
        info->mvolts = 0;
        info->icon_mvolts = 0;
        info->draw = 0;
        info->temperature = 0;
        info->charging = g_battery_info.charging;
        info->installed = g_battery_info.installed;
        return;
    }

    // Calculate the averages and return those values 
    info->mvolts = get_average_of_table(g_voltage_samples,v_count);
    info->icon_mvolts = get_icon_voltage_average();
    info->draw = get_average_of_table(g_amperage_samples,a_count);
    info->temperature = get_average_of_table(g_temperature_samples,t_count);

    // Pass back other known information
    info->charging = g_battery_info.charging;
    info->installed = g_battery_info.installed;
}

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
void clear_voltage_test_tables(void)
{
    int x;
    for (x=0; x<g_vc_sample_count; x++) {
        g_voltage_samples_2_sec[x] = 0;
        g_voltage_samples_5_sec[x] = 0;
        g_voltage_samples_8_sec[x] = 0;
        g_voltage_samples_10_sec[x] = 0;
        g_voltage_samples_20_sec[x] = 0;
    }
    g_next_voltage_sample_2_sec = 0;
    g_next_voltage_sample_5_sec = 0;
    g_next_voltage_sample_8_sec = 0;
    g_next_voltage_sample_10_sec = 0;
    g_next_voltage_sample_20_sec = 0;
    g_2_sample_count = 0;
    g_5_sample_count = 0;
    g_8_sample_count = 0;
    g_10_sample_count = 0;
    g_20_sample_count = 0;
}
#endif // CONFIG_MULTI_RATE_AVERAGING_MONITOR 

void clear_voltage_table(void)
{
    int x;

    g_next_amperage_sample = 0;         // Reset amperage table
    g_next_voltage_sample = 0;          // Reset voltage table
    g_next_icon_voltage_sample = 0;     // Reset battery icon voltage table
    g_next_temp_sample = 0;             // Reset temperature table
    g_voltage_sample_count = 0;         // Clear sample count, no samples left
    g_voltage_checker_needs_reset = 0;
    clear_icon_voltage_average();       // Clear the icon voltage samples...

    // Clear them out to make logging less confusing
    for (x=0; x<g_vc_sample_count; x++) {
        g_voltage_samples[x] = 0;
        g_icon_voltage_samples[x] = 0;
    }
    for (x=0; x<AMPERAGE_AVERAGE_SAMPLE_COUNT; x++)
        g_amperage_samples[x] = 0;
    for (x=0; x<TEMPERATURE_AVERAGE_SAMPLE_COUNT; x++)
        g_temperature_samples[x] = 0;

#ifdef CONFIG_MULTI_RATE_AVERAGING_MONITOR
    clear_voltage_test_tables();
#endif
}

void log_battery_voltage(Battery_Info_T *average_info, Battery_Info_T *current_sample_info)
{
    int mv = 0;
    if (g_watermark_algorithm == VOLTAGE_ALGORITHM_2)
        mv = average_info->icon_mvolts;
    else
        mv = average_info->mvolts;
    printk("Battery: %d mVolts (raw: %d mVolts, %d mA)\n",(int)mv,current_sample_info->raw_mvolts,(int)current_sample_info->draw);
}


int _IOC_Voltage_Sampler(void)
{
    int retVal = 0;
    Battery_Info_T current_sample_info;
    Battery_Info_T average_info;
    struct timeval tv;
    unsigned long curTime = 0;
    unsigned long mv = 0;
    unsigned long amps = 0;
    unsigned int raw_mv = 0;
    int did_ioc_reset = 0;
    unsigned long jiffs = 0;
    int do_print_watermark_hit = 0;

    // Clear structs
    memset(&average_info,0,sizeof(average_info));
    memset(&current_sample_info,0,sizeof(current_sample_info));

enter:
    // If we have a reset pending, do it now...
    if (g_voltage_checker_needs_reset) {
        powdebug("_IOC_Voltage_Sampler() resetting voltage averaging table\n");
        clear_voltage_table();
    }

    // Value already adjusted by the time we get it...
    retVal = get_battery_info(&current_sample_info,0);
    if (retVal != 0)  {
        printk("ERROR %d returned from get_battery_info() !!\n",retVal);
        return(retVal);
    }
    raw_mv = current_sample_info.raw_mvolts;
    amps = current_sample_info.draw;

    // If we just got a 0 value from the table, we don't have valid data yet,
    // so head back...
    // Most likely reason this happened is that the IOC has gone south.  
    // We reset the IOC in this case and try again (only 1 retry, though, if a
    // reset doesn't fix it, it's not the issue
    if (raw_mv == 0) {
        if (did_ioc_reset == 1) {
            printk("WATER: 0 voltage read from IOC again, bailing out of _IOC_Voltage_Sampler()...\n");
            goto exit;
        }
        printk("WATER: 0 voltage read from IOC, resetting IOC...\n");
        globs.hal->IOC_HAL_ResetChip(globs.hal);
        did_ioc_reset = 1;
        goto enter;
    }

    // Add the current voltage sample to the tables
    add_battery_info_to_samples_full(&current_sample_info);

    // Read back out the running averages
    get_battery_info_from_samples(&average_info);

    // Check for a bad voltage read
    if (g_watermark_algorithm == VOLTAGE_ALGORITHM_1)
        mv = (unsigned long) average_info.icon_mvolts;
    else
        mv = (unsigned long) average_info.mvolts;
    if (mv == 0) {
        printk("WATER: 0 voltage read from IOC voltage tables.\n");
        goto exit;
    }

    // Log Voltage for User
    do_gettimeofday(&tv);
    if ((voltage_level_log_freq != 0) && (next_voltage_level_log_time <= tv.tv_sec))  {
        next_voltage_level_log_time = (tv.tv_sec + voltage_level_log_freq);
        log_battery_voltage(&average_info, &current_sample_info);
    }

    // Dump voltage table, forcing it if we're told to force print
    printVoltageTable(0);   
    printVoltageInfo(&current_sample_info, &average_info);

    // Save off current time in jiffies
    jiffs = jiffies;

#ifdef CONFIG_SAFETY_NET
    //
    // Check our critical zone first, make sure adjusted voltages aren't getting out-of-whack
    // We watch the spikes (raw voltage) and if we ever see one penetrate our safety net,
    // send the warning....
    //
    if (g_ioc_voltage_check_enabled && (g_ioc_host_safety_net >= raw_mv))  {
        printk("WATER SAFETY_NET: %d mVolts (raw: %d mV, limit: %d mV)\n",(int)info.mvolts,raw_mv,(int)g_ioc_host_safety_net);
        if (g_fiona_hw.wan_watermark == g_ioc_host_watermark) { 
            waterLog("SAFETY_NET: NO_WAN_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
            handle_fiona_event(BATTERY_NO_WAN_EVENTCODE);
            goto exit;
        }
        else {
            // If we peirced the low or critical watermark, it's critical (no difference between HW setup between
            //  the two "states", so it's critical 
            waterLog("SAFETY_NET: CRITICAL_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
            handle_fiona_event(BATTERY_CRITICAL_EVENTCODE);
            goto exit;
        }
    }
#endif


    // If we don't have enough samples, exit now...
    if (g_voltage_sample_count < MIN_SAMPLES_FOR_WATERMARK_AVERAGE) {
        waterLog("Not enough samples yet...\n");
        printVoltageTable(0);
        goto exit;
    }

    // If we hit our watermark, send a low-battery event...
    if (g_ioc_voltage_check_enabled && (g_ioc_host_watermark >= mv))  {
        waterLog("IOC_Voltage_Sampler() detected a low battery condition (%d<%d)\n",(int)mv,(int)g_ioc_host_watermark);
        do_gettimeofday(&tv);
        curTime = tv.tv_sec;

        // Determine what type of low-battery event we have
        if (mv <= fpowGetCriticalWatermark())  {
            if (!time_after(jiffs,g_critical_watermark_timeout))  {
                powdebug("CRITICAL_BATTERY_EVENT NOT sent (last sent %d sec ago)\n", (int)((jiffs - (g_critical_watermark_timeout - MIN_WATERMARK_JIFF_TIMEOUT))/HZ));
                goto exit;
            }
            do_print_watermark_hit = 1;
            printk("CRITICAL_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
            handle_fiona_event(BATTERY_CRITICAL_EVENTCODE);
            g_critical_watermark_timeout = (jiffs + MIN_WATERMARK_JIFF_TIMEOUT);
            printVoltageTable(PRINT_VTABLE_ON_WATERMARK_HIT);
        }
        else if (mv <= fpowGetLowWatermark())  {
            if (g_ioc_host_watermark == fpowGetWanWatermark())  {
                if (!time_after(jiffs,g_wan_watermark_timeout))  {
                    powdebug("NO_WAN_BATTERY_EVENT NOT sent (last sent %d sec ago)\n",(int)((jiffs - (g_wan_watermark_timeout - MIN_WATERMARK_JIFF_TIMEOUT))/HZ));
                    goto exit;
                }
                do_print_watermark_hit = 1;
                printk("NO_WAN_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
                handle_fiona_event(BATTERY_NO_WAN_EVENTCODE);
                g_wan_watermark_timeout = (jiffs + MIN_WATERMARK_JIFF_TIMEOUT);
            }
            if (!time_after(jiffs, g_low_watermark_timeout))  {
                powdebug("LOW_BATTERY_EVENT NOT sent (last sent %d sec ago)\n",(int)((jiffs - (g_wan_watermark_timeout - MIN_WATERMARK_JIFF_TIMEOUT))/HZ));
                goto exit;
            }

            do_print_watermark_hit = 1;
            printk("LOW_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
            handle_fiona_event(BATTERY_LOW_EVENTCODE);
            g_low_watermark_timeout = (jiffs + MIN_WATERMARK_JIFF_TIMEOUT);
            printVoltageTable(PRINT_VTABLE_ON_WATERMARK_HIT);
        }
        else if (mv <= fpowGetWanWatermark())  {
            if (!time_after(jiffs, g_wan_watermark_timeout))  {
                powdebug("NO_WAN_BATTERY_EVENT NOT sent (last sent %d sec ago)\n",(int)((jiffs - (g_wan_watermark_timeout - MIN_WATERMARK_JIFF_TIMEOUT))/HZ));
                goto exit;
            }
            do_print_watermark_hit = 1;
            printk("NO_WAN_BATTERY_EVENT sent (raw_mv = %d, mv = %d) at %d sec\n",raw_mv,(int)mv,(int)curTime);
            handle_fiona_event(BATTERY_NO_WAN_EVENTCODE);
            g_wan_watermark_timeout = (jiffs + MIN_WATERMARK_JIFF_TIMEOUT);
            printVoltageTable(PRINT_VTABLE_ON_WATERMARK_HIT);
        }
        else  {
            powdebug("False alarm, voltage of %lu > wan watermark of %lu (set limit = %lu\n",(long unsigned int)mv,(long unsigned int)fpowGetWanWatermark(),g_ioc_host_watermark);
        }
    } 
exit:
#ifdef LOG_EVERY_VOLT_READING
    printk("%lu mV  %d mA\n",mv,(int)current_sample_info.draw);
#endif
    
    // Print the watermark hit if needed
    if (do_print_watermark_hit != 0)
       printk("WATERMARK HIT: %d mVolts (raw: %d mV, limit: %d mV)\n",(int)mv,raw_mv,(int)g_ioc_host_watermark);

    return(0);
}


int IOC_Voltage_Sampler(void)
{
    int retVal = 0;

    // First, get the lock
    lockdebug("Getting g_ioc_hw_lock in IOC_Voltage_Sampler()...\n");
    if (down_interruptible(&g_ioc_hw_lock))
        return(-ERESTARTSYS);
    lockdebug("Got g_ioc_hw_lock in IOC_Voltage_Sampler()...\n");

    retVal = _IOC_Voltage_Sampler();

    up(&g_ioc_hw_lock);
    lockdebug("Released g_ioc_hw_lock in IOC_Voltage_Sampler()...\n");

    return(retVal);
}
#endif

int
IOC_Enable_Battery_Voltage_Low_Reporting_At_IOC(unsigned short enable, int level)
{
    int raw_mv = 0;
    int rc = 0;
    int cmd_size = (enable?IOC_CMD_ENABLE_BATTERY_LOW_WARNING_CMD_BYTES:IOC_CMD_DISABLE_BATTERY_LOW_WARNING_CMD_BYTES);
    char cmd[cmd_size];

    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_LOW_WARNING_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_BATTERY_LOW_WARNING_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];

    // If firmware is older than version 43, use older protocol
    if (g_fw_version < 43)  {
        powdebug("IOC: version < 43, calling IOC_Enable_Battery_Voltage_Low_Reporting_PRE_43(%d,%d)\n",enable,level);
        return(IOC_Enable_Battery_Voltage_Low_Reporting_PRE_43(enable,level));
    }

    memclr(response,MAX_BYTES_IOC_RESPONSE+1);
    cmd[0] = ((enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_LOW_REPORTING : \
                                    IOC_CMD_DISABLE_BATTERY_LOW_REPORTING );
    if (enable)  {
        raw_mv = ioc_convert_mvolts_to_raw_voltage(level);
        cmd[1] = raw_mv;
    }

    powdebug("IOC: IOC_Enable_Battery_Voltage_Low_Reporting_At_IOC(%d,%d), raw=0x%x\n",enable,level,raw_mv);

    // Issue the command
    rc = IOC_IssueCmd(cmd, cmd_size, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Battery Low Voltage Reporting", enable, rc, response_size);
        return(1);  // Return error
    }

    powdebug("BATTERY: %sabled low voltage reporting at IOC at %d mVolts.\n",(enable==ENABLE)?"En":"Dis",level);
    return(0);
}

#ifdef CONFIG_SAFETY_NET
int
get_level_safety_net(int level)
{
    if (g_fiona_hw.wan_watermark == level)
        return(g_fiona_hw.wan_safety_net);
    else if (g_fiona_hw.wan_watermark == level)
        return(g_fiona_hw.low_safety_net);
    else
        return(g_fiona_hw.critical_safety_net);
}
#endif

int
IOC_Enable_Battery_Voltage_Low_Reporting(unsigned short enable, int level)
{
#ifdef CONFIG_IOC_VOLTAGE_CHECKER
    if (enable) {
        g_ioc_voltage_check_enabled = 1;
        g_ioc_host_watermark = level;
#ifdef CONFIG_SAFETY_NET
        g_ioc_host_safety_net = get_level_safety_net(level);
#endif
    }
    else {
        g_ioc_voltage_check_enabled = 0;
    }
    waterLog("Set host ioc watermark to %d\n",(enable?((int)g_ioc_host_watermark):0));

#else
    IOC_Enable_Battery_Voltage_Low_Reporting_At_IOC(enable,level);
#endif

    powdebug("BATTERY: %sabled low voltage reporting at %d mVolts.\n",(enable==ENABLE)?"En":"Dis",level);
    return(0);
}

static int
IOC_Enable_Battery_Capacity_Reporting(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_CAPACITY_REPORTING : \
                                    IOC_CMD_DISABLE_BATTERY_CAPACITY_REPORTING);
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_BATTERY_LEVEL_REPORTING_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_BATTERY_LEVEL_REPORTING_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE+1);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Battery Capacity Reporting", enable, rc, response_size);
        return(1);  // Return error
    }
    powdebug("BATTERY: %sabled battery capacity reporting.\n",(enable==ENABLE)?"En":"Dis");
    return(0);
}

static int 
IOC_Enable_Keyboard(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE) ? IOC_CMD_ENABLE_KEYBOARD : IOC_CMD_DISABLE_KEYBOARD );
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_KEYBOARD_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_KEYBOARD_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Keyboard", enable, rc, response_size);
        return(1);  // Return error
    }
    return(0);
}



int
IOC_Enable_Buttons(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE) ? IOC_CMD_ENABLE_BUTTONS : IOC_CMD_DISABLE_BUTTONS );
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_BUTTONS_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_BUTTONS_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Buttons", enable, rc, response_size);
        return(1);  // Return error
    }
    return(0);
}


int
IOC_Enable_ScrollWheel(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE)? IOC_CMD_ENABLE_SCROLLWHEEL : IOC_CMD_DISABLE_SCROLLWHEEL );
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_SCROLLWHEEL_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_SCROLLWHEEL_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Scrollwheel", enable, rc, response_size);
        return(1);  // Return error
    }
    return(0);
}

#ifdef CONFIG_FIONA_PM_IOC
/////////////////////////////////////////////
// IOC Echo Routine
/////////////////////////////////////////////

static int
valid_byte_response(char response)
{
    switch(response)  {
        case IOC_SUCCESS_RESPONSE:
        case IOC_FAILURE_RESPONSE:
        case IOC_DISABLED_RESPONSE:
        case IOC_ENABLED_RESPONSE:
        case IOC_BATTERY_FULL_STATUS_RESPONSE:
        case IOC_NO_BATTERY_STATUS_RESPONSE:
            return(1);
        default:
            return(0);
    }
}


int
IOC_Echo(void)
{
    char    cmd = IOC_CMD_ECHO;
    char    response = '\0';
    int     rc = 0;
    int     found_valid = 0;
    int     valid = 0;
    int     tries = 2;

    // Have IOC echo, verify she's there
    rc = IOC_IssueCmd(&cmd, 1, &response, 1);

    if (rc != 1) {
        printk("Trying to fix IOC wake issue (saw no response, rc=%d)...\n",rc);
        return(1);
    }
    else if (IOC_ECHO_RESPONSE != response) {
        printk("Trying to fix IOC wake issue (saw %x response, rc=%d)...\n",response,rc);
        valid = valid_byte_response(response);
        if (valid)
            found_valid = 1;
        while (valid && tries--) {
            valid = 0;
            if (BTLSR & LSR_DR) {
                powdebug(" calling HAL_Receive for next byte...\n");
                rc = globs.hal->IOC_HAL_Receive(globs.hal,&response,1);
                if (rc == 1)
                    valid = valid_byte_response(response);
                if ((rc != 1) || (IOC_ECHO_RESPONSE != response)) {
                    powdebug("ERROR: IOC_Echo got 0x%02x (rc=0x%02x) back when issuing ECHO Cmd to IOC chip !!\n",(int)response,rc);
                    powdebug("BTRBR = 0x%x\n",BTRBR);
                }
            }
            else {
                powdebug("NO MORE DATA IN IOC RX QUEUE !\n");
            }
        }
    }

    if (IOC_ECHO_RESPONSE != response) {
        printk("ERROR: IOC_Echo got 0x%02x (rc=0x%02x) back when issuing ECHO Cmd to IOC chip !!\n",(int)response,rc);
        if (found_valid)  {
            printk("CORRECTING for bug, we saw a valid response, albeit not the right one..\n");
            return(0);
        }
        return(1);
    }
    return(0);
}

int IOC_Enable_Keyboard_Wake(unsigned short enable)
{   
    // FIXME - add some guts here... -njv
    return(0);
}

int IOC_Enable_Scroll_Wake(unsigned short enable)
{
    // FIXME - add some guts here... -njv
    return(0);
}

int
IOC_Enable_PowerSaving(unsigned short enable)
{
    int rc = 0;
    int x;
    int cmd_size;
    char cmd[MAX(IOC_CMD_ENABLE_POWER_SAVINGS_CMD_BYTES,IOC_CMD_DISABLE_POWER_SAVINGS_CMD_BYTES)];

    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_POWER_SAVINGS_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_POWER_SAVINGS_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    if (enable) {
        cmd_size = 1;
        cmd[0] =  IOC_CMD_ENABLE_POWER_SAVINGS;
    }
    else {
        cmd_size = IOC_CMD_DISABLE_POWER_SAVINGS_CMD_BYTES;
        for (x=0; x<cmd_size; x++)
            cmd[x] = IOC_CMD_DISABLE_POWER_SAVINGS;
    }

    // If going to sleep, disable common wake sources
    if (enable)  {
        // First, disable all sleep sources we don't want (keys, buttons, scrollwheel)
        rc = IOC_Enable_Keyboard_Wake(DISABLE);
        rc = IOC_Enable_Scroll_Wake(DISABLE);
    }

    // Issue the command
    rc = IOC_IssueCmd(cmd, cmd_size, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Power Savings Mode", enable, rc, response_size);
        return(1);  // Return error
    }
    return(0);

}
#endif



////////////////////////////////////////////////////////
//
// Query Function(s)
//
////////////////////////////////////////////////////////
static int
IOC_Query_Cmd(int query_cmd, char *response, char parm)
{
    int     cmd_size;
    int     version = 0;
    char    cmd[2] = { PLACE_HOLDER, PLACE_HOLDER };
    int     response_size;
    int     rc = 0;

    if (query_cmd >= 0x100)  {
        printk("ERROR: IOC_Query_Cmd() failed, query cmd 0x%08x > maximum 0xFF\n",query_cmd);
        return(1);  // return error
    }

    // Validate query cmd; setup command, and specify command size
    switch(query_cmd) {

        // 1-byte commands, 1-byte response
        case IOC_QUERY_PNLCD_POWER_STATUS:
        case IOC_QUERY_IOC_ENABLE_STATUS:
        case IOC_QUERY_KEYBOARD_ENABLE_STATUS:
        case IOC_QUERY_BUTTONS_ENABLE_STATUS:   // non-keyboard statu
        case IOC_QUERY_SCROLLWHEEL_ENABLE_STATUS:
        case IOC_QUERY_POWER_SAVING_STATUS:
            cmd_size = 1;
            response_size = 1;
            break;

        // 1-byte commands, 2-byte response
        case IOC_QUERY_FW_VERSION:
        case IOC_QUERY_BATTERY_LEVEL:
        case IOC_QUERY_BATTERY_CAPACITY:
        case IOC_QUERY_BATTERY_TEMPERATURE:
        case IOC_QUERY_CURRENT_DRAW:
            cmd_size = 1;
            response_size = 2;
            break;

        case IOC_QUERY_BATTERY_INFO:
            cmd_size = IOC_QUERY_BATTERY_INFO_CMD_BYTES;
            response_size = IOC_QUERY_BATTERY_INFO_RESPONSE_BYTES;
            break;

        case IOC_QUERY_CHARGE_OR_DISCHARGE:
            cmd_size = 1;
            rc = get_ioc_firmware_version(&version, 0);
            if ((rc != 0) || (version < 20))
                response_size = 1;
            else
                response_size = 2;
            break;
 
        // 2-byte commands, 1-byte response
        case IOC_QUERY_KEY_MASK_STATUS:
        case IOC_QUERY_KEY_STATUS:
            cmd[1] = parm;
            cmd_size = 2;
            response_size = 1;
            break;

        default:
            printk("Unknown IOC Query Cmd 0x%02x\n",query_cmd);
            return(1);  // return error
    }

    // Add in the actual command
    cmd[0] = query_cmd;

    // Send the query
    rc = (globs.hal->IOC_HAL_SendRcv(globs.hal,cmd,cmd_size,response,response_size));
    if ((rc != response_size) && (VERIFY_BAD_RESPONSE(cmd[0],rc,response)))  {
            printk("ERROR: IOC_Query_Cmd() failed for query cmd 0x%02x\n",query_cmd);
            return(1);  // return error
    }
    return(0);
}


static int
VERIFY_BAD_RESPONSE(char cmd, int response_size, char *response)
{
    switch(cmd)  {
        // Handle non-standard responses
        case IOC_QUERY_BATTERY_TEMPERATURE:
        case IOC_QUERY_BATTERY_LEVEL:
        case IOC_QUERY_BATTERY_CAPACITY:
            if ((response_size == 1) && (*response == IOC_NO_BATTERY_RESPONSE))
                return(0);  // This is OK
            break;
        default:
            break;
    }
    return(1);
}

////////////////////////////////////////////////////////
//
// Sleep / Wake Support Functions
//
////////////////////////////////////////////////////////
static int IOC_Two_Key_Wake_Enable(unsigned short enable)
{
    int     rc = 0;
    char    cmd = (enable ? IOC_CMD_ENABLE_TWO_KEY_WAKE : IOC_CMD_DISABLE_TWO_KEY_WAKE);
    int     response_size = (enable ? IOC_CMD_ENABLE_TWO_KEY_WAKE_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_TWO_KEY_WAKE_RESPONSE_BYTES);
    char    response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("Two Key Wake", enable, rc, response_size);
        return(1);  // Return error
    }

    // Remember how we set it...
    g_last_wake_condition = cmd;

    return(0);
}

static int IOC_One_Key_Wake_Enable(unsigned short enable)
{
    int     rc = 0;
    char    cmd = (enable ? IOC_CMD_ENABLE_ONE_KEY_WAKE : IOC_CMD_DISABLE_ALL_KEY_WAKE);
    int     response_size = (enable ? IOC_CMD_ENABLE_ONE_KEY_WAKE_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_ALL_KEY_WAKE_RESPONSE_BYTES);
    char    response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("One Key Wake", enable, rc, response_size);
        return(1);  // Return error
    }

    // Remember how we set it...
    g_last_wake_condition = cmd;

    return(0);
}

int IOC_Reset_Last_Wake_Condition(void)
{
    int rc = 0;
    
    // Reset the wake condition
    switch( g_last_wake_condition ) {
        case IOC_CMD_ENABLE_TWO_KEY_WAKE:
            rc = IOC_Two_Key_Wake_Enable(1);
            break;
        case IOC_CMD_ENABLE_ALL_KEY_WAKE:
            rc = IOC_All_Key_Wake_Enable(1);
            break;
        case IOC_CMD_DISABLE_ALL_KEY_WAKE:
            rc = IOC_All_Key_Wake_Enable(0);
            break;
        default:
            printk("IOC_Reset_Last_Wake_Condition() - unknown last wake reason = %d\n",g_last_wake_condition);
            break;
    }

    // Turn off low-battery warnings...
    IOC_Enable_Battery_Voltage_Low_Reporting_At_IOC(DISABLE, 0);

    return(rc);
}


static int IOC_All_Key_Wake_Enable(unsigned short enable)
{
    int     rc = 0;
    char    cmd = (enable ? IOC_CMD_ENABLE_ALL_KEY_WAKE : IOC_CMD_DISABLE_ALL_KEY_WAKE);
    int     response_size = (enable ? IOC_CMD_ENABLE_ALL_KEY_WAKE_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_ALL_KEY_WAKE_RESPONSE_BYTES);
    char    response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        IOC_Enable_Err("All Key Wake", enable, rc, response_size);
        return(1);  // Return error
    }

    // Remember how we set it...
    g_last_wake_condition = cmd;
    return(0);
}


static int IOC_Get_Wake_Source(int *source)
{
    int     rc = 0;
    char    cmd = IOC_QUERY_IOC_WAKE_SOURCE;
    int     response_size = IOC_QUERY_IOC_WAKE_SOURCE_RESPONSE_BYTES;
    char    response[MAX_BYTES_IOC_RESPONSE+1];
    memclr(response,MAX_BYTES_IOC_RESPONSE);

    // Check return pointer and initialize if valid, return error if not
    if (NULL == source)
        return(1);  // Return error
    *source = WAKE_INVALID;

    // Issue the command
    rc = IOC_IssueCmd(&cmd, 1, response, response_size);
    if (rc != response_size)  {
        printk("IOC: Error querying wake source (rc=%d, response_size=%d).\n",rc,response_size);
        return(1);  // Return error
    }
    // Verify it's the right packet type for response
    if (response[0] != 0xE0) {
        printk("IOC: Wrong packet type 0x%x returned from IOC_WAKE_SOURCE query.\n",response[0]);
        return(1);  // Return error
    }

    // Send back the wake reason
    switch(response[1]) {
        case IOC_WOKEN_BY_AC_ADAPTER:
            *source = AC_PLUG_INSERTED_WAKE;
            break;
        case IOC_WOKEN_BY_LOW_BATTERY:
            *source = LOW_BATTERY_WAKE;
            break;
        case IOC_WOKEN_BY_UART:
            *source = PXA_UART_WAKE;
            break;
        case IOC_WOKEN_BY_KEY_PRESS:
            *source = KEY_PRESS_WAKE;
            break;
        default:
            printk("IOC: Unknown wake type 0x%x returned from IOC_WAKE_SOURCE query.\n",response[1]);
            break;
    }
    return(0);
}


////////////////////////////////////////////////////////
//
// Flash Read/Write Functions
//
////////////////////////////////////////////////////////

static int 
WaitForGoodResponse(void)
{
    int rc = 0;
    char response[2];
    rc = globs.hal->IOC_HAL_Receive(globs.hal,response,1);
    if (rc != 1) {
        printk("Error waiting for response from IOC during flash\n");
        return(1);  // Indicate error
    }
    if (response[0] != 0xFA)  {
        printk("Error 0x%x returned waiting for response from IOC during flash\n",response[0]);
        return(1);
    }
    return(0);
}

static void
SendData( char ch )
{
    int rc = 0;
    rc = globs.hal->IOC_HAL_Send(globs.hal,&ch,1);
    if (rc != 1) printk("Error sending %d to IOC during flash\n",ch);
}

static int
atohex(char ch)
{
    if ((ch >= '0') && (ch <= '9'))
        return(ch-'0');
    else if ((ch >= 'a') && (ch <= 'f'))
        return(ch-'a'+10);
    else if ((ch >= 'A') && (ch <= 'F'))
        return(ch-'A'+10);
    printk("ERROR - non-numeric value '%02x' passed to atohex()\n",ch);
    return(0);
}

#define MAX_DOTS_LIMIT  (520)
#define USE_SEPARATE_BUFFER
static int
IOC_Flash_Write(char *fw_data, int orig_length)
{
    unsigned int    t1 ;
    unsigned int    t2 ;
    int             length = orig_length;
    char            *tbuf = NULL;
    char            *data = NULL;
    unsigned int    addr = 0; 
    unsigned char   byte_count,addr_high,addr_low,next_byte,data_type,csum;
    unsigned int    n_pos=0;
    int             n_position=0;
    int             dummy, cycle;
    int             nD = 0;
    unsigned char   next_addr_high,next_addr_low;
    unsigned int    n_next_addr;
    unsigned int    cur_count = 0;
    unsigned int    line_start = 0;
    unsigned int    line_count = 0;    

    unsigned int    cur_flash_addr;
    int             data_index=0;
    int             version = 0;
    int             rc = 0;
    int             reflash_try_count = MAX_IOC_REFLASH_ATTEMPTS;

    // Allocation temporary local buffer, bail if we can't get it
    tbuf = kmalloc(30000,GFP_KERNEL);
    if (tbuf == NULL) {
        return -ENOMEM;
    }
    data = kmalloc(length,GFP_KERNEL);
    if (data == NULL) {
        kfree(tbuf);
        return -ENOMEM;
    }
#ifdef USE_SEPARATE_BUFFER
    memcpy(data,fw_data,length);
#else
    data = fw_data;
#endif
    
    while( cur_count < length) {
       while( data[cur_count] != ':' ) {
                cur_count++;
       }
       //save the line start count.
       line_start = cur_count;
       cur_count++;
     
        t1 = atohex(data[cur_count++]);
        t2 = atohex( data[cur_count++] );
        byte_count = t1*16+t2;
        //end of data.
        if( byte_count == 0)
        {
          continue;
        }

        t1 = atohex( data[cur_count++] );
        t2 = atohex( data[cur_count++] );
        addr_high = t1*16+t2;
        t1 = atohex( data[cur_count++] );
        t2 = atohex( data[cur_count++] );
        addr_low = t1*16+t2;
        addr =  addr_high*256+addr_low;
        if( addr > 0x2000 )
        {
            break;
        }
        t1 = atohex( data[cur_count++] );
        t2 = atohex( data[cur_count++] );
        data_type = t1*16+t2;
        if( data_type != 0 )
        {
          continue;  //
        }
        t1 = atohex( data[line_start+13+byte_count*2+3]); //
        t2 = atohex( data[line_start+13+byte_count*2+4]);
        next_addr_high = t1*16+t2;
        t1= atohex( data[line_start+13+byte_count*2+5]);
        t2= atohex( data[line_start+13+byte_count*2+6]);
        next_addr_low = t1*16+t2;
        n_next_addr = next_addr_high*256+next_addr_low;

        //the address is continuous
          //to copy  the data sequence to buffer.
          for(cycle=0;cycle<byte_count;cycle++)
          {
              t1 = atohex( data[cur_count++] );
              t2 = atohex( data[cur_count++] );
              next_byte = t1*16+t2;
              tbuf[n_pos++]= next_byte;
             
          }
          
          //whether the address is not continuous ? 
          if( n_next_addr == 0x400e )
              break;
          //caculate the dummy ROM.
          dummy = n_next_addr-addr-byte_count;
          //write the dummy ROM to buffer.
          for(cycle=0;cycle< dummy/2;cycle++)
          {
            tbuf[n_pos++] = 0xFF;
            tbuf[n_pos++] = 0x3F;
          }
                
    }
    //if total length can not divided by 8;
    while( n_pos%8 != 0)
    {
      tbuf[n_pos++] = 0xFF;
      tbuf[n_pos++] = 0x3F;
    }
    //every package the length is 8 byte.
    n_pos /= 0x08;
    csum =0;
    for(cycle =0; cycle< n_pos;cycle++)
    {
       data[cycle*12] = 8;
       data[cycle*12+1] = ((cycle*8)>>8)&0xFF;
       data[cycle*12+2] = (cycle*8)&0xFF;
       csum += 8;
       csum += data[cycle*12+1];
       csum += data[cycle*12+2];
       for(nD=0;nD<8;nD++)
       {
         data[cycle*12+3+nD] = tbuf[n_position++];
         csum += data[cycle*12+3+nD];
       }
       data[cycle*12+11] = csum;
       csum =0;
    }

    // Convert length to line count...
    length = cycle-1;

    //release RAM.
    kfree(tbuf);
    n_pos *= 8;

    // Get the version for a FW bug workaround
    rc = get_ioc_firmware_version(&version, 0);

    // Start up the flash
start_fw_flash:
    SendData( IOC_CMD_PROGRAMMER_START_PROGRAMMING );
    msleep( 1500 );

    /*  IF FW version is older than V24, DON'T wait for an OK response from
     *  IOC in response to the 0x65 (Update FW) command.
     *  This is a bug work-around for versions < V24, which didn't return response
     *  byte from 0x65 command
     */

#define IOC_FW_FLASH_VERIFY_BUG_VERSION 23
    if ((rc != 0) || (version <= IOC_FW_FLASH_VERIFY_BUG_VERSION))  {
        // Wait for the 0xFA to come back.  If we timeout, try again (but not forever)
        if (0 != WaitForGoodResponse())  {
            if (reflash_try_count--)
                goto start_fw_flash;  
            else
                goto complete_flash_exit;
        }
        // Reset flash attempt count for next line
        reflash_try_count = MAX_IOC_REFLASH_ATTEMPTS;
    }

stage_2_fw_flash:
    SendData( IOC_CMD_PROGRAMMER_DATA_LENGTH );
    msleep(1);
    SendData((((n_pos&0xFFFF)>>8)&0xFF));
    msleep(1);
    SendData( ( n_pos&0xFF) );
    msleep(1);
    SendData( (0x60+ ((n_pos&0xFFFF)>>8) + ((n_pos&0xFF)&0xFF)) ); 
    if (0 != WaitForGoodResponse())  {
        if (reflash_try_count--)
            goto stage_2_fw_flash;  
        else 
            goto complete_flash_exit;
    }
    // Reset flash attempt count for next line
    reflash_try_count = MAX_IOC_REFLASH_ATTEMPTS;

    // Send off the data
    while (  line_count <= length) {
        if( data[line_count*12+1] > 0x1F ) {
            printk("IOC Flash write complete (data > 0x1F).\n");
            goto complete_flash_exit;
        }

flash_line:
        SendData( 0x61 );
        SendData( (data[line_count*12+0]&0xFF) ); //byte count
        SendData( (data[line_count*12+1]&0xFF) );//High byte address
        SendData( (data[line_count*12+2]&0xFF) );//low byte address

        data_index = 0;
        for (;data_index< (data[line_count*12+0]&0xFF)/2;data_index++ )
        {
            SendData( data[3+data_index*2+line_count*12]&0xFF );
            SendData( data[data_index*2+4+line_count*12]&0xFF );
        }
        SendData( (data[line_count*12+data_index*2+3]&0xFF) ); //check sum.
        msleep(5);

#if 1
        // Check response, resend if needed
        if (0 != WaitForGoodResponse())  {
            if (reflash_try_count--)
                goto flash_line;
            else
                goto complete_flash_exit;
        }

        // Reset flash attempt count for next line
        reflash_try_count = MAX_IOC_REFLASH_ATTEMPTS;
#endif

        cur_flash_addr = (data[line_count*12+1]&0xFF)*256;
        cur_flash_addr |= (data[line_count*12+2]&0xFF);

        // Give progress status feedback
        if (cur_flash_addr && (cur_flash_addr % MAX_DOTS_LIMIT) == 0)
            printk("\n");
        else if (cur_flash_addr)
            printk(".");

        line_count++;
    }
    printk("\n");

complete_flash_exit:
    msleep(100);
    SendData( 0x62 );
    SendData( 0x00 );
#ifdef USE_SEPARATE_BUFFER
    kfree(data);
#endif
    return(orig_length);
}



////////////////////////////////////////////////////////
//
// Segment On/Off Functions
//
////////////////////////////////////////////////////////

static char pnlcd_segments0[PN_LCD_SEGMENT_COUNT]; // Saved PNLCD state.
static char pnlcd_segments1[PN_LCD_SEGMENT_COUNT]; // Working copy.

static char *pnlcd_segments = pnlcd_segments0;     // Point to saved state.
static int  force_issue_cmd = 0;                   // Set to "flush" cache.

static int
IOC_IssuePNLCDCmd(char *cmd, int cmdSize, char *response, int responseSize)
{
    int  i, issue_cmd = force_issue_cmd, skip = 1, result = responseSize;
    char startSeg, endSeg, segment = SEGMENT_ON;
    
    // Make a copy of the current PNLCD segment state.
    //
    memcpy(pnlcd_segments1, pnlcd_segments0, PN_LCD_SEGMENT_COUNT);
    
    // Update the copy of the PNLCD segment state.
    //
    switch ( cmd[0] )
    {
        // Supported commands.
        //
        case IOC_CMD_PNLCD_VERTICAL_SEQUENTIAL_OFF:
            segment = SEGMENT_OFF;
        case IOC_CMD_PNLCD_VERTICAL_SEQUENTIAL_ON:
            skip = 2;
        goto sequential_common;
       
        case IOC_CMD_PNLCD_SEQUENTIAL_OFF:
            segment = SEGMENT_OFF;
        case IOC_CMD_PNLCD_SEQUENTIAL_ON:
        sequential_common:      
            startSeg = cmd[1];
            endSeg   = cmd[2] + 1;

            for (i = startSeg; i < endSeg; i += skip)
                pnlcd_segments1[i] = segment;
        break;

        case IOC_CMD_PNLCD_MISC_OFF:
            segment = SEGMENT_OFF;
        case IOC_CMD_PNLCD_MISC_ON:
            startSeg = 2;
            endSeg   = cmd[1] + 2;

            for (i = startSeg; i < endSeg; i++)
                pnlcd_segments1[(int)cmd[i]] = segment;
        break;

        // Not-yet supported commands (and not used by framework).
        //
        case IOC_CMD_PNLCD_MISC_VERTICAL_PAIR_OFF:
        case IOC_CMD_PNLCD_MISC_VERTICAL_PAIR_ON:
        case IOC_CMD_PNLCD_MISC_PAIR_OFF:
        case IOC_CMD_PNLCD_MISC_PAIR_ON:
        default:
            issue_cmd = 1;
        break;
    }
    
    // If the PNLCD state changed, then update the PNLCD segment state and
    // issue the appropriate command.
    //
    if ( issue_cmd || (0 != memcmp(pnlcd_segments1, pnlcd_segments0, PN_LCD_SEGMENT_COUNT)) )
    {
        memcpy(pnlcd_segments0, pnlcd_segments1, PN_LCD_SEGMENT_COUNT);
        
        // Only issue the command if we haven't been delayed.
        //
        if ( !pnlcd_delayed() )
            result = IOC_IssueCmd(cmd, cmdSize, response, responseSize);
    }

    return ( result );
}

static int
IOC_Sequential_Segment(unsigned short on, char startSeg, char endSeg, int vertical)
{
    int     err = 0;
    char    the_cmd[3];
    char    result[IOC_CMD_PNLCD_LIGHT_RESPONSE_BYTES];

    NDEBUG("@@@@ IOC_Sequential_Segment(on=%02x, start=%02x, end=%02x)\n",on,startSeg,endSeg);

    // Range-check values to assure segment < 200
    if (startSeg > MAX_PN_LCD_SEGMENT) {
            NDEBUG("@@@@ IOC_Sequential_Segment(start=%02x out of range)\n",startSeg);
            err = startSeg;
            goto exit;
    }
    else if (endSeg > MAX_PN_LCD_SEGMENT)  {
            NDEBUG("@@@@ IOC_Sequential_Segment(end=%02x out of range)\n",endSeg);
            err = endSeg;
            goto exit;
    }

    // Turn on sequential segments
    if (vertical)
        the_cmd[0] = (on ? IOC_CMD_PNLCD_VERTICAL_SEQUENTIAL_ON : IOC_CMD_PNLCD_VERTICAL_SEQUENTIAL_OFF);
    else
        the_cmd[0] = (on ? IOC_CMD_PNLCD_SEQUENTIAL_ON : IOC_CMD_PNLCD_SEQUENTIAL_OFF);
    the_cmd[1] = startSeg;         // The starting segment number to turn on
    the_cmd[2] = endSeg;           // The ending segment number to turn on
    err = IOC_IssuePNLCDCmd(the_cmd,3,result,IOC_CMD_PNLCD_LIGHT_RESPONSE_BYTES);
    if (err == 1)
        err = 0;  // no errors...
exit:
    return(err);  // Return Error - not implemented yet
}


static int
IOC_Multi_Segment(unsigned short on, char num_segs, char *segment_array)
{
    int     sample_count = 0;
    int     x = 0;
    int     err = 0;
    char    the_cmd[256];
    char    *cur_byte = the_cmd;
    char    result = 0;

    NDEBUG("@@@@ IOC_Multi_Segment(on=%02x, numSegs=%02x, array=%08x)\n",on,num_segs,(int)segment_array);

    sample_count = num_segs;     // Count is given in pairs

    // Make sure we don't overflow firmware buffers
    if (num_segs > (MAX_COMMAND_BYTES-3))  {
        err = num_segs;
        goto exit;
    }
    for (x=0;x<num_segs;x++) {
        if (segment_array[x] > MAX_PN_LCD_SEGMENT) {
            err = segment_array[x];
            goto exit;
        }
    }

    // Turn on multiple segments
    *cur_byte++ = (on ? IOC_CMD_PNLCD_MISC_ON : IOC_CMD_PNLCD_MISC_OFF);
    *cur_byte++ = num_segs;         // The number of segments to turn on or off
    for (x=0;x<num_segs;x++)
        *cur_byte++ = segment_array[x];  // The next segment number to turn on
    err = IOC_IssuePNLCDCmd(the_cmd,(sample_count+2),&result,IOC_CMD_PNLCD_LIGHT_RESPONSE_BYTES);
    if (err == 1)
        err = 0;  // no errors...

exit:
    return(err);  
}

static int
IOC_Multi_Segment_Pairs(unsigned short on, char num_pairs, char *paired_segment_array, int vertical)
{
    int     sample_count = 0;
    int     x = 0;
    int     err = 0;
    char    the_cmd[256];
    char    *cur_byte = the_cmd;
    char    result = 0;

    NDEBUG("@@@@ IOC_Multi_Segment_Pairs(on=%02x, numPairs=%02x, array=%08x)",on,num_pairs,(int)paired_segment_array);

    sample_count = (num_pairs*2);     // Count is given in pairs

    // Range-check values to assure segment < 200
    if (sample_count > (MAX_COMMAND_BYTES-3))  {
        err = sample_count;
        goto exit;
    }
    for (x=0;x<sample_count;x++) {
        if (paired_segment_array[x] > MAX_PN_LCD_SEGMENT) {
            err = paired_segment_array[x];
            goto exit;
        }
    }

    // Build command to turn on/off multiple segments
    if (vertical)
        *cur_byte++ = (on ? IOC_CMD_PNLCD_MISC_VERTICAL_PAIR_ON : IOC_CMD_PNLCD_MISC_VERTICAL_PAIR_OFF);
    else
        *cur_byte++ = (on ? IOC_CMD_PNLCD_MISC_PAIR_ON : IOC_CMD_PNLCD_MISC_PAIR_OFF);
    *cur_byte++ = sample_count;        // The number of segments to turn on or off
    for (x=0;x<sample_count;x++)
        *cur_byte++ = paired_segment_array[x];  // The next segment number to turn on

    err = IOC_IssuePNLCDCmd(the_cmd,(sample_count+2),&result,IOC_CMD_PNLCD_LIGHT_RESPONSE_BYTES);
    if (err == 1)
        err = 0;  // no errors...

exit:
    return(err);  
}

static void set_which_pnlcd_segments(char which, char *segments)
{
    char start, count;
    int i;
    
    // First, do all the segments, if any, that are transitioning.
    //
    for (i = start = which, count = 0; i < (PN_LCD_SEGMENT_COUNT - 2); i += 2)
    {
        if ( segments[i] != segments[i+2] )
        {
            IOC_Sequential_Segment((unsigned short)segments[(int)start], start, (start + (count * 2)), VERT_SEQ);
            start = i+2; count = 0;
        }
        else
            count++;
    }
    
    // Now, do any "left overs."
    //
    IOC_Sequential_Segment((unsigned short)segments[(int)start], start, (start + (count * 2)), VERT_SEQ);
}

void set_pnlcd_segments(char *segments)
{
    set_which_pnlcd_segments(0, segments); // Evens (right column).
    set_which_pnlcd_segments(1, segments); // Odds  (left  column).
}

void restore_pnlcd_segments(void)
{
    force_issue_cmd = 1; set_pnlcd_segments(pnlcd_segments);
    force_issue_cmd = 0;
}

char *get_pnlcd_segments(void)
{
    return(pnlcd_segments);
}

int
IOC_Enable_PNLCD(unsigned short enable)
{
    int rc = 0;
    char cmd = ((enable==ENABLE) ?  IOC_CMD_ENABLE_PNLCD :  IOC_CMD_DISABLE_PNLCD );
    int response_size = (enable==ENABLE) ?  IOC_CMD_ENABLE_PNLCD_RESPONSE_BYTES : \
                                            IOC_CMD_DISABLE_PNLCD_RESPONSE_BYTES;
    char response[MAX_BYTES_IOC_RESPONSE+1];
    
    // Just clear the PNLCD on DISABLE to eliminate the "sparkle" problem.
    if (enable!=ENABLE) {
        IOC_Sequential_Segment(SEGMENT_OFF, FIRST_PNLCD_SEGMENT, LAST_PNLCD_SEGMENT, SEQ);
    } else {
        memclr(response,MAX_BYTES_IOC_RESPONSE);
        rc = IOC_IssueCmd(&cmd, 1, response, response_size);
        if (rc != response_size)  {
            IOC_Enable_Err("PNLCD", enable, rc, response_size);
            return(1);  // Return error
        }
    }
    return(0);
}


/////////////////////////////////////////////
// Receive Engine
/////////////////////////////////////////////
/*
Packet Header   Packet Size     Purpose
    0x00        1               Disabled Response
    0x01        1               Enabled Response
    0xFA        1               Success / Complete Response
    0xFB        1               Failure / Complete Response
    0xFE        1               Echo Response
    0xE0        2               Response Packet
    0xE1        2               Battery Discharge Current
    0xE2        2               Battery Charge Current
    0xE3        2               IOC Crush during programming - must reprogram
    0xF0        2               Keyboard / scroll wheel information
    0xF1        2               Battery Capacity Information
    0xF2        2               Battery Voltage Low Warning
*/

static int
ioc_process_rx_packet(struct fiona_struct *k)
{
    int err = 0;

    switch(k->packet[0]) {

    case IOC_DISABLED_RESPONSE:             // Disabled
    case IOC_ENABLED_RESPONSE:              // Enabled
    case IOC_SUCCESS_RESPONSE:              // Success
    case IOC_FAILURE_RESPONSE:              // Failure
    case IOC_ECHO_RESPONSE:                 // Echo
    case IOC_BATTERY_FULL_STATUS_RESPONSE:  // Full Battery
    case IOC_NO_BATTERY_STATUS_RESPONSE:    // No Battery
    case IOC_NO_BATTERY_RESPONSE:           // Generic No Battery Response
    case IOC_TWO_BYTE_RESPONSE:             // Generic 2-byte Response Packet
    case IOC_FOUR_BYTE_RESPONSE:            // Generic 4-byte Response Packet
        ioc_response_seen(k->packet);
        err = ioc_put_packet_irq(k);
        break;

    case IOC_BATTERY_DISCHARGE_RESPONSE:// Battery Discharge Current
    case IOC_BATTERY_CHARGE_RESPONSE:   // Battery Charge Current
        ioc_response_seen(k->packet);
        update_battery_info(k, &g_battery_info);
        err = ioc_put_packet_irq(k);
        break;

    case IOC_BATTERY_LOW_RESPONSE:      // Battery Low Voltage Warning
        ioc_response_seen(k->packet);
        update_battery_info(k, &g_battery_info);

        printk("IOC returned BATTERY_LOW_RESPONSE !!!\n");

       // handle_fiona_event(BATTERY_LOW_EVENTCODE);
        PDEBUG("IOC LOW BATTERY condition seen.\n");
        break;

    case IOC_AC_IN_RESPONSE:                // AC Adapter plugged in
        ioc_response_seen(k->packet);
        update_ac_adapter_status(1);
        handle_fiona_event(AC_IN_EVENTCODE);
        PDEBUG("IOC AC Adapter inserted.\n");
        break;

    case IOC_AC_OUT_RESPONSE:               // AC Adapter removed
        ioc_response_seen(k->packet);
        update_ac_adapter_status(0);

        // Clear the voltage averaging table because removing the AC
        // can and usually will cause a large swing in voltage
        // (in the negative direction).  By clearing the table, we reduce
        // propogation delay effect of the table to a minimum.
        ioc_reset_voltage_averages();

        handle_fiona_event(AC_OUT_EVENTCODE);
        PDEBUG("IOC AC Adapter removed.\n");
        break;

    case IOC_BATTERY_CAPACITY_RESPONSE: // Battery Capacity Data
        ioc_response_seen(k->packet);
        update_battery_info(k, &g_battery_info);
        //handle_critical_battery_level(k->packet);
        PDEBUG("IOC Packet type IOC_BATTERY_CAPACITY_RESPONSE seen.\n",k->packet[0]);
        break;

    case IOC_CRUSH_RESPONSE:      // IOC Firmware Flash Crush Condition Seen
        err = ioc_put_packet_irq(k);
        PDEBUG("IOC CRUSH condition seen.\n");
        break;

    case IOC_KEY_SCROLL_RESPONSE:      // Keyboard / Scroll Wheel Information
        PDEBUG("IOC Keyboard packet seen %x\n",k->packet[1]);
        ioc_response_seen(k->packet);
        handle_keyboard_event(k->packet[1]);
        break;


    default:
        printk("Unknown IOC Packet type 0x%02x\n",k->packet[0]);
        err = 1;
        break;
    }
    return(err);
}


void fiona_receive(struct fiona_struct *k)
{
    k->packet[k->pos++] = BTRBR;


#ifdef DEBUG_IOC_COMMAND_RECEIVE
    printk("IOC rx byte received : 0x%02x\n",(int)k->packet[(k->pos-1)]);
#endif

#ifdef DEBUG_WAKE_BUG
    if (BTLSR & LSR_DR)
        printk("*** IOC RX DATA STILL AVAILABLE...\n");
#endif

    if (k->pos == 1) {
        switch(k->packet[0]) {

        // Four Byte Response Packet
        case IOC_FOUR_BYTE_RESPONSE:            // 4 byte packet
            k->length = 5;
            break;

        // Single Byte Response Packets
        case IOC_DISABLED_RESPONSE:             // Disabled
        case IOC_ENABLED_RESPONSE:              // Enabled
        case IOC_SUCCESS_RESPONSE:              // Success
        case IOC_FAILURE_RESPONSE:              // Failure
        case IOC_ECHO_RESPONSE:                 // Echo
        case IOC_BATTERY_FULL_STATUS_RESPONSE:  // Full Battery
        case IOC_NO_BATTERY_STATUS_RESPONSE:    // No Battery
        case IOC_NO_BATTERY_RESPONSE:           // Generic No Battery Response
        case IOC_AC_IN_RESPONSE:                // AC Adapter plugged in
        case IOC_AC_OUT_RESPONSE:               // AC Adapter removed
            k->length = 1;
            break;

        // Two Byte Response Packets
        case IOC_TWO_BYTE_RESPONSE:             // Generic 2-byte Response Packet
        case IOC_BATTERY_DISCHARGE_RESPONSE:    // Battery Discharge Current
        case IOC_BATTERY_CHARGE_RESPONSE:       // Battery Charge Current

        // Two Byte Unsolicited Response Packets
        case IOC_CRUSH_RESPONSE:                // IOC Firmware Flash Crush Seen
        case IOC_KEY_SCROLL_RESPONSE:           // Keyboard/Scrollwheel Info
        case IOC_BATTERY_LOW_RESPONSE:          // Battery Low Voltage Warning
        case IOC_BATTERY_CAPACITY_RESPONSE:     // Battery Capacity Data
            k->length = 2;
            break;

        default:
            /* Don't know this packet type
             * Set to single byte, let next case statement
             * report unknown and drop it on the floor
             */
            k->length = 1;
            break;
        }
    }

    if (k->pos == k->length) {
        dprint_fiona_packet(k,"IOC RX packet      ");

        // Transfer the packet to the proper queue
        ioc_process_rx_packet(k);
        k->pos = k->length = 0;

        tasklet_schedule(&ioc_hal_tasklet);
    }
}


static void ioc_response_seen(char *packet)
{
    // Don't do anything right now;
    // Later, we transition rx engine states here
}

#endif  // CONFIG_IOC
