/******************************************************************
 *
 *	File:	ppoke.c
 *
 *	Author:	Nick Vaccaro
 *
 *	Description:
 *		Driver for accessing system memory space (Peek and Poke)
 ******************************************************************/
/*************************************
 *  Flags for running test code
 *  undef to disable, define to enable 
 *************************************/
#include <linux/module.h>
#include <linux/version.h>
#include <linux/linkage.h>
#include <asm/uaccess.h>

char kernel_version [] = UTS_RELEASE;

/*************************************
 *  Current revision of driver
 *************************************/
#define kDriverRevision	4

#include <linux/kernel.h>		/* printk() */
#include <linux/fs.h>			/* everything... */
#include <linux/errno.h>		/* error codes */
#include <asm/io.h>			    /* ioremap */
#include <linux/ioport.h>		/* request_region and port accessors */
#include <linux/init.h>			/* __init */
#include <linux/proc_fs.h>

/* Constants for PXA board */
#include <asm/hardware.h>
#include "ppoke.h"

#ifdef CONFIG_PEEK_POKE
static struct proc_dir_entry    *ppoke_dir;
#define PEEK_PROC_ENTRY_NAME     "peek"
#define POKE_PROC_ENTRY_NAME     "poke"
#define PEEK_POKE_PROC_DIR_NAME  "ppoke"
#endif

// Define PPOKE_DEBUG to enable debugging
#undef PPOKE_DEBUG	// turn on debugging

#undef PDEBUG
#ifdef PPOKE_DEBUG 
#  ifdef __KERNEL__
#    define PDEBUG(fmt,args...) printk( KERN_CRIT"ppoke: " fmt, ## args)
#  else
#    define PDEBUG(fmt,args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt,args...)
#endif

#undef PDEBUGG
#define PDEBUGG(fmt,args...) /* do nothing */

int ppoke_open(struct inode* inNode, struct file* inFile);
int ppoke_close(struct inode* inNode, struct file* inFile);
void ppoke_release(void);
int ppoke_ioctl(struct inode *inNode, 
				struct file *inFile, 
				unsigned int inCommand, 
				unsigned long inArg);


/* ********** Private Constants  */

#define kPeekPokeMajorNumber	207
#define kMinorNumber		1


/* ********** File Operations struct  */

//        write:          ppoke_write,
static struct file_operations ppoke_fops = {
        owner:          THIS_MODULE,
        read:           NULL,
        write:          NULL,
        ioctl:          ppoke_ioctl,
        open:           ppoke_open,
        release:        ppoke_close,
};
                                                                                                           

typedef struct DrvrGlobalsRec {
	int	majorNumber;			// ppoke Major Number
} DrvrGlobalsRec, *DrvrGlobalsPtr;


/* Globals */
DrvrGlobalsRec gDrvrGlobals;


/*************************************************
 * Local Prototypes
 ************************************************/
static int SystemWriteByte(unsigned long addr, unsigned char value);
static int SystemWriteShort(unsigned long addr, unsigned short value);
static int SystemWriteLong(unsigned long addr, unsigned long value);

static int SystemReadByte(unsigned long addr, unsigned char *value);
static int SystemReadShort(unsigned long addr, unsigned short *value);
static int SystemReadLong(unsigned long addr, unsigned long *value);

/***************/

static int SystemWriteByte(unsigned long addr, unsigned char value)
{
	* (volatile unsigned char*) addr = value;
	return(NOERR);
}

static int SystemWriteShort(unsigned long addr, unsigned short value)
{
	* (volatile unsigned short *) addr = value;
	return (NOERR);
}

static int SystemWriteLong(unsigned long addr, unsigned long value)
{
	* (volatile unsigned long *) addr = value;
	return (NOERR);
}


static int SystemReadByte(unsigned long addr, unsigned char *value)
{
	*value = * (volatile unsigned char *) addr;
	return (NOERR);
}

static int SystemReadShort(unsigned long addr, unsigned short *value)
{
	*value = * (volatile unsigned short *) addr;
	return (NOERR);
}

static int SystemReadLong(unsigned long addr, unsigned long *value)
{
	*value = *(volatile unsigned long *) addr;
	PDEBUGG("SystemReadLong: value read at 0x%x was %x\n",(int) addr,(int) *value);
	return (NOERR);
}


/***************/

int ppoke_ioctl(struct inode *inNode, 
				struct file *inFile, 
				unsigned int inCommand, 
				unsigned long inArg)
{
	int err = NOERR;
	PPokeRecPtr recPtr = (PPokeRec*)inArg;
	PPokeRec theRec;
	unsigned long *virtualP = NULL;
	unsigned char byteVal = 0;
	unsigned short shortVal = 0;
	unsigned long longVal = 0;

	PDEBUG("Module - ppoke_ioctl called. inCommand = %x\n", inCommand);
	
	if (_IOC_TYPE(inCommand) != kIOCTLPPOKEMagicNumber) return -EINVAL;
	if (_IOC_NR(inCommand) > PPOKE_IOCTLLast) return -EINVAL;
	
	// Get user data
	if (copy_from_user(&theRec,(void *)inArg, sizeof(theRec)))
		return(-EFAULT);
	
	if (inCommand == PPOKE_IOCTLPOKE)
		PDEBUG("ppoke_ioctl - PEEK ");
	else if (inCommand == PPOKE_IOCTLPEEK)
		PDEBUG("ppoke_ioctl - POKE ");
	PDEBUG("ppoke_ioctl - addr=0x%x, value=0x%x, size=0x%x }\n", (unsigned int) theRec.addr, (unsigned int)theRec.value, (unsigned int) theRec.size);

	// Remap the IO space
	virtualP = ioremap((int)theRec.addr,0x100);
	if (virtualP == NULL)
	{
		printk("ppoke_ioctl: ioremap failed for 0x%x\n",(unsigned int)theRec.addr);
		return(0);
	}

	/* Add verification of user space before accessing inArg pointer - FIXME */
	
	PDEBUG("ppoke_ioctl - ioremap( 0x%x -> 0x%x )\n", (unsigned int) theRec.addr,(unsigned int) virtualP);
	
	if (inCommand == PPOKE_IOCTLPOKE) {
		switch(theRec.size) {
			case kByteSize:
				err = SystemWriteByte((int) virtualP, theRec.value);
				break;			
			case kShortSize:
				err = SystemWriteShort((int) virtualP, theRec.value);
				break;
			case kLongSize:
				err = SystemWriteLong((int) virtualP, theRec.value);
				break;
			default:
				err = -EINVAL;
				PDEBUG("ppoke_ioctl - Bad size for poke (%d)\n",theRec.size);
				break;
		}
	}
	else if (inCommand == PPOKE_IOCTLPEEK) {
		switch(theRec.size) {
			case kByteSize:
				err = SystemReadByte((int) virtualP, (unsigned char *) &byteVal);
				if (!err)
					put_user(byteVal, (unsigned char *) &recPtr->value);
				break;			
			case kShortSize:
				err = SystemReadShort((int) virtualP, (unsigned short *) &shortVal);
				if (!err)
					put_user(shortVal, (unsigned short *) &recPtr->value);
				break;
			case kLongSize:
				err = SystemReadLong((int) virtualP, &longVal);
				if (!err)
					put_user(longVal, (unsigned long *) &recPtr->value);
				break;
			default:
				err = -EINVAL;
				PDEBUG("ppoke_ioctl - unknown peek size (%d)\n", theRec.size);
				break;
		}
	}
	else {
		err = -EINVAL;
		PDEBUG("ppoke_ioctl - unknown inCommand = 0x%x\n", inCommand);
	}
	
	// Release the mapping
	iounmap(virtualP);
	return err;
}

/***************/
#if 0
static ssize_t ppoke_write(struct file *fp, const char *buf, size_t count, loff_t *ppos)
{
	volatile unsigned long *virtualP= 0;
	volatile unsigned long *userP= 0;
	unsigned long *userDataPtr = (unsigned long*) buf;
	unsigned long userData = 0;

    return(count);

	// Parse out address
	if (buf == NULL)
		return(0);	// bad doggie

	userP =(volatile unsigned long*) (*userDataPtr++);

	userData = (*userDataPtr++);
	virtualP = ioremap((int)userP,0x100);
	if (virtualP == NULL)
	{
		printk("ppoke_write: ioremap failed for 0x%x\n",(int)userP);
		return(0);
	}

	*virtualP = userData;

	iounmap((void*)virtualP);
	return(count);

}
#endif

/***************/

int ppoke_open(struct inode* inNode, struct file* inFile)
{
        return 0;
}
                                                                                                          
int ppoke_close(struct inode* inNode, struct file* inFile)
{
#ifdef MODULE
	PDEBUG("PPOKE - module closed.\n");
#else
	PDEBUG("PPOKE - driver closed.\n");
#endif

	return 0;
}

void do_peek(int addr, int *value, int size)
{
	unsigned long *virtualP = NULL;
    char            byteVal = 0;
    unsigned short  shortVal = 0;
    unsigned long   longVal = 0;
    int err = 0;


	// Remap the IO space
	virtualP = ioremap(addr,0x100);
	if (virtualP == NULL)
	{
		printk("do_peek(): ioremap failed for 0x%x\n",addr);
		return;
	}

	switch(size) {
		case 1:
		    err = SystemReadByte((int) virtualP, (unsigned char *) &byteVal);
            *value = (unsigned long) byteVal;
		    break;			
		case 2:
		    err = SystemReadShort((int) virtualP, (unsigned short *) &shortVal);
            *value = (unsigned long) shortVal;
		    break;
		case 4:
		    err = SystemReadLong((int) virtualP, &longVal);
            *value = (unsigned long) longVal;
		    break;
		default:
		    err = -EINVAL;
		    PDEBUG("do_peek() - unknown peek size (%d)\n", size);
		    break;
	}
	
	// Release the mapping
	iounmap(virtualP);
	return;
}
void do_poke(int addr, int value, int size)
{
	unsigned long *virtualP = NULL;
    int err = 0;

	// Remap the IO space
	virtualP = ioremap(addr,0x100);
	if (virtualP == NULL)
	{
		printk("do_poke(): ioremap failed for 0x%x\n",addr);
		return;
	}

    switch(size) {
        case 1:
            err = SystemWriteByte((int) virtualP, value);
            break;			
        case 2:
            err = SystemWriteShort((int) virtualP, value);
            break;
        case 4:
            err = SystemWriteLong((int) virtualP, value);
            break;
        default:
            err = -EINVAL;
            PDEBUG("do_poke() - Bad size for poke (%d)\n",size);
            break;
        }
	
	// Release the mapping
	iounmap(virtualP);
	return;
}


#ifdef CONFIG_PEEK_POKE

/******************************************************************************
 *
 * /proc fs support
 *
 ******************************************************************************/

static int
parse_proc_peek_write_format(const char *cmd, int *addr, int *count)
{
    int     rc = 0;
    char    size;

    rc = sscanf(cmd,"%x:%c\n",addr, &size);
    switch (size)  {
        case 'b':
        case 'B':
            *count = 1;
            break;
        case 'w':
        case 'W':
        case 's':
        case 'S':
            *count = 2;
            break;
        case 'l':
        case 'L':
        default:
            *count = 4;
            break;
    }
    return(0);  // No error
}

static int
parse_proc_poke_write_format(const char *cmd, int *addr, int *data_to_write, int *count)
{
    int     rc = 0;
    char    size;
    int     bytes_to_write = 3;

    //printk("Parsing command = %s\n",cmd);
    rc = sscanf(cmd,"%x:%x:%c\n",addr, data_to_write, &size);
    if (bytes_to_write != 3)  {
            printk("Improper format.  Must be hexAddr:value:sizeOfValue.\n" \
                    "Example:  echo '0x48000000:0xdeadbeef:4' >> /proc/poke\n");
            return(1);
    }
    switch (size)  {
        case 'b':
        case 'B':
            *count = 1;
            break;
        case 'w':
        case 'W':
        case 's':
        case 'S':
            *count = 2;
            break;
        case 'l':
        case 'L':
            *count = 4;
            break;
        default:
            printk("Invalid size field.  Defaulting to long value.\n");
            *count = 4;
            break;
    }
    return(0);  // No error
}


static int proc_write_poke(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char    cmd[256];
    int     rc = 0;
    int     size = 0;
    int     value = 0;
    int     addr = 0;

    if (count > 256) {
        return -EINVAL;
    }

    // Get user data
    if (copy_from_user(cmd, buffer, count)) {
        return -EFAULT;
    }

    // Parse the user format
    rc = parse_proc_poke_write_format(cmd, &addr, &value, &size);
    if (rc != 0)
        return -EFAULT;

    // Poke the address with the value
    printk("Poking %d bytes at address 0x%x with 0x%x...\n",size,addr,value);
    do_poke(addr,value,size);
    return count;
}

static int proc_write_peek(struct file *file,
                 const char *buffer,
                 unsigned long count,
                 void *data)
{
    char    cmd[256];
    int     rc = 0;
    int     size = 0;
    int     addr = 0;
    int     value = 0;

    if (count > 256) {
        return -EINVAL;
    }

    // Get user data
    if (copy_from_user(cmd, buffer, count)) {
        return -EFAULT;
    }

    // Parse the user format
    rc = parse_proc_peek_write_format(cmd, &addr, &size);
    if (rc != 0)
        return -EFAULT;
    do_peek(addr,&value,size);
    switch(size)  {
        case 1:
            printk("Peek 0x%08x = 0x%02x\n",addr,value);
            break;
        case 2:
            printk("Peek 0x%08x = 0x%04x\n",addr,value);
            break;
        case 4:
            printk("Peek 0x%08x = 0x%08x\n",addr,value);
            break;
        default:
            break;
    }
    return count;
}

/******************************************************************************/

static int peek_poke_proc_init(void)
{
    int rc = 0;

    struct proc_dir_entry *peek_file;
    struct proc_dir_entry *poke_file;

#ifdef CREATE_PPOKE_PROC_DIR
    // Create /proc/peek and /proc/poke
    //
    ppoke_dir = proc_mkdir(PEEK_POKE_PROC_DIR_NAME, NULL);
    if (NULL == ppoke_dir ) {
        printk("Out of mem setting up Peek and Poke proc interface...\n");
        rc = -ENOMEM;
        goto exit;
    }

    // Set the owner
    ppoke_dir->owner = THIS_MODULE;
#else
    ppoke_dir = NULL;
#endif

    // create peek file entry functions
    peek_file = create_proc_entry(PEEK_PROC_ENTRY_NAME, 0644, ppoke_dir);
    if (!peek_file) {
        rc = -ENOMEM;
        if (ppoke_dir != NULL)
            remove_proc_entry(PEEK_POKE_PROC_DIR_NAME, NULL);
        goto exit;
    }

    peek_file->owner = THIS_MODULE;
    peek_file->data  = NULL;
    peek_file->read_proc = NULL; // proc_read_peek;
    peek_file->write_proc    = proc_write_peek;

    /* All's good with peek */
    if (ppoke_dir != NULL)
        printk("/proc/ppoke/peek initialized.\n");
    else
        printk("/proc/peek initialized.\n");

    // create peek file entry functions
    poke_file = create_proc_entry(POKE_PROC_ENTRY_NAME, 0644, ppoke_dir);
    if (!poke_file) {
        rc = -ENOMEM;
        remove_proc_entry(PEEK_POKE_PROC_DIR_NAME, NULL);
        goto exit;
    }

    poke_file->owner = THIS_MODULE;
    poke_file->data  = NULL;
    poke_file->read_proc = NULL; // proc_read_poke;
    poke_file->write_proc    = proc_write_poke;

    /* All's good with poke */
    if (ppoke_dir != NULL)
        printk("/proc/ppoke/poke initialized.\n");
    else
        printk("/proc/poke initialized.\n");
    return 0;

exit:
    return(rc);
}

static void ppoke_cleanup_procfs(void)
{
    remove_proc_entry(PEEK_PROC_ENTRY_NAME, ppoke_dir);
    remove_proc_entry(POKE_PROC_ENTRY_NAME, ppoke_dir);
    if (ppoke_dir != NULL)
        remove_proc_entry(PEEK_POKE_PROC_DIR_NAME, NULL);

    printk("/proc file entries for peek and poke removed.\n");
}
#endif


/********** Init and Cleanup Functions *********/

int __init ppoke_init(void)
{
	short   driverRev = (short) kDriverRevision;


	// Register this driver module.
	gDrvrGlobals.majorNumber = register_chrdev(kPeekPokeMajorNumber, "ppoke", &ppoke_fops);
	if (gDrvrGlobals.majorNumber < 0) {
		PDEBUG("Module - Error: Major number was not assigned!");
		return gDrvrGlobals.majorNumber;
	}
		
        if (gDrvrGlobals.majorNumber == 0)
                gDrvrGlobals.majorNumber = kPeekPokeMajorNumber;


        printk("ppoke driver: Major: %d  Rev. %d  Kernel v.%s\n",gDrvrGlobals.majorNumber, driverRev, kernel_version);

#ifdef CONFIG_PEEK_POKE
    // Initialize the /proc interface
    peek_poke_proc_init();
#endif

	return 0;
}



void ppoke_release(void)
{
	PDEBUG("in ppoke_release, unregistering char driver.\n");

	// Unregister this driver module.
	unregister_chrdev(gDrvrGlobals.majorNumber, "ppoke");

#ifdef CONFIG_PEEK_POKE
    // Destroy the /proc interface
    ppoke_cleanup_procfs();
#endif
}
                                                                                                          
/***************/
module_init(ppoke_init);
module_exit(ppoke_release);

MODULE_AUTHOR("Nick Vaccaro");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("loadable peek/poke driver");

