/*
 *  linux/drivers/mmc/pxa.c - PXA MMCI driver
 *
 *  Copyright (C) 2003 Russell King, All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  This hardware is really sick:
 *   - No way to clear interrupts.
 *   - Have to turn off the clock whenever we touch the device.
 *   - Doesn't tell you how many data blocks were transferred.
 *  Yuck!
 *
 *	1 and 3 byte data transfers not supported
 *	max block length up to 1023
 *
 * Lab126 changes Copyright (C) 2005, 2006 Lab126, Inc.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/dma-mapping.h>
#include <linux/mmc/host.h>
#include <linux/mmc/protocol.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <linux/sem.h>
#include <linux/ipc.h>
#include <linux/suspend.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/sizes.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/mmc.h>

#ifdef CONFIG_ARCH_LAB126
#ifdef CONFIG_ARCH_FIONA
#include <asm/arch/fiona.h>
#ifdef CONFIG_FIONA_PM_MMC
#include <asm/arch/fpow.h>
#endif
#else
#error unsupported board configuration
#endif

#include <mmc/mmcnotify.h>
#endif

#include "pxamci.h"

#ifdef CONFIG_MMC_DEBUG
//#define DBG(x...)	printk(KERN_DEBUG x)
#define _DBG(x...)	do { } while (0)
#define DBG(x...)	printk(x)
#define MDEBUG(x...)	printk(x)
#else
#define _DBG(x...)	do { } while (0)
#define DBG(x...)	do { } while (0)
#define MDEBUG(x...)	do { } while (0)
#endif

#if defined(CONFIG_MMC_DEBUG) || defined(CONFIG_PM_DEBUG)
#define pinfo(x...) printk(x)
#else
#define pinfo(x...) do {} while(0)
#endif

#define DRIVER_NAME	"pxa2xx-mci"

#define NR_SG	1

extern int mmc_suspend_host(struct mmc_host *host, u32 state);
extern int mmc_resume_host(struct mmc_host *host);

struct pxamci_host {
	struct mmc_host		*mmc;
	spinlock_t		lock;
	struct resource		*res;
	void			*base;
	int			irq;
	int			dma;
	unsigned int		clkrt;
	unsigned int		cmdat;
	unsigned int		imask;
	unsigned int		power_mode;
	struct pxamci_platform_data *pdata;

	struct mmc_request	*mrq;
	struct mmc_command	*cmd;
	struct mmc_data		*data;

	dma_addr_t		sg_dma;
	struct pxa_dma_desc	*sg_cpu;
	unsigned int		dma_len;

	unsigned int		dma_dir;
};

static inline unsigned int ns_to_clocks(unsigned int ns)
{
	return (ns * (CLOCKRATE / 1000000) + 999) / 1000;
}

static void pxamci_stop_clock(struct pxamci_host *host)
{
	_DBG("PXAMCI: pxamci_stop_clock called \n");
	if (readl(host->base + MMC_STAT) & STAT_CLK_EN) {
		unsigned long timeout = 10000;
		unsigned int v;

		writel(STOP_CLOCK, host->base + MMC_STRPCL);

		do {
			v = readl(host->base + MMC_STAT);
			if (!(v & STAT_CLK_EN))
				break;
			udelay(1);
		} while (timeout--);

		if (v & STAT_CLK_EN)
			dev_err(mmc_dev(host->mmc), "unable to stop clock\n");
	}
}

static void pxamci_enable_irq(struct pxamci_host *host, unsigned int mask)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	host->imask &= ~mask;
	writel(host->imask, host->base + MMC_I_MASK);
	spin_unlock_irqrestore(&host->lock, flags);
}

static void pxamci_disable_irq(struct pxamci_host *host, unsigned int mask)
{
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	host->imask |= mask;
	writel(host->imask, host->base + MMC_I_MASK);
	spin_unlock_irqrestore(&host->lock, flags);
}

static void pxamci_setup_data(struct pxamci_host *host, struct mmc_data *data)
{
	unsigned int nob = data->blocks;
	unsigned int timeout;
	u32 dcmd;
	int i;

	host->data = data;

	if (data->flags & MMC_DATA_STREAM)
		nob = 0xffff;

	writel(nob, host->base + MMC_NOB);
	writel(1 << data->blksz_bits, host->base + MMC_BLKLEN);

	timeout = ns_to_clocks(data->timeout_ns) + data->timeout_clks;
	writel((timeout + 255) / 256, host->base + MMC_RDTO);

	if (data->flags & MMC_DATA_READ) {
		host->dma_dir = DMA_FROM_DEVICE;
		dcmd = DCMD_INCTRGADDR | DCMD_FLOWTRG;
		DRCMRTXMMC = 0;
		DRCMRRXMMC = host->dma | DRCMR_MAPVLD;
	} else {
		host->dma_dir = DMA_TO_DEVICE;
		dcmd = DCMD_INCSRCADDR | DCMD_FLOWSRC;
		DRCMRRXMMC = 0;
		DRCMRTXMMC = host->dma | DRCMR_MAPVLD;
	}

	dcmd |= DCMD_BURST32 | DCMD_WIDTH1;

	host->dma_len = dma_map_sg(mmc_dev(host->mmc), data->sg, data->sg_len,
				   host->dma_dir);

	for (i = 0; i < host->dma_len; i++) {
		if (data->flags & MMC_DATA_READ) {
			host->sg_cpu[i].dsadr = host->res->start + MMC_RXFIFO;
			host->sg_cpu[i].dtadr = sg_dma_address(&data->sg[i]);
		} else {
			host->sg_cpu[i].dsadr = sg_dma_address(&data->sg[i]);
			host->sg_cpu[i].dtadr = host->res->start + MMC_TXFIFO;
		}
		host->sg_cpu[i].dcmd = dcmd | sg_dma_len(&data->sg[i]);
		host->sg_cpu[i].ddadr = host->sg_dma + (i + 1) *
					sizeof(struct pxa_dma_desc);
	}
	host->sg_cpu[host->dma_len - 1].ddadr = DDADR_STOP;
	wmb();

	DDADR(host->dma) = host->sg_dma;
	DCSR(host->dma) = DCSR_RUN;
}

static void pxamci_start_cmd(struct pxamci_host *host, struct mmc_command *cmd, unsigned int cmdat)
{

#ifndef CONFIG_ARCH_LAB126
	WARN_ON(host->cmd != NULL);
#endif

	host->cmd = cmd;

	if (cmd->flags & MMC_RSP_BUSY)
		cmdat |= CMDAT_BUSY;

	switch (cmd->flags & (MMC_RSP_MASK | MMC_RSP_CRC)) {
	case MMC_RSP_SHORT | MMC_RSP_CRC:
		cmdat |= CMDAT_RESP_SHORT;
		break;
	case MMC_RSP_SHORT:
		cmdat |= CMDAT_RESP_R3;
		break;
	case MMC_RSP_LONG | MMC_RSP_CRC:
		cmdat |= CMDAT_RESP_R2;
		break;
	default:
		break;
	}

	writel(cmd->opcode, host->base + MMC_CMD);
	writel(cmd->arg >> 16, host->base + MMC_ARGH);
	writel(cmd->arg & 0xffff, host->base + MMC_ARGL);
	writel(cmdat, host->base + MMC_CMDAT);
	writel(host->clkrt, host->base + MMC_CLKRT);

	writel(START_CLOCK, host->base + MMC_STRPCL);

	pxamci_enable_irq(host, END_CMD_RES);
}

static void pxamci_finish_request(struct pxamci_host *host, struct mmc_request *mrq)
{
	_DBG("PXAMCI: request done\n");
	host->mrq = NULL;
	host->cmd = NULL;
	host->data = NULL;
	mmc_request_done(host->mmc, mrq);
}

static int pxamci_cmd_done(struct pxamci_host *host, unsigned int stat)
{
	struct mmc_command *cmd = host->cmd;
	int i;
	u32 v;

	if (!cmd)
		return 0;

	host->cmd = NULL;

	/*
	 * Did I mention this is Sick.  We always need to
	 * discard the upper 8 bits of the first 16-bit word.
	 */
	v = readl(host->base + MMC_RES) & 0xffff;
	for (i = 0; i < 4; i++) {
		u32 w1 = readl(host->base + MMC_RES) & 0xffff;
		u32 w2 = readl(host->base + MMC_RES) & 0xffff;
		cmd->resp[i] = v << 24 | w1 << 8 | w2 >> 8;
		v = w2;
	}

	if (stat & STAT_TIME_OUT_RESPONSE) {
		cmd->error = MMC_ERR_TIMEOUT;
	} else if (stat & STAT_RES_CRC_ERR && cmd->flags & MMC_RSP_CRC) {
#ifdef CONFIG_PXA27x
		/*
		 * workaround for erratum #42:
		 * Intel PXA27x Family Processor Specification Update Rev 001
		 */
		if (cmd->opcode == MMC_ALL_SEND_CID ||
		    cmd->opcode == MMC_SEND_CSD ||
		    cmd->opcode == MMC_SEND_CID) {
			/* a bogus CRC error can appear if the msb of
			   the 15 byte response is a one */
			if ((cmd->resp[0] & 0x80000000) == 0)
				cmd->error = MMC_ERR_BADCRC;
		} else {
			DBG("ignoring CRC from command %d - *risky*\n",cmd->opcode);
		}
#else
		cmd->error = MMC_ERR_BADCRC;
#endif
	}

	pxamci_disable_irq(host, END_CMD_RES);
	if (host->data && cmd->error == MMC_ERR_NONE) {
		pxamci_enable_irq(host, DATA_TRAN_DONE);
	} else {
		pxamci_finish_request(host, host->mrq);
	}

	return 1;
}

static int pxamci_data_done(struct pxamci_host *host, unsigned int stat)
{
	struct mmc_data *data = host->data;

	if (!data)
		return 0;

	DCSR(host->dma) = 0;
	dma_unmap_sg(mmc_dev(host->mmc), data->sg, host->dma_len, host->dma_dir);

	if (stat & STAT_READ_TIME_OUT)
		data->error = MMC_ERR_TIMEOUT;
	else if (stat & (STAT_CRC_READ_ERROR|STAT_CRC_WRITE_ERROR))
		data->error = MMC_ERR_BADCRC;

	/*
	 * There appears to be a hardware design bug here.  There seems to
	 * be no way to find out how much data was transferred to the card.
	 * This means that if there was an error on any block, we mark all
	 * data blocks as being in error.
	 */
	if (data->error == MMC_ERR_NONE)
		data->bytes_xfered = data->blocks << data->blksz_bits;
	else
		data->bytes_xfered = 0;

	pxamci_disable_irq(host, DATA_TRAN_DONE);

	host->data = NULL;
	if (host->mrq->stop && data->error == MMC_ERR_NONE) {
		pxamci_stop_clock(host);
		pxamci_start_cmd(host, host->mrq->stop, 0);
	} else {
		pxamci_finish_request(host, host->mrq);
	}

	return 1;
}

static irqreturn_t pxamci_irq(int irq, void *devid, struct pt_regs *regs)
{
	struct pxamci_host *host = devid;
	unsigned int ireg;
	int handled = 0;

	ireg = readl(host->base + MMC_I_REG);

	_DBG("PXAMCI: irq %08x\n", ireg);

	if (ireg) {
		unsigned stat = readl(host->base + MMC_STAT);

		_DBG("PXAMCI: stat %08x\n", stat);

		if (ireg & END_CMD_RES)
			handled |= pxamci_cmd_done(host, stat);
		if (ireg & DATA_TRAN_DONE)
			handled |= pxamci_data_done(host, stat);
	}

	return IRQ_RETVAL(handled);
}

static void pxamci_request(struct mmc_host *mmc, struct mmc_request *mrq)
{
	struct pxamci_host *host = mmc_priv(mmc);
	unsigned int cmdat;

#ifndef CONFIG_ARCH_LAB126
	/* mrq can be NULL on card abort case */
	WARN_ON(host->mrq != NULL);
#endif

	host->mrq = mrq;

	pxamci_stop_clock(host);

	cmdat = host->cmdat;
	host->cmdat &= ~CMDAT_INIT;

	if (mrq->data) {
		pxamci_setup_data(host, mrq->data);

		cmdat &= ~CMDAT_BUSY;
		cmdat |= CMDAT_DATAEN | CMDAT_DMAEN;
		if (mrq->data->flags & MMC_DATA_WRITE)
			cmdat |= CMDAT_WRITE;

		if (mrq->data->flags & MMC_DATA_STREAM)
			cmdat |= CMDAT_STREAM;
	}

	pxamci_start_cmd(host, mrq->cmd, cmdat);
}

static int pxamci_get_ro(struct mmc_host *mmc)
{
	struct pxamci_host *host = mmc_priv(mmc);

	if (host->pdata && host->pdata->get_ro)
		return host->pdata->get_ro(mmc->dev);

	/* Host doesn't support read only detection so assume writeable */
	return 0;
}

static void pxamci_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct pxamci_host *host = mmc_priv(mmc);

	DBG("pxamci_set_ios: clock %u power %u vdd %u.%02u\n",
	    ios->clock, ios->power_mode, ios->vdd / 100,
	    ios->vdd % 100);

	if (ios->clock) {
		unsigned int clk = CLOCKRATE / ios->clock;
		if(clk > (1<<6)) clk = (1<<6);
		host->clkrt = fls(clk) - 1;
		MDEBUG("PXAMCI: Starting HW clock in pxamci_set_ioc\n");
		pxa_set_cken(CKEN12_MMC, 1);

		/*
		 * we write clkrt on the next command
		 */
	} else {
		pxamci_stop_clock(host);
		MDEBUG("PXAMCI: Stopping HW clock in pxamci_set_ioc\n");
		pxa_set_cken(CKEN12_MMC, 0);
	}

	if (host->power_mode != ios->power_mode) {
		host->power_mode = ios->power_mode;

		if (host->pdata && host->pdata->setpower)
			host->pdata->setpower(mmc->dev, ios->vdd);

		if (ios->power_mode == MMC_POWER_ON)
			host->cmdat |= CMDAT_INIT;
	}

	DBG("pxamci_set_ios: clkrt = %x cmdat = %x\n",
	    host->clkrt, host->cmdat);
}

#ifdef CONFIG_ARCH_LAB126

void pxamci_clear(struct mmc_host *mmc)
{
	struct pxamci_host *host = mmc_priv(mmc);

	pxamci_stop_clock(host);
	writel(TXFIFO_WR_REQ|RXFIFO_RD_REQ|CLK_IS_OFF|STOP_CMD|
	       END_CMD_RES|PRG_DONE|DATA_TRAN_DONE,
	       host->base + MMC_I_MASK);

	DRCMRRXMMC = 0;
	DRCMRTXMMC = 0;
}

int pxamci_card_inserted(void)
{
	return mmc_query_insertion_state();
}

#define MMC_NOTIFY_THREAD   "mmcdd"

static DECLARE_COMPLETION(mmc_detect_thread_exited);
static DECLARE_MUTEX(detect_host_lock);

static struct {
	pid_t thread_pid;
	wait_queue_head_t event_wq;
	struct mmc_host *host;
	unsigned int event_count;
} mmc_detect_data;

#undef NOTIFY_SEM

#ifdef NOTIFY_SEM
static void mmc_detect_notify_sem(void)
{
	int sem_id;

	// retrieve the notification semaphore
	sem_id = sys_semget(MMC_NOTIFY_SEM_KEY, 1, 0666);
	if (sem_id >= 0) {
		struct sembuf sem_operations[1];

		// signal the semaphore
		sem_operations[0].sem_num = 0;		// which semaphore to use
		sem_operations[0].sem_op = 1;		// add one from the semaphore value
		sem_operations[0].sem_flg = 0;		// set wait flag

		// do the signal
		sys_semop(sem_id, sem_operations, 1);

	} else {
		// semaphore doesn't exist; ignore as nobody may be waiting on it

	}
}
#endif

static void mmc_detect_notify(int enabled, int notify)
{
	int err;

	char *envp[] = {
		"HOME=/",
		"TERM=linux",
		"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
		NULL
	};

	char *argv[] = {
		MMC_NOTIFY_SCRIPT,
		enabled ? "1" : "0",
		notify ? "1" : "0",
		NULL
	};

	MDEBUG("mmc_detect_notify(CARD_%s)\n", (enabled ? "INSERTED" : "REMOVED"));
	if ((err = call_usermodehelper(argv[0], argv, envp, 0)) < 0) {
		printk("mmc: notification exec failed: \"" MMC_NOTIFY_SCRIPT "\", err %d\n", err);
	}
}

#ifdef CONFIG_FIONA_PM_MMC
static void mmc_detect_notify_signal(struct mmc_host *host);

struct mmc_host *g_pm_mmc_resume_host = NULL;

static int pm_mmc_power_mode = FPOW_MODE_ON;

static int mmc_bus_fpow_getmode(void *private);
static FPOW_ERR mmc_bus_fpow_setmode(void *private, u32 state, u32 mode);

static fpow_component *mmc_fpow_dev = NULL;

static int mmc_register_with_fpow(struct device *dev, fpow_component **fpow_dev)
{
	struct fpow_registration_rec reg;
	int rc = 0;

	memclr(&reg, sizeof(reg));
	strcpy((char*)&reg.name, "mmc");

	reg.device_class = FPOW_CLASS_MMC;
	reg.supported_modes = FPOW_MODE_ON_SUPPORTED | FPOW_MODE_SLEEP_SUPPORTED | FPOW_MODE_OFF_SUPPORTED;
	reg.private = (void *)dev;
	reg.getmode = mmc_bus_fpow_getmode;
	reg.setmode = mmc_bus_fpow_setmode;
	*fpow_dev = fpow_register(&reg);
	if (*fpow_dev == NULL) {
		printk("MMC: Failed to register with fpow !!\n");
		rc = -1;
	}

	return rc;
}

static FPOW_ERR mmc_bus_fpow_setmode(void *private, u32 state, u32 mode)
{
	FPOW_ERR ret = FPOW_NO_ERR;

	powdebug("mmc_bus_fpow_setmode(mode=%s)\n", fpow_mode_to_str(mode));

	// if we're already in the correct mode, we don't need to do anything
	if (pm_mmc_power_mode != mode) { 

		switch (mode)  {

			case FPOW_MODE_ON :
				pm_mmc_power_mode = mode;

				mmc_detect_notify_signal(g_pm_mmc_resume_host);
				g_pm_mmc_resume_host = NULL;
				break;

			case FPOW_MODE_SLEEP :
			case FPOW_MODE_OFF :
				pm_mmc_power_mode = mode;
				break;

			default :
				ret = -EINVAL;
				break;
		}

	}

	return ret;
}

static int mmc_bus_fpow_getmode(void *private)
{
	powdebug("mmc_bus_fpow_getmode()\n");

	return pm_mmc_power_mode;
}

static void mmc_perform_suspend(void)
{
	// force an unmount of the card, without notification
	mmc_detect_notify(0, 0);
}
#endif


#define SETTLE_COUNT_MAX		10
#define SETTLE_DELAY			10

static int mmc_detect_notify_thread(void *arg)
{
	int last_notify = -1;
	int settle_count;

	daemonize(MMC_NOTIFY_THREAD);
	allow_signal(SIGKILL);

	while (1) {
		wait_event_interruptible(mmc_detect_data.event_wq, mmc_detect_data.event_count != 0);

#ifdef CONFIG_PM
		if (current->flags & PF_FREEZE) {
			refrigerator(PF_FREEZE);

			if (!signal_pending(current)) {
				continue;
			}
		}
#endif
		if (signal_pending(current)) {
			break;
		}

		settle_count = 0;

		while (++settle_count < SETTLE_COUNT_MAX) {
			int prev_event_count = mmc_detect_data.event_count;

			msleep(SETTLE_DELAY);

			if (prev_event_count != mmc_detect_data.event_count) {
				settle_count = 0;
			}
		}

		mmc_detect_data.event_count = 0;

#ifdef CONFIG_FIONA_PM_MMC
		if (pm_mmc_power_mode == FPOW_MODE_ON) { 
#endif

			int insertion_state = mmc_query_insertion_state();
			int notify = insertion_state || last_notify != insertion_state;

			struct mmc_host *host;

			down(&detect_host_lock);
			host = mmc_detect_data.host;
			mmc_detect_data.host = NULL;
			up(&detect_host_lock);

			if (host != NULL) {
				mmc_detect_change(host);

				flush_scheduled_work();
			}

			last_notify = insertion_state;

#ifdef NOTIFY_SEM
			// notify any waiting processes
			mmc_detect_notify_sem();
#endif

			// execute user-level notify script
			mmc_detect_notify(insertion_state, notify);

#ifdef CONFIG_FIONA_PM_MMC
		}
#endif

	}

	complete_and_exit(&mmc_detect_thread_exited, 0);

	return 0;
}

static void mmc_detect_notify_init(void)
{
	pid_t pid;

	memset(&mmc_detect_data, 0, sizeof(mmc_detect_data));

	init_waitqueue_head(&mmc_detect_data.event_wq);

	pid = kernel_thread(mmc_detect_notify_thread, NULL, CLONE_KERNEL);
	if (pid >= 0) {
		mmc_detect_data.thread_pid = pid;

	} else {
		mmc_detect_data.thread_pid = 0;
		printk("mmc: thread " MMC_NOTIFY_THREAD " create failed (%d)\n", pid);

	}
}

static void mmc_detect_notify_done(void)
{
	int ret;

	if (mmc_detect_data.thread_pid > 0) {
		ret = kill_proc(mmc_detect_data.thread_pid, SIGKILL, 1);
		if (ret == 0) {
			wait_for_completion(&mmc_detect_thread_exited);
		}
	}
}

static void mmc_detect_notify_signal(struct mmc_host *host)
{
	down(&detect_host_lock);
	mmc_detect_data.host = host;
	up(&detect_host_lock);

	++mmc_detect_data.event_count;

	pinfo("mmc_detect_notify_signal()\n");
	wake_up(&mmc_detect_data.event_wq);
}

static int proc_status_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = sprintf(page, "%d\n", mmc_query_insertion_state());

	*eof = 1;

	return len;
}

static int proc_ro_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = sprintf(page, "%d\n", mmc_query_readonly_state());

	*eof = 1;

	return len;
}

static void mmc_proc_init(void)
{
	struct proc_dir_entry *proc_mmc_parent, *proc_mmc_status, *proc_mmc_ro;

	proc_mmc_parent = create_proc_entry("mmc", S_IFDIR | S_IRUGO | S_IXUGO, NULL);
	if (proc_mmc_parent != NULL) {
		proc_mmc_status = create_proc_entry("status", S_IRUGO, proc_mmc_parent);
		if (proc_mmc_status != NULL) {
			proc_mmc_status->data = NULL;
			proc_mmc_status->read_proc = proc_status_read;
			proc_mmc_status->write_proc = NULL;
		}

		proc_mmc_ro = create_proc_entry("readonly", S_IRUGO, proc_mmc_parent);
		if (proc_mmc_ro != NULL) {
			proc_mmc_ro->data = NULL;
			proc_mmc_ro->read_proc = proc_ro_read;
			proc_mmc_ro->write_proc = NULL;
		}
	}
}
#endif

static struct mmc_host_ops pxamci_ops = {
	.request	= pxamci_request,
	.get_ro		= pxamci_get_ro,
	.set_ios	= pxamci_set_ios,
#ifdef CONFIG_ARCH_LAB126
	.clear		= pxamci_clear,
	.card_inserted	= pxamci_card_inserted,
#endif
};

static void pxamci_dma_irq(int dma, void *devid, struct pt_regs *regs)
{
	printk(KERN_ERR "DMA%d: IRQ???\n", dma);
	DCSR(dma) = DCSR_STARTINTR|DCSR_ENDINTR|DCSR_BUSERR;
}

static irqreturn_t pxamci_detect_irq(int irq, void *devid, struct pt_regs *regs)
{
	MDEBUG("pxamci_detect_irq() enter\n");

	mmc_detect_notify_signal(devid);

	MDEBUG("pxamci_detect_irq() exit\n");
	return IRQ_HANDLED;
}

static int pxamci_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct mmc_host *mmc;
	struct pxamci_host *host = NULL;
	struct resource *r;
	int ret, irq;

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!r || irq == NO_IRQ)
		return -ENXIO;

	r = request_mem_region(r->start, SZ_4K, DRIVER_NAME);
	if (!r)
		return -EBUSY;

	mmc = mmc_alloc_host(sizeof(struct pxamci_host), dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto out;
	}

	mmc->ops = &pxamci_ops;
	mmc->f_min = CLOCKRATE_MIN;
	mmc->f_max = CLOCKRATE_MAX;

	/*
	 * We can do SG-DMA, but we don't because we never know how much
	 * data we successfully wrote to the card.
	 */
	mmc->max_phys_segs = NR_SG;

	/*
	 * Our hardware DMA can handle a maximum of one page per SG entry.
	 */
	mmc->max_seg_size = PAGE_SIZE;

	host = mmc_priv(mmc);
	host->mmc = mmc;
	host->dma = -1;
	host->pdata = pdev->dev.platform_data;
	mmc->ocr_avail = host->pdata ?
			 host->pdata->ocr_mask :
			 MMC_VDD_32_33|MMC_VDD_33_34;

	host->sg_cpu = dma_alloc_coherent(dev, PAGE_SIZE, &host->sg_dma, GFP_KERNEL);
	if (!host->sg_cpu) {
		ret = -ENOMEM;
		goto out;
	}

	spin_lock_init(&host->lock);
	host->res = r;
	host->irq = irq;
	host->imask = MMC_I_MASK_ALL;

	host->base = ioremap(r->start, SZ_4K);
	if (!host->base) {
		ret = -ENOMEM;
		goto out;
	}

	/*
	 * Ensure that the host controller is shut down, and setup
	 * with our defaults.
	 */
	pxamci_stop_clock(host);
	writel(0, host->base + MMC_SPI);
	writel(64, host->base + MMC_RESTO);
	writel(host->imask, host->base + MMC_I_MASK);

	host->dma = pxa_request_dma(DRIVER_NAME, DMA_PRIO_LOW,
					pxamci_dma_irq, host);
	if (host->dma < 0) {
		ret = -EBUSY;
		goto out;
	}

	MDEBUG("MMC DEBUG: Request %d RISING IRQ for mmc\n",host->irq);

	set_irq_type(host->irq,IRQT_RISING);

	ret = request_irq(host->irq, pxamci_irq, 0, DRIVER_NAME, host);
	if (ret)
		goto out;

	dev_set_drvdata(dev, mmc);

	if (host->pdata && host->pdata->init)
		host->pdata->init(dev, pxamci_detect_irq, mmc);

	mmc_add_host(mmc);

	return 0;

 out:
	if (host) {
		if (host->dma >= 0)
			pxa_free_dma(host->dma);
		if (host->base)
			iounmap(host->base);
		if (host->sg_cpu)
			dma_free_coherent(dev, PAGE_SIZE, host->sg_cpu, host->sg_dma);
	}
	if (mmc)
		mmc_free_host(mmc);
	release_resource(r);
	return ret;
}

static int pxamci_remove(struct device *dev)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);

	dev_set_drvdata(dev, NULL);

	if (mmc) {
		struct pxamci_host *host = mmc_priv(mmc);

		if (host->pdata && host->pdata->exit)
			host->pdata->exit(dev, mmc);

		mmc_remove_host(mmc);

		pxamci_stop_clock(host);
		writel(TXFIFO_WR_REQ|RXFIFO_RD_REQ|CLK_IS_OFF|STOP_CMD|
			   END_CMD_RES|PRG_DONE|DATA_TRAN_DONE,
			   host->base + MMC_I_MASK);

		DRCMRRXMMC = 0;
		DRCMRTXMMC = 0;

		free_irq(host->irq, host);
		pxa_free_dma(host->dma);
		iounmap(host->base);
		dma_free_coherent(dev, PAGE_SIZE, host->sg_cpu, host->sg_dma);

		release_resource(host->res);

		mmc_free_host(mmc);
	}
	return 0;
}

#ifdef CONFIG_PM
static int pxamci_suspend(struct device *dev, u32 state, u32 level)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);
	int ret = 0;

	pinfo("MMC: pxamci_suspend(%d)\n",level);

	if (mmc && level == SUSPEND_DISABLE) {

#ifdef CONFIG_FIONA_PM_MMC
		mmc_perform_suspend();
#endif

		pinfo("MMC: pxamci_suspend(%d) calling mmc_suspend_host\n",level);
		ret = mmc_suspend_host(mmc, state);
	}

	return ret;
}

static int pxamci_resume(struct device *dev, u32 level)
{
	struct mmc_host *mmc = dev_get_drvdata(dev);
	int ret = 0;

	pinfo("MMC: pxamci_resume(%d)\n",level);

	if (mmc && level == RESUME_ENABLE) {
		pinfo("MMC: pxamci_resume(%d) calling mmc_resume_host\n",level);
		ret = mmc_resume_host(mmc);
	}

	return ret;
}
#else
#define pxamci_suspend	NULL
#define pxamci_resume	NULL
#endif

static struct device_driver pxamci_driver = {
	.name		= DRIVER_NAME,
	.bus		= &platform_bus_type,
	.probe		= pxamci_probe,
	.remove		= pxamci_remove,
	.suspend	= pxamci_suspend,
	.resume		= pxamci_resume,
};

static int __init pxamci_init(void)
{
#ifdef CONFIG_ARCH_LAB126
	mmc_proc_init();
	mmc_detect_notify_init();

#ifdef CONFIG_FIONA_PM_MMC
	if (mmc_register_with_fpow(NULL, &mmc_fpow_dev)) {
		printk("MMC failed to register with Fiona Power Manager.\n");
	}
#endif
#endif

	return driver_register(&pxamci_driver);
}

static void __exit pxamci_exit(void)
{
#ifdef CONFIG_ARCH_LAB126
	mmc_detect_notify_done();

#ifdef CONFIG_FIONA_PM_MMC
	fpow_unregister(mmc_fpow_dev);
	mmc_fpow_dev = NULL;
#endif
#endif
	driver_unregister(&pxamci_driver);
}

module_init(pxamci_init);
module_exit(pxamci_exit);

MODULE_DESCRIPTION("PXA Multimedia Card Interface Driver");
MODULE_LICENSE("GPL");
