/******************************************************************
 *
 *  File:   usb_device_pm.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   08/14/06
 *
 *  Copyright 2006, Lab126, Inc.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *  Description:
 *      USB Device Fiona Power Management System Support
 *
 ******************************************************************/

#include <asm/arch/fpow.h>	// FPOW-related constants



// Static global holding current FPOW power state of the driver
static FPOW_POWER_MODE	g_usb_device_power_mode = FPOW_MODE_ON; 

// Static global holding fpow device entry returned from fpow_register()
struct fpow_component *g_usb_device_fpow_component_ptr = NULL;


static int
usb_device_power_state(void *private, FPOW_POWER_MODE mode)
{
	int	retVal = 0;
	int	force_off = (FPOW_MODE_OFF == mode);

	powdebug("usb_device_power_state(%s)\n", fpow_mode_to_str(mode));

	// Switch to the appropriate power level
	switch (mode)  {
	case FPOW_MODE_ON:
		// Nothing needs to be done - the part is already on.
            	break;

	case FPOW_MODE_SLEEP:	// non-drivemode with the power switch on
	case FPOW_MODE_OFF:	// power switch off
		pdc_suspend_controller(force_off);
		break;

	default:
		retVal = -FPOW_COMPONENT_TRANSITION_ERR;
		break;
	}
	return(retVal);
}


/*******************************************************************
 * Routine:     usb_device_fpow_setmode
 *
 * Purpose:     Handles Fiona Power Management callbacks
 *              for doing USB Device power management.
 *
 * Paramters:   private - value passed at registration - currently ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                        Currently supported Modes:
 *                          FPOW_MODE_ON
 *                          FPOW_MODE_SLEEP
 *                          FPOW_MODE_OFF
 *
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - not enough power to transition
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: none
 *******************************************************************/

static FPOW_ERR
usb_device_fpow_setmode(void *private, u32 state, u32 mode)
{
	int	retVal = FPOW_NO_ERR;

	powdebug("usb_device_fpow_setmode(state=%s, mode=%s)\n",\
        		fpow_state_to_str(state), fpow_mode_to_str(mode));
	DBG_MSG("%s(%s)\n", __FUNCTION__, fpow_mode_to_str(mode));

	// Switch to the appropriate power level
	switch (mode)  {
		case FPOW_MODE_ON:
		case FPOW_MODE_SLEEP:
		case FPOW_MODE_OFF:
			retVal = usb_device_power_state(private,
						(FPOW_POWER_MODE) mode); 
			if (retVal == FPOW_NO_ERR)
				g_usb_device_power_mode = mode;
			break;

		default:
			retVal = -EINVAL;
			break;
	}

	powdebug("USB Device: driver now in %s mode\n",
			fpow_mode_to_str(g_usb_device_power_mode));
	DBG_MSG("USB Device: driver now in %s mode\n",
			fpow_mode_to_str(g_usb_device_power_mode));
	return(retVal);
}

static int
usb_device_fpow_getmode(void *private)
{
	powdebug("usb_device_fpow_getmode(0x%x) returning %s\n", (int) private,
			fpow_mode_to_str(g_usb_device_power_mode));
	DBG_MSG("usb_device_fpow_getmode(0x%x) returning %s\n", (int) private,
			fpow_mode_to_str(g_usb_device_power_mode));
	
	return (g_usb_device_power_mode);
}

int 
usb_device_register_with_fpow(void *private, struct fpow_component **fpow_dev)
{
	struct fpow_registration_rec   reg;
	int rc = 0;

	powdebug("usb_device_register_with_fpow(0x%x, 0x%x)\n",(int)private,
			(int)fpow_dev);
	DBG_MSG("usb_device_register_with_fpow(0x%x, 0x%x)\n",(int)private,
			(int)fpow_dev);

	memclr(&reg,sizeof(reg));

	strcpy(reg.name,"usbd");
	reg.device_class = FPOW_CLASS_USB_DEVICE;
	reg.supported_modes = ( FPOW_MODE_ON_SUPPORTED | \
				FPOW_MODE_SLEEP_SUPPORTED | \
				FPOW_MODE_OFF_SUPPORTED);
	reg.private = private;
	reg.getmode = usb_device_fpow_getmode;
	reg.setmode = usb_device_fpow_setmode;
	*fpow_dev = fpow_register(&reg);
	if (*fpow_dev == NULL) {
		printk("USB Device: Failed to register with fpow !!\n");
		rc = -1;
	}
	return(rc);
}

void
usb_device_unregister_with_fpow(struct fpow_component *fpow_dev)
{
	fpow_unregister(fpow_dev);
}
