/*
 * linux/sound/soc-dpm.h -- ALSA SoC Dynamic Power Management
 *
 * Author:		Liam Girdwood
 * Created:		Aug 11th 2005
 * Copyright:	Wolfson Microelectronics. PLC.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
 
#ifndef __LINUX_SND_SOC_DPM_H
#define __LINUX_SND_SOC_DPM_H

#include <linux/device.h>
#include <linux/types.h>
#include <sound/control.h>
#include <sound/soc.h>

/*
 * SoC dynamic power managment
 *
 * We can have upto 4 power domains
 * 	1. Codec domain - VREF, VMID
 *     Usually controlled at codec probe/remove, although can be set
 *     at stream time if power is not needed for sidetone, etc.
 *  2. Platform/Machine domain - physically connected inputs and outputs
 *     Is platform/machine and user action specific, is set in the .asoundrc file and
 *     by userspace e.g when HP are inserted
 *  3. Path domain - Internal codec path mixers
 *     Are automatically set when mixer and mux settings are
 *     changed by the user.
 *  4. Stream domain - DAC's and ADC's.
 *     Enabled when stream playback/capture is started.  
 */

/* codec domain - vmid, vref handled by the codec driver */  
#define SND_SOC_DPM_CLOCK(wname, preg, pshift, pinvert) \
{.id = snd_soc_dpm_clock, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = NULL, .num_kcontrols = 0}
/* not a true power register, but is used by dpm to minimise pops */ 
#define SND_SOC_DPM_MUTE(wname, preg, pshift, pinvert) \
{.id = snd_soc_dpm_mute, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = NULL, .num_kcontrols = 0}
 
/* platform domain */
#define SND_SOC_DPM_INPUT(wname) \
{.id = snd_soc_dpm_input, .name = wname, .kcontrols = NULL, \
 .num_kcontrols = 0}
#define SND_SOC_DPM_OUTPUT(wname) \
{.id = snd_soc_dpm_output, .name = wname, .kcontrols = NULL, \
 .num_kcontrols = 0}
#define SND_SOC_DPM_MIC(wname) \
{.id = snd_soc_dpm_mic, .name = wname, .kcontrols = NULL, \
 .num_kcontrols = 0}
#define SND_SOC_DPM_HP(wname) \
{.id = snd_soc_dpm_hp, .name = wname, .kcontrols = NULL, \
 .num_kcontrols = 0}
#define SND_SOC_DPM_AMP(wname, wext_event) \
{.id = snd_soc_dpm_amp, .name = wname, .kcontrols = NULL, \
 .num_kcontrols = 0, .ext_event = wext_event}
 
/* path domain */
#define SND_SOC_DPM_VOLUME(wname, preg, pshift, pinvert, wcontrols, wncontrols) \
{.id = snd_soc_dpm_volume, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = wcontrols, .num_kcontrols = wncontrols}
#define SND_SOC_DPM_MUX(wname, preg, pshift, pinvert, wcontrols) \
{.id = snd_soc_dpm_mux, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert,	.kcontrols = wcontrols, .num_kcontrols = 1} 
#define SND_SOC_DPM_MIXER(wname, preg, pshift, pinvert, wcontrols, wncontrols) \
{.id = snd_soc_dpm_mixer, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = wcontrols, .num_kcontrols = wncontrols} 
#define SND_SOC_DPM_MICBIAS(wname, preg, pshift, pinvert) \
{.id = snd_soc_dpm_micbias, .name = wname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = NULL, .num_kcontrols = 0} 

/* stream domain */
#define SND_SOC_DPM_DAC(wname, stname, preg, pshift, pinvert, wcontrols, wncontrols) \
{.id = snd_soc_dpm_dac, .name = wname, .sname = stname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = wcontrols, .num_kcontrols = wncontrols}
#define SND_SOC_DPM_ADC(wname, stname, preg, pshift, pinvert, wcontrols, wncontrols) \
{.id = snd_soc_dpm_adc, .name = wname, .sname = stname, .reg = preg, .shift = pshift, \
 .invert = pinvert, .kcontrols = wcontrols, .num_kcontrols = wncontrols}

#define SOC_DPM_SINGLE(xname, reg, shift, mask, invert) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .info = snd_soc_info_volsw, \
  .get = snd_soc_dpm_get_volsw, .put = snd_soc_dpm_put_volsw, \
  .private_value =  SOC_SINGLE_VALUE(reg, shift, mask, invert) }
#define SOC_DPM_DOUBLE(xname, reg, shift_left, shift_right, mask, invert, power) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = (xname), .info = snd_soc_info_volsw, \
  .get = snd_soc_dpm_get_volsw, .put = snd_soc_dpm_put_volsw, \
  .private_value = (reg) | ((shift_left) << 8) | ((shift_right) << 12) | ((mask) << 16) | ((invert) << 24) }
#define SOC_DPM_ENUM(xname, xenum) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .info = snd_soc_info_enum_double, \
  .get = snd_soc_dpm_get_enum_double, .put = snd_soc_dpm_put_enum_double, \
  .private_value = (unsigned long)&xenum }

#define SND_SOC_DPM_STREAM_START		0x1
#define SND_SOC_DPM_STREAM_STOP			0x2
#define SND_SOC_DPM_STREAM_SUSPEND		0x4
#define SND_SOC_DPM_STREAM_PAUSE_PUSH	0x8
#define SND_SOC_DPM_STREAM_PAUSE_RELEASE	0x10  

struct snd_soc_dpm_widget;
enum snd_soc_dpm_type;
struct snd_soc_dpm_path;
struct snd_soc_dpm_pin;
	
int snd_soc_dpm_put_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_dpm_get_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_dpm_get_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_dpm_put_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol);
int snd_soc_dpm_new_control(struct snd_soc_codec *codec, const struct snd_soc_dpm_widget *widget);
int snd_soc_dpm_connect_input(struct snd_soc_codec *codec, const char *target_name, 
	const char *control_name, const char *src_name);
int snd_soc_dpm_sync(struct snd_soc_codec *codec);
int snd_soc_dpm_stream_event(struct snd_soc_codec *codec, const char * stream, int event);
int snd_soc_dpm_codec_event(struct snd_soc_codec *codec, int event);
void snd_soc_dpm_sys_remove(struct device *dev);
int snd_soc_dpm_sys_add(struct device *dev);
int snd_soc_dpm_codec_mute(struct snd_soc_codec *codec, int mute);
int snd_soc_dpm_set_connection(struct snd_soc_codec *codec, const struct snd_soc_dpm_pin *pin);

enum snd_soc_dpm_type {
	snd_soc_dpm_input = 0,
	snd_soc_dpm_output,
	snd_soc_dpm_mux,
	snd_soc_dpm_mixer,
	snd_soc_dpm_volume,
	snd_soc_dpm_adc,
	snd_soc_dpm_dac,
	snd_soc_dpm_clock,
	snd_soc_dpm_vref,
	snd_soc_dpm_micbias,
	snd_soc_dpm_mute,
	snd_soc_dpm_mic,
	snd_soc_dpm_hp,
	snd_soc_dpm_amp,
};

struct snd_soc_dpm_path {
	char *name;
	char *long_name;
	struct snd_soc_dpm_widget *in;
	struct snd_soc_dpm_widget *out;
	snd_kcontrol_t *kcontrol;
	
	u8 connect:1;	/* applies to widget inputs only */
	u8 walked:1;	/* path has been walked */
	
	struct list_head list_in;
	struct list_head list_out;
	struct list_head list;
};

struct snd_soc_dpm_widget {
	enum snd_soc_dpm_type id;
	char *name;
	char *sname;
	struct snd_soc_codec *codec;
	struct list_head list;
	
	/* dpm control */
	short reg;						/* negative reg = no direct dpm */
	unsigned char shift;
	unsigned int value;				/* widget saved value */ 
	unsigned char power:1;			/* block power status */
	unsigned char invert:1;
	unsigned char active:1;			/* active stream on DAC, ADC's */
	unsigned char connected:1;		/* connected codec pin */
	unsigned char sync:1;			/* sync complete */
	unsigned char ext:1;			/* has external widgets */
	unsigned char muted;			/* muted for pop reduction */
	int (*ext_event)(struct snd_soc_codec*, int);
	
	/* kcontrols that relate to this widget */
	int num_kcontrols;
	const snd_kcontrol_new_t * kcontrols;
	
	/* widget input and outputs */
	struct list_head inputs;
	struct list_head outputs;
};

/* codec pin connections */
struct snd_soc_dpm_pin {
	char *name;
	int connected; /* this is dynamic for insertion/removal events e.g. hp, mic */
};

#endif
