/*
 * Copyright 2005 Openedhand Ltd.
 *
 * Author: Richard Purdie <richard@openedhand.com>
 *
 * Based on WM8753.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/i2c.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>
#include <sound/initval.h>

#include "wm8750.h"

#define AUDIO_NAME "WM8750"
#define WM8750_VERSION "0.4"

/*
 * Debug
 */
 
#define PFX AUDIO_NAME
#define WM8750_DEBUG 0

#ifdef WM8750_DEBUG
#define dbg(format, arg...) printk(KERN_DEBUG PFX ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif
#define err(format, arg...) printk(KERN_ERR PFX ": " format "\n" , ## arg)
#define info(format, arg...) printk(KERN_INFO PFX ": " format "\n" , ## arg)
#define warn(format, arg...) printk(KERN_WARNING PFX ": " format "\n" , ## arg)


/* 
 * WM8750 2 wire address is determined by CSB pin state during powerup */
#define I2C_DRIVERID_WM8750 0xfde

static unsigned short normal_i2c[] = {  0x1a, 0x1b,
                                       I2C_CLIENT_END };

/* Magic definition of all other variables and things */
I2C_CLIENT_INSMOD;

static struct i2c_driver wm8750_i2c_driver;
static struct i2c_client client_template;
	
/*
 * wm8750 register cache
 * We can't read the WM8750 register space when we 
 * are using 2 wire for device control, so we cache them instead. 
 */
static const u16 wm8750_reg[] = {
	0x0097, 0x0097, 0x0079, 0x0079,  /*  0 */
	0x0000, 0x0008, 0x0000, 0x000a,  /*  4 */
	0x0000, 0x0000, 0x00ff, 0x00ff,  /*  8 */
	0x000f, 0x000f, 0x0000, 0x0000,  /* 12 */
	0x0000, 0x007b, 0x0000, 0x0032,  /* 16 */
	0x0000, 0x00c3, 0x00c3, 0x00c0,  /* 20 */
	0x0000, 0x0000, 0x0000, 0x0000,  /* 24 */
	0x0000, 0x0000, 0x0000, 0x0000,  /* 28 */
	0x0000, 0x0000, 0x0050, 0x0050,  /* 32 */
	0x0050, 0x0050, 0x0050, 0x0050,  /* 36 */
	0x0079, 0x0079, 0x0079,          /* 40 */
};

#define WM8750_HWFMT \
	(SND_SOC_HWFMT_I2S | SND_SOC_HWFMT_LEFT_J | SND_SOC_HWFMT_RIGHT_J | \
	SND_SOC_HWFMT_CBS_CFS | SND_SOC_HWFMT_CBM_CFM |\
	SND_SOC_HWFMT_NB_NF | SND_SOC_HWFMT_NB_IF | SND_SOC_HWFMT_IB_NF | \
	SND_SOC_HWFMT_IB_IF)

/* priv1 is srate register setting */
static struct snd_soc_hw_mode wm8750_hwfmt[] = {
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 8000,   64, 0x000c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 8000,   96, 0x000e},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 8000,  128, 0x004c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 8000,  192, 0x004e},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 11025, 64,  0x0030},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 11025, 96,  0x0032},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 11025, 128, 0x0070},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 11025, 192, 0x0072},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 12000, 64,  0x0010},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 12000, 96,  0x0012},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 12000, 128, 0x0050},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 12000, 192, 0x0052},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 16000, 64,  0x0014},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 16000, 96,  0x0016},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 16000, 128, 0x0054},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 16000, 192, 0x0056},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 22050, 64,  0x0034},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 22050, 96,  0x0036},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 22050, 128, 0x0074},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 22050, 192, 0x0076},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 24000, 64,  0x0038},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 24000, 96,  0x003a},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 24000, 128, 0x0078},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 24000, 192, 0x007a},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 32000,  64, 0x0018},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 32000,  96, 0x001a},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 32000, 128, 0x0058},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 32000, 192, 0x005a},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 44100,  64, 0x0020},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 44100,  96, 0x0022},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 44100, 128, 0x0060},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 44100, 192, 0x0062},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 48000,  64, 0x0000},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 48000,  96, 0x0002},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 48000, 128, 0x0040},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 48000, 192, 0x0042},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 88200,  32, 0x003c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 88200,  48, 0x003e},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 88200,  64, 0x007c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 88200,  96, 0x007e},

	{WM8750_HWFMT, SND_SOC_HWBITS(16), 96000,  32, 0x001c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 96000,  48, 0x001e},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 96000,  64, 0x005c},
	{WM8750_HWFMT, SND_SOC_HWBITS(16), 96000,  96, 0x005e},
};

/*
 * read wm8750 register cache
 */
static inline unsigned int wm8750_read_reg_cache(struct snd_soc_codec * codec, 
	unsigned int reg)
{
	u16 *cache = codec->reg_cache;
	if (reg > WM8750_CACHE_REGNUM)
		return -1;
	return cache[reg];
}

/*
 * write wm8750 register cache
 */
static inline void wm8750_write_reg_cache(struct snd_soc_codec * codec, 
	unsigned int reg, unsigned int value)
{
	u16 *cache = codec->reg_cache;
	if (reg > WM8750_CACHE_REGNUM)
		return;
	cache[reg] = value;
}

static int wm8750_write(struct snd_soc_codec * codec, unsigned int reg, 
	unsigned int value)
{
	u8 data[2];
	struct i2c_client *i2c = (struct i2c_client*)codec->control_data;

	/* data is 
	 *   D15..D9 WM8750 register offset
	 *   D8...D0 register data
	 */
	data[0] = (reg << 1) | ((value >> 8) & 0x0001);
	data[1] = value & 0x00ff;

	wm8750_write_reg_cache (codec, reg, value);
	if (i2c_master_send(i2c, data, 2) == 2)
		return 0;
	else
		return -1;
}

#define wm8750_reset(c)	wm8750_write(c, WM8750_RESET, 0)

/*
 * WM8750 Controls
 */
static const char* wm8750_bass[] = {"Linear Control", "Adaptive Boost"};
static const char* wm8750_bass_filter[] = { "130Hz @ 48kHz", "200Hz @ 48kHz" };
static const char* wm8750_treble[] = {"8kHz", "4kHz"};
static const char* wm8750_3d_lc[] = {"200Hz", "500Hz"};
static const char* wm8750_3d_uc[] = {"2.2kHz", "1.5kHz"};
static const char* wm8750_3d_func[] = {"Capture", "Playback"};
static const char* wm8750_alc_func[] = {"Off", "Right", "Left", "Stereo"};
static const char* wm8750_ng_type[] = {"Constant PGA Gain", "Mute ADC Output"};
static const char* wm8750_line_mux[] = {"Line 1", "Line 2", "Line 3", "PGA", "Differential"};
static const char* wm8750_pga_sel[] = {"Line 1", "Line 2", "Line 3", "Differential"};
static const char* wm8750_out3[] = {"VREF", "ROUT1 + Vol", "MonoOut", "ROUT1"};
static const char* wm8750_diff_sel[] = {"Line 1", "Line 2"};
static const char* wm8750_adcpol[] = {"Normal", "L Invert", "R Invert", "L + R Invert"};
static const char* wm8750_deemph[] = {"None", "32Khz", "44.1Khz", "48Khz"};
static const char* wm8750_mono_mux[] = {"Stereo", "Mono (Left)", "Mono (Right)", "Digital Mono"};

static const struct soc_enum wm8750_enum[] = {
SOC_ENUM_SINGLE(WM8750_BASS, 7, 2, wm8750_bass),
SOC_ENUM_SINGLE(WM8750_BASS, 6, 2, wm8750_bass_filter),
SOC_ENUM_SINGLE(WM8750_TREBLE, 6, 2, wm8750_treble),
SOC_ENUM_SINGLE(WM8750_3D, 5, 2, wm8750_3d_lc),
SOC_ENUM_SINGLE(WM8750_3D, 6, 2, wm8750_3d_uc),
SOC_ENUM_SINGLE(WM8750_3D, 7, 2, wm8750_3d_func),
SOC_ENUM_SINGLE(WM8750_ALC1, 7, 4, wm8750_alc_func),
SOC_ENUM_SINGLE(WM8750_NGATE, 1, 2, wm8750_ng_type),
SOC_ENUM_SINGLE(WM8750_LOUTM1, 0, 5, wm8750_line_mux),
SOC_ENUM_SINGLE(WM8750_ROUTM1, 0, 5, wm8750_line_mux),
SOC_ENUM_SINGLE(WM8750_LADCIN, 6, 4, wm8750_pga_sel), /* 10 */
SOC_ENUM_SINGLE(WM8750_RADCIN, 6, 4, wm8750_pga_sel),
SOC_ENUM_SINGLE(WM8750_ADCTL2, 7, 4, wm8750_out3),
SOC_ENUM_SINGLE(WM8750_ADCIN, 8, 2, wm8750_diff_sel),
SOC_ENUM_SINGLE(WM8750_ADCDAC, 5, 4, wm8750_adcpol),
SOC_ENUM_SINGLE(WM8750_ADCDAC, 1, 4, wm8750_deemph),
SOC_ENUM_SINGLE(WM8750_ADCIN, 6, 4, wm8750_mono_mux), /* 16 */

};

static const snd_kcontrol_new_t wm8750_snd_controls[] = {

SOC_SINGLE("Left Capture Volume", WM8750_LINVOL, 0, 63, 0),
SOC_SINGLE("Left Capture ZC Switch", WM8750_LINVOL, 6, 1, 0),
SOC_SINGLE("Left Capture Switch", WM8750_LINVOL, 7, 1, 1),
SOC_SINGLE("Right Capture Volume", WM8750_RINVOL, 0, 63, 0),
SOC_SINGLE("Right Capture ZC Switch", WM8750_RINVOL, 6, 1, 0),
SOC_SINGLE("Right Capture Switch", WM8750_RINVOL, 7, 1, 1),

SOC_SINGLE("Left Out1 Playback ZC Switch", WM8750_LOUT1V, 7, 1, 0),
SOC_SINGLE("Right Out1 Playback ZC Switch", WM8750_ROUT1V, 7, 1, 0),

SOC_SINGLE("Left Out2 Playback ZC Switch", WM8750_LOUT2V, 7, 1, 0),
SOC_SINGLE("Right Out2 Playback ZC Switch", WM8750_ROUT2V, 7, 1, 0),

SOC_ENUM("Playback De-emphasis", wm8750_enum[15]),

SOC_ENUM("Capture Polarity", wm8750_enum[14]),
SOC_SINGLE("Playback 6dB Attenuate", WM8750_ADCDAC, 7, 1, 0),
SOC_SINGLE("Capture 6dB Attenuate", WM8750_ADCDAC, 8, 1, 0),

SOC_SINGLE("Left PCM Volume", WM8750_LDAC, 0, 255, 0),
SOC_SINGLE("Right PCM Volume", WM8750_RDAC, 0, 255, 0),

SOC_ENUM("Bass Boost", wm8750_enum[0]),
SOC_ENUM("Bass Filter", wm8750_enum[1]),
SOC_SINGLE("Bass Volume", WM8750_BASS, 0, 15, 1),

SOC_SINGLE("Treble Volume", WM8750_TREBLE, 0, 15, 0),
SOC_ENUM("Treble Cut-off", wm8750_enum[2]),

SOC_SINGLE("3D Switch", WM8750_3D, 0, 1, 0),
SOC_SINGLE("3D Volume", WM8750_3D, 1, 15, 0),
SOC_ENUM("3D Lower Cut-off", wm8750_enum[3]),
SOC_ENUM("3D Upper Cut-off", wm8750_enum[4]),
SOC_ENUM("3D Mode", wm8750_enum[5]),

SOC_SINGLE("ALC Capture Target Volume", WM8750_ALC1, 0, 7, 0),
SOC_SINGLE("ALC Capture Max Volume", WM8750_ALC1, 4, 7, 0),
SOC_ENUM("ALC Capture Function", wm8750_enum[6]),
SOC_SINGLE("ALC Capture ZC Switch", WM8750_ALC2, 7, 1, 0),
SOC_SINGLE("ALC Capture Hold Time", WM8750_ALC2, 0, 15, 0),
SOC_SINGLE("ALC Capture Decay Time", WM8750_ALC3, 4, 15, 0),
SOC_SINGLE("ALC Capture Attack Time", WM8750_ALC3, 0, 15, 0),
SOC_SINGLE("ALC Capture NG Threshold", WM8750_NGATE, 3, 31, 0),
SOC_ENUM("ALC Capture NG Type", wm8750_enum[4]),
SOC_SINGLE("ALC Capture NG Switch", WM8750_NGATE, 0, 1, 0),

SOC_SINGLE("Left ADC Capture Volume", WM8750_LADC, 0, 255, 0),
SOC_SINGLE("Right ADC Capture Volume", WM8750_RADC, 0, 255, 0),

SOC_SINGLE("ZC Timeout Switch", WM8750_ADCTL1, 0, 1, 0),
SOC_SINGLE("Playback Invert Switch", WM8750_ADCTL1, 1, 1, 0),

SOC_SINGLE("Right Out2 Playback Invert Switch", WM8750_ADCTL2, 4, 1, 0),


/* Unimplemented */
/* ADCDAC Bit 0 - ADCHPD */
/* ADCDAC Bit 4 - HPOR */
/* ADCTL1 Bit 2,3 - DATSEL */
/* ADCTL1 Bit 4,5 - DMONOMIX */
/* ADCTL1 Bit 6,7 - VSEL */
/* ADCTL2 Bit 2 - LRCM */
/* ADCTL2 Bit 3 - TRI */
/* ADCTL3 Bit 5 - HPFLREN */
/* ADCTL3 Bit 6 - VROI */
/* ADCTL3 Bit 7,8 - ADCLRM */
/* ADCIN Bit 4 - LDCM */ 
/* ADCIN Bit 5 - RDCM */

SOC_SINGLE("Left Mic Boost", WM8750_LADCIN, 4, 3, 0),
SOC_SINGLE("Right Mic Boost", WM8750_RADCIN, 4, 3, 0),

SOC_SINGLE("Left Bypass Left Playback Volume", WM8750_LOUTM1, 4, 7, 1),
SOC_SINGLE("Right Bypass Left Playback Volume", WM8750_LOUTM2, 4, 7, 1),

SOC_SINGLE("Left Bypass Right Playback Volume", WM8750_ROUTM1, 4, 7, 1),
SOC_SINGLE("Right Bypass Right Playback Volume", WM8750_ROUTM2, 4, 7, 1),

SOC_SINGLE("Left Bypass Mono Playback Volume", WM8750_MOUTM1, 4, 7, 1),
SOC_SINGLE("Right Bypass Mono Playback Volume", WM8750_MOUTM2, 4, 7, 1), 

SOC_SINGLE("Mono Playback ZC Switch", WM8750_MOUTV, 7, 1, 0),

};

/* add non dpm controls */
static int wm8750_add_controls(struct snd_soc_codec *codec)
{
	int err, i;

	for (i = 0; i < ARRAY_SIZE(wm8750_snd_controls); i++) {
		if ((err = snd_ctl_add(codec->card, snd_soc_cnew(&wm8750_snd_controls[i],codec, NULL))) < 0)
			return err;
	}
	return 0;
}

/*
 * DPM Controls
 */

/* Left Mixer */
static const snd_kcontrol_new_t wm8750_left_mixer_controls[] = {
SOC_DPM_SINGLE("Playback Switch", WM8750_LOUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Left Bypass Switch", WM8750_LOUTM1, 7, 1, 0),
SOC_DPM_SINGLE("Right Playback Switch", WM8750_LOUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Right Bypass Switch", WM8750_LOUTM2, 7, 1, 0),
};

/* Right Mixer */
static const snd_kcontrol_new_t wm8750_right_mixer_controls[] = {
SOC_DPM_SINGLE("Left Playback Switch", WM8750_ROUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Left Bypass Switch", WM8750_ROUTM1, 7, 1, 0),
SOC_DPM_SINGLE("Playback Switch", WM8750_ROUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Right Bypass Switch", WM8750_ROUTM2, 7, 1, 0),
};

/* Mono Mixer */
static const snd_kcontrol_new_t wm8750_mono_mixer_controls[] = {
SOC_DPM_SINGLE("Left Playback Switch", WM8750_MOUTM1, 8, 1, 0),
SOC_DPM_SINGLE("Left Bypass Switch", WM8750_MOUTM1, 7, 1, 0),
SOC_DPM_SINGLE("Right Playback Switch", WM8750_MOUTM2, 8, 1, 0),
SOC_DPM_SINGLE("Right Bypass Switch", WM8750_MOUTM2, 7, 1, 0),
};

/* Left Line Mux */
static const snd_kcontrol_new_t wm8750_left_line_controls =
SOC_DPM_ENUM("Route", wm8750_enum[8]);

/* Right Line Mux */
static const snd_kcontrol_new_t wm8750_right_line_controls =
SOC_DPM_ENUM("Route", wm8750_enum[9]);

/* Left PGA Mux */
static const snd_kcontrol_new_t wm8750_left_pga_controls =
SOC_DPM_ENUM("Route", wm8750_enum[10]);

/* Right PGA Mux */
static const snd_kcontrol_new_t wm8750_right_pga_controls =
SOC_DPM_ENUM("Route", wm8750_enum[11]);

/* Out 3 Mux */
static const snd_kcontrol_new_t wm8750_out3_controls =
SOC_DPM_ENUM("Route", wm8750_enum[12]);

/* Differential Mux */
static const snd_kcontrol_new_t wm8750_diffmux_controls =
SOC_DPM_ENUM("Route", wm8750_enum[13]);

/* Mono ADC Mux */
static const snd_kcontrol_new_t wm8750_monomux_controls =
SOC_DPM_ENUM("Route", wm8750_enum[16]);

/* LOUT 1 Volume */
static const snd_kcontrol_new_t wm8750_lout1_vol_controls =
SOC_DPM_SINGLE("Left Out1 Playback Volume", WM8750_LOUT1V, 0, 127, 0);

/* ROUT 1 Volume */
static const snd_kcontrol_new_t wm8750_rout1_vol_controls =
SOC_DPM_SINGLE("Right Out1 Playback Volume", WM8750_ROUT1V, 0, 127, 0);

/* LOUT 2 Volume */
static const snd_kcontrol_new_t wm8750_lout2_vol_controls =
SOC_DPM_SINGLE("Left Out2 Playback Volume", WM8750_LOUT2V, 0, 127, 0);

/* ROUT 2 Volume */
static const snd_kcontrol_new_t wm8750_rout2_vol_controls =
SOC_DPM_SINGLE("Right Out2 Playback Volume", WM8750_ROUT2V, 0, 127, 0);

/* Mono Volume */
static const snd_kcontrol_new_t wm8750_mout_vol_controls =
SOC_DPM_SINGLE("Mono Playback Volume", WM8750_MOUTV, 0, 127, 0);

static const struct snd_soc_dpm_widget wm8750_dpm_widgets[] = {
	SND_SOC_DPM_CLOCK("WM8750 Clock", WM8750_PWR1, 0, 1),
	SND_SOC_DPM_MUTE("DAC Mute", WM8750_ADCDAC, 3, 0),
	SND_SOC_DPM_MIXER("Left Mixer", -1, 0, 0, &wm8750_left_mixer_controls[0], ARRAY_SIZE(wm8750_left_mixer_controls)),
	SND_SOC_DPM_MIXER("Right Mixer", -1, 0, 0, &wm8750_right_mixer_controls[0], ARRAY_SIZE(wm8750_right_mixer_controls)),
	SND_SOC_DPM_MIXER("Mono Mixer", WM8750_PWR2, 2, 0, &wm8750_mono_mixer_controls[0], ARRAY_SIZE(wm8750_mono_mixer_controls)),

	SND_SOC_DPM_VOLUME("Right Out 2", WM8750_PWR2, 3, 0, &wm8750_rout2_vol_controls, 1),
	SND_SOC_DPM_VOLUME("Left Out 2", WM8750_PWR2, 4, 0, &wm8750_lout2_vol_controls, 1),
	SND_SOC_DPM_VOLUME("Right Out 1", WM8750_PWR2, 5, 0, &wm8750_rout1_vol_controls, 1),
	SND_SOC_DPM_VOLUME("Left Out 1", WM8750_PWR2, 6, 0, &wm8750_lout1_vol_controls, 1),
	SND_SOC_DPM_DAC("Right DAC", "Right Playback", WM8750_PWR2, 7, 0, NULL, 0),
	SND_SOC_DPM_DAC("Left DAC", "Left Playback", WM8750_PWR2, 8, 0, NULL, 0),

	SND_SOC_DPM_MICBIAS("Mic Bias", WM8750_PWR1, 1, 0),
	SND_SOC_DPM_ADC("Right ADC", "Right Capture", WM8750_PWR1, 2, 0, NULL, 0),
	SND_SOC_DPM_ADC("Left ADC", "Left Capture", WM8750_PWR1, 3, 0, NULL, 0),

	SND_SOC_DPM_MUX("Left PGA Mux", WM8750_PWR1, 5, 0, &wm8750_left_pga_controls),
	SND_SOC_DPM_MUX("Right PGA Mux", WM8750_PWR1, 4, 0, &wm8750_right_pga_controls),
	SND_SOC_DPM_MUX("Left Line Mux", -1, 0, 0, &wm8750_left_line_controls),
	SND_SOC_DPM_MUX("Right Line Mux", -1, 0, 0, &wm8750_right_line_controls),

	SND_SOC_DPM_MUX("Out3 Mux", -1, 0, 0, &wm8750_out3_controls),
	SND_SOC_DPM_VOLUME("Out 3", WM8750_PWR2, 1, 0, NULL, 0),
	SND_SOC_DPM_VOLUME("Mono Out 1", WM8750_PWR2, 2, 0, &wm8750_mout_vol_controls, 1),
	
	SND_SOC_DPM_MUX("Differential Mux", -1, 0, 0, &wm8750_diffmux_controls),
	SND_SOC_DPM_MUX("Left ADC Mux", -1, 0, 0, &wm8750_monomux_controls),
	SND_SOC_DPM_MUX("Right ADC Mux", -1, 0, 0, &wm8750_monomux_controls),

	SND_SOC_DPM_OUTPUT("LOUT1"),
	SND_SOC_DPM_OUTPUT("ROUT1"),
	SND_SOC_DPM_OUTPUT("LOUT2"),
	SND_SOC_DPM_OUTPUT("ROUT2"),
	SND_SOC_DPM_OUTPUT("MONO"),
	SND_SOC_DPM_OUTPUT("OUT3"),

	SND_SOC_DPM_INPUT("LINPUT1"),
	SND_SOC_DPM_INPUT("LINPUT2"),
	SND_SOC_DPM_INPUT("LINPUT3"),
	SND_SOC_DPM_INPUT("RINPUT1"),
	SND_SOC_DPM_INPUT("RINPUT2"),
	SND_SOC_DPM_INPUT("RINPUT3"),
};

static const char* intercon[][3] = {
	/* left mixer */
	{"Left Mixer", "Playback Switch", "Left DAC"},
	{"Left Mixer", "Left Bypass Switch", "Left Line Mux"},
	{"Left Mixer", "Right Playback Switch", "Right DAC"},
	{"Left Mixer", "Right Bypass Switch", "Right Line Mux"},

	/* right mixer */
	{"Right Mixer", "Left Playback Switch", "Left DAC"},
	{"Right Mixer", "Left Bypass Switch", "Left Line Mux"},
	{"Right Mixer", "Playback Switch", "Right DAC"},
	{"Right Mixer", "Right Bypass Switch", "Right Line Mux"},

	/* left out 1 */
	{"Left Out 1", NULL, "Left Mixer"},
	{"LOUT1", NULL, "Left Out 1"},

	/* left out 2 */
	{"Left Out 2", NULL, "Left Mixer"},
	{"LOUT2", NULL, "Left Out 2"},
	
	/* right out 1 */
	{"Right Out 1", NULL, "Right Mixer"},
	{"ROUT1", NULL, "Right Out 1"},

	/* right out 2 */
	{"Right Out 2", NULL, "Right Mixer"},
	{"ROUT2", NULL, "Right Out 2"},

	/* mono mixer */
	{"Mono Mixer", "Left Playback Switch", "Left DAC"},
	{"Mono Mixer", "Left Bypass Switch", "Left Line Mux"},
	{"Mono Mixer", "Right Playback Switch", "Right DAC"},
	{"Mono Mixer", "Right Bypass Switch", "Right Line Mux"},
	
	/* mono out */
	{"Mono Out 1", NULL, "Mono Mixer"},
	{"MONO1", NULL, "Mono Out 1"},

	/* out 3 */
	{"Out3 Mux", "VREF", "VREF"},
	{"Out3 Mux", "ROUT1 + Vol", "ROUT1"},
	{"Out3 Mux", "ROUT1", "Right Mixer"},
	{"Out3 Mux", "MonoOut", "MONO1"},
	{"Out 3", NULL, "Out3 Mux"},
	{"OUT3", NULL, "Out 3"},

	/* Left Line Mux */
	{"Left Line Mux", "Line 1", "LINPUT1"},
	{"Left Line Mux", "Line 2", "LINPUT2"},
	{"Left Line Mux", "Line 3", "LINPUT3"},
	{"Left Line Mux", "PGA", "Left PGA Mux"},
	{"Left Line Mux", "Differential", "Differential Mux"},

	/* Right Line Mux */
	{"Right Line Mux", "Line 1", "RINPUT1"},
	{"Right Line Mux", "Line 2", "RINPUT2"},
	{"Right Line Mux", "Line 3", "RINPUT3"},
	{"Right Line Mux", "PGA", "Right PGA Mux"},
	{"Right Line Mux", "Differential", "Differential Mux"},

	/* Left PGA Mux */
	{"Left PGA Mux", "Line 1", "LINPUT1"},
	{"Left PGA Mux", "Line 2", "LINPUT2"},
	{"Left PGA Mux", "Line 3", "LINPUT3"},
	{"Left PGA Mux", "Differential", "Differential Mux"},

	/* Right PGA Mux */
	{"Right PGA Mux", "Line 1", "RINPUT1"},
	{"Right PGA Mux", "Line 2", "RINPUT2"},
	{"Right PGA Mux", "Line 3", "RINPUT3"},
	{"Right PGA Mux", "Differential", "Differential Mux"},

	/* Differential Mux */
	{"Differential Mux", "Line 1", "LINPUT1"},
	{"Differential Mux", "Line 1", "RINPUT1"},
	{"Differential Mux", "Line 2", "LINPUT2"},
	{"Differential Mux", "Line 2", "RINPUT2"},

	/* Left ADC Mux */
	{"Left ADC Mux", "Stereo", "Left PGA Mux"},
	{"Left ADC Mux", "Mono (Left)", "Left PGA Mux"},
	{"Left ADC Mux", "Digital Mono", "Left PGA Mux"},

	/* Right ADC Mux */
	{"Right ADC Mux", "Stereo", "Right PGA Mux"},
	{"Right ADC Mux", "Mono (Right)", "Right PGA Mux"},
	{"Right ADC Mux", "Digital Mono", "Right PGA Mux"},

	/* ADC */
	{"Left ADC", NULL, "Left ADC Mux"},
	{"Right ADC", NULL, "Right ADC Mux"},
	
	/* terminator */
	{NULL, NULL, NULL},
};

static int wm8750_add_widgets(struct snd_soc_codec *codec)
{
	int i;
	
	for(i = 0; i < ARRAY_SIZE(wm8750_dpm_widgets); i++) {
		snd_soc_dpm_new_control(codec, &wm8750_dpm_widgets[i]);
	}
	
	/* set up audio path interconnects */
	for(i = 0; intercon[i][0] != NULL; i++) {
		snd_soc_dpm_connect_input(codec, intercon[i][0], intercon[i][1], intercon[i][2]);
	}

	snd_soc_dpm_sync(codec);
	return 0;
}


static int wm8750_pcm_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_codec *codec = pcm_c->codec;
	u16 iface = 0;
	
	/* set master/slave audio interface */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK) {
		case SND_SOC_HWFMT_CBM_CFM:
			iface |= 0x0040;
			break;
		case SND_SOC_HWFMT_CBS_CFS:
			break;
	}
	
	/* interface format */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) {
		case SND_SOC_HWFMT_I2S:
			iface |= 0x0002;
			break;
		case SND_SOC_HWFMT_RIGHT_J:
			break;
		case SND_SOC_HWFMT_LEFT_J:
			iface |= 0x0001;
			break;
		case SND_SOC_HWFMT_DSP_A:
			iface |= 0x0003;
			break;
		case SND_SOC_HWFMT_DSP_B:
			iface |= 0x0013;
			break;
	}
		
	/* bit size */
	switch(pcm_c->hw_runtime.hbits) {
		case SND_SOC_HWBITS(16):
			break;
		case SND_SOC_HWBITS(20):
			iface |= 0x0004;
			break;
		case SND_SOC_HWBITS(24):
			iface |= 0x0008;
			break;
		case SND_SOC_HWBITS(32):
			iface |= 0x000c;
			break;
	}

	/* clock inversion */
	switch(pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) {
		case SND_SOC_HWFMT_NB_NF:
			break;
		case SND_SOC_HWFMT_IB_IF:
			iface |= 0x0090;
			break;
		case SND_SOC_HWFMT_IB_NF:
			iface |= 0x0080;
			break;
		case SND_SOC_HWFMT_NB_IF:
			iface |= 0x0010;
			break;
	}
	
	/* set rate */
	wm8750_write(codec, WM8750_SRATE, pcm_c->hw_runtime.priv1);
	wm8750_write(codec, WM8750_IFACE, iface);
	
	return 0;
}

static void wm8750_device_release(struct device * dev)
{
}

static int wm8750_dpm_event(struct snd_soc_codec *codec, int event)
{
	u16 pwr_reg = wm8750_read_reg_cache(codec, WM8750_PWR1) & 0xfe3f;
	
	switch (event) {
		case SNDRV_CTL_POWER_D0: /* full On */
			/* set vmid to 50k and unmute dac */
			wm8750_write(codec, WM8750_PWR1, pwr_reg | 0x00c0);
			break;
		case SNDRV_CTL_POWER_D1: /* partial On */
		case SNDRV_CTL_POWER_D2: /* partial On */
			/* set vmid to 5k for quick power up */
			wm8750_write(codec, WM8750_PWR1, pwr_reg | 0x01c0);
			break;
		case SNDRV_CTL_POWER_D3hot: /* Off, with power */
			/* mute dac and set vmid to 500k, enable VREF */
			wm8750_write(codec, WM8750_PWR1, pwr_reg | 0x0140);
			break;
		case SNDRV_CTL_POWER_D3cold: /* Off, without power */
			wm8750_write(codec, WM8750_PWR1, 0x0000);
			break;
	}
	
	return 0;
}

static struct snd_soc_pcm_codec wm8750_pcm_client = {
	.name = "WM8750",
	.playback = {
		.sname = "Playback",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 |
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,
	},
	.nplayback = 1,
	.capture = {
		.sname = "Capture",
		.rates = SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_16000 | 
		SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_48000 | SNDRV_PCM_RATE_44100 |
		SNDRV_PCM_RATE_96000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,
		.channels_min = 1,
		.channels_max = 2,
	},
	.ncapture = 1,
	.ops = {
		.prepare = wm8750_pcm_prepare,
	},
	.hw = {
		.num_hmodes = ARRAY_SIZE(wm8750_hwfmt),
		.hmodes = &wm8750_hwfmt[0],
	},
};

/*
 * initialise the WM8750 driver
 * register the mixer and dsp interfaces with the kernel 
 */
static int wm8750_probe(struct snd_soc_codec *codec)
{
	int reg;
	
	info("WM8750 Audio Codec %s", WM8750_VERSION);
	
	wm8750_reset(codec);
	
	/* charge output caps */
	wm8750_dpm_event(codec, SNDRV_CTL_POWER_D2);

	/* set the update bits */
	reg = wm8750_read_reg_cache(codec, WM8750_LDAC);
	wm8750_write(codec, WM8750_LDAC, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_RDAC);
	wm8750_write(codec, WM8750_RDAC, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_LOUT1V);
	wm8750_write(codec, WM8750_LOUT1V, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_ROUT1V);
	wm8750_write(codec, WM8750_ROUT1V, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_LOUT2V);
	wm8750_write(codec, WM8750_LOUT2V, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_ROUT2V);
	wm8750_write(codec, WM8750_ROUT2V, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_LINVOL);
	wm8750_write(codec, WM8750_LINVOL, reg | 0x0100);
	reg = wm8750_read_reg_cache(codec, WM8750_RINVOL);
	wm8750_write(codec, WM8750_RINVOL, reg | 0x0100);

	wm8750_add_controls(codec);
	wm8750_add_widgets(codec);
	
	/* wait for caps to finish charging - liam make this thread */
	ssleep(1);
	wm8750_dpm_event(codec, SNDRV_CTL_POWER_D3hot);

	return 0;
}

/* 
 * unregister interfaces and clean up
 */
static void wm8750_remove(struct snd_soc_codec *codec)
{
	wm8750_dpm_event(codec, SNDRV_CTL_POWER_D3cold);
}

/*
 * Attach WM8750 2 wire client 
 */
static int wm8750_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{		
	int ret = 0;
	struct snd_soc_codec *codec;
	struct i2c_client *i2c;

	client_template.adapter = adap;
	client_template.addr = addr;
	
	if ((codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL)) == NULL)
		return -ENOMEM;
 
	if ((i2c = kzalloc(sizeof(struct i2c_client), GFP_KERNEL)) == NULL){
		kfree(codec);
		return -ENOMEM;
	}
	memcpy(i2c, &client_template, sizeof(struct i2c_client));
	codec->control_data = i2c;
	i2c->dev.platform_data = codec;

	if((ret = i2c_attach_client(i2c)) < 0) {
		err("failed to attach codec at addr %x\n", addr);
		kfree(i2c);
		kfree(codec);
		return ret;
	}

	codec->name = "WM8750";
	codec->longname = "WM8750 Portable Codec";
	codec->owner = THIS_MODULE;
	codec->pdev.name = "alsa-wm8750";
	codec->pdev.id = -1;
	codec->pdev.dev.release = wm8750_device_release;
	codec->probe = wm8750_probe;
	codec->remove = wm8750_remove;
	codec->read = wm8750_read_reg_cache;
	codec->write = wm8750_write;
	codec->dpm_event = wm8750_dpm_event;
	codec->pcms = &wm8750_pcm_client;
	codec->npcms = 1;
	if ((codec->reg_cache = 
			kzalloc(sizeof(u16) * ARRAY_SIZE(wm8750_reg), GFP_KERNEL)) == NULL)
		return -ENOMEM;
	memcpy(codec->reg_cache, wm8750_reg, sizeof(u16) * ARRAY_SIZE(wm8750_reg));

	if((ret = snd_soc_register_codec(codec)) < 0) {
		err("can't register codec");
		i2c_del_driver(&wm8750_i2c_driver);
		kfree(i2c);
		kfree(codec->reg_cache);
		kfree(codec);
	}

	return ret;
}

static int wm8750_i2c_detach(struct i2c_client *client)
{
	struct snd_soc_codec *codec = 
		(struct snd_soc_codec*)client->dev.platform_data;
			
	i2c_detach_client(client);
	snd_soc_unregister_codec(codec);
	kfree(codec->control_data);
	kfree(codec->reg_cache);
	kfree(codec);
	return 0;
}

static int wm8750_i2c_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, wm8750_i2c_probe);
}

/* WM8750 2 Wire layer */ 
static struct i2c_driver wm8750_i2c_driver = {
	.name =           "i2c wm8750 driver",
	.id =             I2C_DRIVERID_WM8750,
	.flags =          I2C_DF_NOTIFY,
	.attach_adapter = wm8750_i2c_attach,
	.detach_client =  wm8750_i2c_detach,
	.command =        NULL,
};

static struct i2c_client client_template = {
	.name =   "WM8750",
	.flags =  I2C_CLIENT_ALLOW_USE,
	.driver = &wm8750_i2c_driver,
};

static int __init wm8750_init(void)
{	
	int ret = 0;

	info("WM8750 Audio Codec %s", WM8750_VERSION);

	if ((ret = i2c_add_driver(&wm8750_i2c_driver)) != 0)
		err("can't add i2c driver");
	
	return ret;
}

static void __exit wm8750_exit(void)
{
	i2c_del_driver(&wm8750_i2c_driver);
}

module_init(wm8750_init);
module_exit(wm8750_exit);

MODULE_DESCRIPTION("Soc WM8750 driver");
MODULE_AUTHOR("Liam Girdwood");
MODULE_LICENSE("GPL");
